#pragma once

#include "2d/plot_2dtypes.h"
#include "widgets/plot_widgets.h"

#include <QComboBox>
#include <QMimeData>

namespace plot {

class TCalcLimitsByFrameContent
{
    inline void Clear();
    void CalcScrLimits(int N);
    void CalcScrLimits(const QwtPlotCurve &CurveData);
    inline bool IsIn(const QwtPlotCurve &CurveData, const int Index) const;

    TCurves *Curves;
    TInfoForAllPlots *PlotInfo;
    QVector<double> XMins, YMins, XMaxs, YMaxs;
    double MinPlotsX, MaxPlotsX, MinPlotsY, MaxPlotsY;
public:
    void operator ()(TInfoForAllPlots *InfoForAll, TCurves *Src, const int ApplyTo, int N = -1);
};

class TComboBox: public QComboBox
{
    void hidePopup() { QComboBox::hidePopup(); emit activated(currentIndex()); }
    void showPopup() { QComboBox::showPopup(); emit highlighted(currentIndex()); }
};

class T2dPlotControl: public TCommonPlotControl
{
    void CreateActionsAndMenus();

public:
    void GUICreate();

    QPushButton  *BResetMCCurve;

    QComboBox    *CBChooseActiveCurve;
    QPushButton  *BAutoScaleX;
    QPushButton  *BAutoScaleY;
    QActionGroup *AGFitOrders;

    QToolButton  *TBCurveColour;
    QToolButton  *TBCurveThickness;
    QActionGroup *AGCurveColours;
    QActionGroup *AGCurveThickness;

};

class T2dMime : public QMimeData
{
    Q_OBJECT
public:
    TCurvesData data;
};

class TInfoFor2dPlot : virtual public TInfoForAllPlots
{
public:
    TSIUnit XAxisUnit;
    TSIUnit YAxisUnit;

    void UpdateAxisLegends() { if (AutoLegendEnabled) { XLegend = XAxisUnit.Legend; YLegend = YAxisUnit.Legend; } }
};

class TPure2dPlotPrivate;
class T2dPlot;
namespace math { class TMathProcessing; }
class TPure2dPlot : public TCommonPlot, public TInfoFor2dPlot, public TCurves
{
    Q_OBJECT

    friend class T2dPlot;

public slots:
    void s_Refresh();
    void ResetMCCurve();

    void on_BAutoScaleX_clicked();
    void on_BAutoScaleY_clicked();
    void on_BExpandSelectRect_clicked()  { ExpandSelectionRect(); ReDrawSelectedRectangle(TSelectEvent::SRO_Mouse); }

signals:
    void MathUpdated();
    void ShowPlotControl();
    void sync_active_curve(int);

public:
    enum TModes {Mode_ManualConstruct, Mode_Select, Mode_Normal};

protected:
    TPure2dPlotPrivate * const d_ptr;
    void dragEnterEvent(QDragEnterEvent *event);
    void dropEvent(QDropEvent *event);

public:
    Q_DECLARE_PRIVATE(TPure2dPlot);

    void ReSetAxisScales();
    using TCommonPlot::UpdateLabels;
    void UpdateLabels(bool IsDetailedInfoVisible);
    void HandleAxisMultiplier();
    void UpdateAxisLegends();
    void SetAutoScaleEnabled(bool Enabled);
    void SetZoomBySelection(const bool State);

    bool eventFilter(QObject *Object, QEvent *Event);

    void TraceCurve(const QPoint &Pos, int NOfCurve = 0);
    void SelectionApplied();
    void SetFitOrder(int Order);
    void EnableSelectMode(const bool State);
    void SelectRectChanged(const QRectF &Rect);

    void StartMCMode();
    void AddPointToMCCurve(const int After, const QPointF &p2);
    void MoveSelectedPointInMCCurve(const QPoint &Pos);
    void ExitMCMode();
    TCurveData GetMCCurve() const;

    void DefineClickInfo(const QPoint &Pos);

    virtual void SetActiveCurve(int Active);

    math::TMathProcessing *MathProcessing;
    bool isMathReady();
    int FitOrder;
    TCalcLimitsByFrameContent LimitsStretcher;

    TPure2dPlot(QObject *Parent = 0, QWidget *parent = 0);
    ~TPure2dPlot();
    QObject *Parent; //for postEvent(Parent, blablabla);

    void Refresh();
    void RefreshInc();
    void paintEvent(QPaintEvent *e);
    bool incRefreshGuard;

    void setVertLineVisible(int N, bool v);
    void SetNumOfVertLines(int N);
    void MoveVertLine(double XPos, int N);
    void SetVertLinePen(QPen &Pen, int N = 0);
    QVector<QwtPlotMarker *> VertLines;
};

class T2dPlot: public QWidget
{
    Q_OBJECT
public slots:
    void on_BFixX_toggled()          { Pure2dPlot->BFixXIsChecked = PlotControl->BFixX->isChecked(); }
    void on_BFixY_toggled()          { Pure2dPlot->BFixYIsChecked = PlotControl->BFixY->isChecked(); }
    void on_FontSize_changed()       { Pure2dPlot->SetFontSize(PlotControl->AGFontSizes->checkedAction()->text().toInt()); }
    void on_CurveThickness_changed() {
        for (int i = 0; i < Pure2dPlot->SourceData.size(); i++)
            if (Pure2dPlot->SourceData.at(i).data())
                Pure2dPlot->SourceData.at(i).data()->Pen.setWidth(PlotControl->AGCurveThickness->checkedAction()->text().toInt());
        Refresh();
    }

    void on_BSave_clicked();
    void on_FitOrder_changed();
    void on_BMarkers_clicked(bool);
    void UpdateAllMarkersInfo();
    void on_BAutoScale_toggled();
    void on_CurveColour_changed();
    void on_BSavePlotPixmap_clicked();
    void on_BDetailedMarkersInfo_clicked();
    void on_CBChooseActiveCurve_highlighted(int Index);
    void on_SyncTimer_timeout();
    void on_CBChooseActiveCurve_activated(int Index);
    void sync_active_curve(int);

public slots:
    void on_BSelectByRect_clicked();
    void on_BCancel_clicked();
    void StartMCCurveConstructing();
    void on_BApply_clicked();
    void SetSelectionModeEnabled(bool Enabled);
    void HideSelectionIfAny()                                               { if (Pure2dPlot->SelectionModeEnabled) on_BCancel_clicked(); }

signals:
    void AddDataToContainer(plot::TCurvesDataOwner, plot::TMetaStorage);
    void ManualConstructingDone();
    void CurveColourChanged(QColor Colour, int AC);

public:
    void GUICreate(bool child = false);
    void ConnectPlotControlSlots();
    void FillCBChooseActiveCurve();
    void PreparePlotControlForMC(bool MCStarts);
    void RestoreEnvironmentState();

    QObject *Parent;
    QTimer *SyncTimer;
    TPure2dPlot *Pure2dPlot;
    TMarkersInfo *MarkersInfo;
    T2dPlotControl *PlotControl;

    T2dPlot(QObject *Parent = 0, bool child = false, QWidget *wparent = 0);
    void Refresh() { Pure2dPlot->Refresh(); }

    void SetGridVisible(const bool Visible);
    void SetAutoScalingEnabled(const bool AutoScaling);
    QImage MakePreview() { return Pure2dPlot->MakePreview(PlotControl); }
};

}
