#include "plot_2dtypes.h"
#include <numeric>
#include <algorithm>
#include <limits>

double max(const QVector<double> &data)
{
    if (!data.size())
        return -std::numeric_limits<double>::max();

    return *std::max_element(data.constBegin(), data.constEnd());
}
double min(const QVector<double> &data)
{
    if (!data.size())
        return std::numeric_limits<double>::max();

    return *std::min_element(data.constBegin(), data.constEnd());
}
QPair<double, double> minmax(const QVector<double> &data)
{
    if (!data.size())
        return QPair<double, double>(std::numeric_limits<double>::max(), -std::numeric_limits<double>::max());

    std::pair<QVector<double>::const_iterator, QVector<double>::const_iterator> minmax;
    minmax = std::minmax_element(data.constBegin(), data.constEnd());

    return QPair<double, double>(*minmax.first, *minmax.second);
}
double rms(const QVector<double> &data)
{
    double disp = 0;
    double average = avg(data);

    for (int i = 0; i < data.size(); i++)
        disp += pow(data.at(i) - average, 2);

    if (data.size())
        disp /= data.size();

    return sqrt(disp);
}
double avg(const QVector<double> &data)
{
    double Average = std::accumulate(data.constBegin(), data.constEnd(), 0.);

    if (data.size())
        Average /= data.size();

    return Average;
}

namespace plot {

TCurves::TCurves(QwtPlot *_Plot)
{
    ActiveCurve = -1;
    Plot = _Plot;
    MCCurve = new QwtPlotCurve;
    MCCurve->setSymbol(new QwtSymbol(QwtSymbol::Ellipse, QBrush(Qt::gray), QPen(Qt::blue), QSize(8, 8)));
    MCCurve->attach(Plot);
}

void TCurves::VerifyActiveCurve()
{
    if (ActiveCurve >= QwtCurves.size()) {
        SetActiveCurve(-1);
        MCCurve->setSamples(QVector<QPointF>());
    }
}
void TCurves::GetClosestPoint(double *x, double *y, const QPoint &Pos, int N) const
{
    int idx = QwtCurves.at(N)->closestPoint(Pos, 0);
    if (idx < 0 || idx >= (int)QwtCurves.at(N)->dataSize()) {
        *x = *y = 0;
        return;
    }

    *x = QwtCurves.at(N)->sample(idx).x();
    *y = QwtCurves.at(N)->sample(idx).y();
}

//requires ownership in gui thread
class TQwtCurveData : public QwtSeriesData<QPointF>
{
public:
    TQwtCurveData(const QWeakPointer<TCurveData> &_data) : data(_data) { }
    int size() const {
        if (!data)
            return 0;

        return data.data()->size();
    }
    QPointF sample(int i) const {
        return QPointF(data.data()->XSeq.at(i), data.data()->YSeq.at(i));
    }
    QRectF boundingRect() const {
        if (!data)
            return QRectF();

        QPair<double, double> _minmaxx = minmax(data.data()->XSeq);
        double xmin = _minmaxx.first;
        double xmax = _minmaxx.second;
        QPair<double, double> _minmax = minmax(data.data()->YSeq);
        double ymin = _minmax.first;
        double ymax = _minmax.second;
        return QRectF(QPointF(xmin, ymin), QPointF(xmax, ymax));
    }

    QWeakPointer<TCurveData> data;
};
class TQwtCurveDataCompressedX : public TQwtCurveData
{
public:
    TQwtCurveDataCompressedX(const QWeakPointer<TCurveData> &_data) : TQwtCurveData(_data) { }
    QPointF sample(int i) const {
        return QPointF(data.data()->XSeq.at(0) + i * data.data()->XSeq.at(1), data.data()->YSeq.at(i));
    }
    QRectF boundingRect() const {
        if (!data)
            return QRectF();

        double xmin = data.data()->getXat(0);
        double xmax = data.data()->getXat(data.data()->size()-1);
        if (xmin > xmax)
            qSwap(xmin, xmax);
        QPair<double, double> _minmax = minmax(data.data()->YSeq);
        double ymin = _minmax.first;
        double ymax = _minmax.second;
        return QRectF(QPointF(xmin, ymin), QPointF(xmax, ymax));
    }
};

void TCurves::ReSyncDataAndQwtCurves()
{
    int target = PaintData.size();
    int curr = QwtCurves.size();

    if (!target)
        MCCurve->setSamples(QVector<QPointF>());

    for (int i = target; i < curr; i++)
        delete QwtCurves[i];

    QwtCurves.resize(target);
    LastPaintSize.resize(target);
    for (int i = curr; i < target; i++) {
        LastPaintSize[i] = 0;
        QwtCurves[i] = new QwtPlotCurve;
        QwtCurves[i]->setRenderHint(QwtPlotItem::RenderAntialiased, true);
        QwtCurves[i]->attach(Plot);
    }

    for (int i = 0; i < QwtCurves.size(); i++)
        if (PaintData.at(i)) {
            LastPaintSize[i] = PaintData.at(i).data()->size();
            if (PaintData.at(i).data()->isCompressedXCurve())
                QwtCurves[i]->setData(new TQwtCurveDataCompressedX(PaintData.at(i)));
            else
                QwtCurves[i]->setData(new TQwtCurveData(PaintData.at(i)));
            //QwtCurves[i]->setSamples(PaintData.at(i).data()->XSeq, PaintData.at(i).data()->YSeq);
            QwtCurves[i]->setPen(PaintData.at(i).data()->Pen);
        } else
            QwtCurves[i]->setSamples(QVector<QPointF>());
}
void TCurves::DeleteAllCurves()
{
    for (int i = 0; i < QwtCurves.size(); i++)
        delete QwtCurves[i];

    QwtCurves.clear();
    MCCurve->setSamples(QVector<QPointF>());
}
void TCurves::UnselectCurves()
{
    for (int i = 0; i < QwtCurves.size(); i++)
        if (PaintData.at(i))
            QwtCurves[i]->setPen(PaintData.at(i).data()->Pen);
}
void TCurves::SelectCurve(int N)
{
    for (int i = 0; i < QwtCurves.size(); i++)
        QwtCurves[i]->setPen(DefaultPenForInactiveCurves);

    if (PaintData.at(N)) {
        QPen FatPen = PaintData.at(N).data()->Pen;
        FatPen.setWidth(3);
        QwtCurves[N]->setPen(FatPen);
    }
}
void TCurves::DefineLimits(QVector<double> &MaxX, QVector<double> &MaxY, QVector<double> &MinX, QVector<double> &MinY, int N)
{
    if (N < 0)
        for (int i = 0; i < PaintData.size(); i++)
            DefineLimits(MaxX, MaxY, MinX, MinY, i);
    else if (N < PaintData.size() && PaintData.at(N)) {
        if (!PaintData.at(N).data()->isCompressedXCurve()) {
            QPair<double, double> _minmax = minmax(PaintData.at(N).data()->XSeq);
            MinX.push_back(_minmax.first);
            MaxX.push_back(_minmax.second);
        } else {
            double xmin = PaintData.at(N).data()->getXat(0);
            double xmax = PaintData.at(N).data()->getXat(PaintData.at(N).data()->size()-1);
            if (xmin > xmax)
                qSwap(xmin, xmax);
            MinX.push_back(xmin);
            MaxX.push_back(xmax);
        }
        QPair<double, double> _minmax = minmax(PaintData.at(N).data()->YSeq);
        MinY.push_back(_minmax.first);
        MaxY.push_back(_minmax.second);
    }
}

void TCurves::CleanData()
{
    TCurvesData data;

    for (int i = 0; i < SourceData.size(); i++)
        if (SourceData.at(i))
            data.append(SourceData.at(i));

    SourceData = data;
}

}
