#include "videoplayer.h"
#include "gui/status.hpp"
#include "tools.hpp"

#include "oscilloscope.h"
#include "remote_nst_dataid.hpp"
#include "mobile/remote_mobile.hpp"
#include <QCheckBox>
#include <QBuffer>
#include <QSettings>


void VideoPlayer::on_Camera_ImgReceived(QImage Img)
{
    if (refresh_limit_timer.isValid() && refresh_limit_timer.elapsed() < 20 + refresh_limit_last_elapsed) //~50 refreshes per second max
        return;
    refresh_limit_timer.start();

    Img.detach();
    QRectF cropp(Img.width()*(1 - cw) / 2,
                 Img.height()*(1 - ch) / 2,
                 Img.width()*cw,
                 Img.height()*ch);

    if (!Img.rect().isValid() || !cropp.isValid()) {
        qDebug() << "invalid frame";
        return;
    }

    if (!rotxy->isChecked())
        ritem->image = Img.copy(cropp.toRect()).mirrored(flipx->isChecked(), flipy->isChecked());
    else {
        QTransform m;
        m.rotate(90);
        m.scale(flipx->isChecked() ? -1 : 1, flipy->isChecked() ? -1 : 1);
        ritem->image = Img.copy(cropp.toRect()).transformed(m);
    }

    QPlot->setMinimumSize(ritem->image.size() / 32);
    QPlot->replot();

    frames++;

    if (inst_idx == 0)
    if (I_MODULE *net = ModuleManager::instance->find1("net")) {
        MobileServer *m = (MobileServer *)net->basecast();
        if (m->any_need_data("video_need")) {
            QByteArray vdata;
            QBuffer buf(&vdata);
            buf.open(QBuffer::WriteOnly);
            ritem->image.save(&buf, "bmp");
            buf.close();

            m->setPropIdVar(DataID::VideoData, vdata, 1);
        }
    }

    refresh_limit_last_elapsed = refresh_limit_timer.elapsed() / 0.2; //time budget, %
}



void VideoPlayer::calcOscContrast(QByteArray data)
{
    const EG_OscData *sdata = (const EG_OscData *)data.constData();

    if (sdata->Channel >= 4)
        return;

    osc_contrast = 0;
    for (int i = 0; i < sdata->Size; i++)
        osc_contrast += sdata->Data[i];
    osc_contrast /= sdata->Size;

    new_osc_contrast = true;
}

void VideoPlayer::on_BAutoFocusOsc_clicked()
{
    double speed = 1.0; //mkm/s
    double z1 = 0;
    double z2 = zPos->edit->maximum();//mkm

    double contrast = 0;
    double old = 0;
    double max = 0;
    double zmax = 0;
    int counter = 0;

    I_MODULE *oscm = ModuleManager::instance->find1("oscilloscope");
    if (!oscm)
        return;
    Oscilloscope *osc = (Oscilloscope *)oscm->basecast();

    setEnabled(false);
    osc->setEnabled(false);

    int old_ch[4];
    for (int i = 0; i < 4; i++) {
        old_ch[i] = osc->CBChannel[i]->currentIndex();
        osc->CBChannel[i]->setCurrentIndex(srcNone);
    }

    int srcch = srcADC4;
    {
        QSettings settings("Param/ccd.conf", QSettings::IniFormat);
        if (settings.value("use_pmt", false).toBool())
            srcch = srcADCHR1;
    }
    osc->CBChannel[0]->setCurrentIndex(osc->CBChannel[0]->findData(srcch));

    osc->start->change_state_man_with_sig(true);

    DoubleChanger *extra_dac = osc->DacsWidget->Extra_Dac_1;
    extra_dac->setValue(0);
    extra_dac->setNewValue();

    isInterrupted = false;

    QPointer<StatusBar> bar = new StatusBar(tr("PMT Autofocus"), 4);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    zPos->setValue(0);
    zPos->setNewValue();
    SleepExport::msleep_gui(50);

    int time = 50;
    QElapsedTimer timer;

    const int daclow = 35000;
    const int dachigh = 55000;
    double dacspeed = 0.1;
    int dacsign = 1;

    counter = 0;
    do {
        extra_dac->setValue(extra_dac->value() + dacsign*dacspeed);
        extra_dac->setNewValue();
        SleepExport::msleep_gui(50);

        new_osc_contrast = false;
        connect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));

        time = 500;
        timer.start();
        while (!new_osc_contrast && !isInterrupted && !timer.hasExpired(time)) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
        if (timer.hasExpired(time)) {
            qDebug() << "AFO: no osc signal";
            new Notification(tr("PMT Autofocus") + ": " + tr("No data from controller"), Notification::Warn, 10);
            doBeep();

            destructStatus(bar);
            isInterrupted = true;
        }

        disconnect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));
        if (isInterrupted)
            continue;

        old = contrast;
        contrast = osc_contrast;

        if ((old > daclow && contrast < daclow) ||
                (old < dachigh && contrast > dachigh)) {
            dacsign *= -1;
            dacspeed /= 2;
            qDebug() << "dac overflow" << old << contrast << dacspeed;
        }
        counter++;
    } while ((contrast < daclow || contrast > dachigh) && counter < 50 && extra_dac->value() < 0.9 && !isInterrupted);

    if (counter >= 50 || extra_dac->value() >= 0.9 || contrast != contrast || isInterrupted) {
        destructStatus(bar);

        setEnabled(true);
        osc->setEnabled(true);

        for (int i = 0; i < 4; i++)
            osc->CBChannel[i]->setCurrentIndex(old_ch[i]);

        osc->start->change_state_man_with_sig(false);

        extra_dac->setValue(0);
        extra_dac->setNewValue();

        if (!isInterrupted) {
            qDebug() << "AFO: no signal";
            new Notification(tr("PMT Autofocus") + ": " + tr("No PMT signal"), Notification::Warn, 10);
            doBeep();
        }

        return;
    }

    qDebug() << "AFO dac" << contrast << extra_dac->value();

    counter = 0;
    do {
        bar->setValue(1+counter);
        qDebug() << "AFO pass" << counter << z1 << z2;

        zPos->setValue(z1);
        zPos->setNewValue();
        SleepExport::msleep_gui(50);

        do {
            if (contrast > max) {
                max = contrast;
                zmax = zPos->value();
            }

            new_osc_contrast = false;
            connect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));

            time = 500;
            timer.start();
            while (!new_osc_contrast && !isInterrupted && !timer.hasExpired(time)) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }
            if (timer.hasExpired(time)) {
                qDebug() << "AFO: no osc signal2";
                new Notification(tr("PMT Autofocus") + ": " + tr("No data from controller"), Notification::Warn, 10);
                doBeep();

                destructStatus(bar);
                isInterrupted = true;
            }

            disconnect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));
            if (isInterrupted)
                continue;

            old = contrast;
            contrast = osc_contrast;

            if (contrast > dachigh) {
                dacsign = -1;
                dacspeed = 0.05;
                double olddac = extra_dac->value();

                do {
                    extra_dac->setValue(extra_dac->value() + dacsign*dacspeed);
                    extra_dac->setNewValue();
                    SleepExport::msleep_gui(50);

                    new_osc_contrast = false;
                    connect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));

                    time = 500;
                    timer.start();
                    while (!new_osc_contrast && !isInterrupted && !timer.hasExpired(time)) {
                        qApp->processEvents();
                        SleepExport::msleep2(10);
                    }
                    if (timer.hasExpired(time)) {
                        qDebug() << "AFO: no osc signal3";
                        new Notification(tr("PMT Autofocus") + ": " + tr("No data from controller"), Notification::Warn, 10);
                        doBeep();

                        destructStatus(bar);
                        isInterrupted = true;
                    }

                    disconnect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));
                    if (isInterrupted)
                        continue;

                    old = contrast;
                    contrast = osc_contrast;

                    if ((old > daclow && contrast < daclow) ||
                            (old < dachigh && contrast > dachigh)) {
                        dacsign *= -1;
                        dacspeed /= 2;
                        qDebug() << "dac overflow" << old << contrast << dacspeed;
                    }
                } while ((contrast < daclow || contrast > dachigh) && !isInterrupted);
                max *= extra_dac->value() / olddac;

                qDebug() << "AFO dac" << contrast << extra_dac->value() << max;
            }

            qDebug() << "AFO_0" << contrast << zPos->value();

            zPos->setValue(zPos->value() + speed);
            zPos->setNewValue();
            SleepExport::msleep_gui(50);
        } while ((zPos->value() + speed) <= z2 && !isInterrupted);

        double dz = (z2 - z1)/8;
        double zrange = zPos->edit->maximum();

        z1 = qBound(0., zmax - dz/2, zrange);
        z2 = qBound(0., zmax + dz/2, zrange);

        speed /= 4;
        counter++;
    } while (speed >= 0.2 && counter < 3 && !isInterrupted);

    qDebug() << "AFO" << counter << zmax;

    zPos->setValue(zmax);
    zPos->setNewValue();

    destructStatus(bar);

    setEnabled(true);
    osc->setEnabled(true);

    for (int i = 0; i < 4; i++)
        osc->CBChannel[i]->setCurrentIndex(old_ch[i]);

    osc->start->change_state_man_with_sig(false);

    extra_dac->setValue(0);
    extra_dac->setNewValue();
}

#define DIFF(a, b) (abs(long(a & 0xFF) - long(b & 0xFF)) + abs(long((a >> 8) & 0xFF) - long((b >> 8) & 0xFF)) + abs(long((a >> 16) & 0xFF) - long((b >> 16) & 0xFF)))
void VideoPlayer::on_BAutoFocus_clicked()
{
    double speed = 2.0;//mkm/s
    double z1 = 0;
    double z2 = zPos->edit->maximum();//mkm

    double max = 0;
    double zmax = 0;
    int counter = 0;

    setEnabled(false);
    bool old_start = m_BStart->on_off;
    if (!m_BStart->on_off)
        m_BStart->change_state_man_with_sig(true);

    isInterrupted = false;

    QPointer<StatusBar> bar = new StatusBar(tr("Optical Autofocus"), 3);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    int time;
    QElapsedTimer timer;

    do {
        bar->setValue(counter);
        qDebug() << "AF pass" << counter << z1 << z2;

        zPos->setValue(z1);
        zPos->setNewValue();
        SleepExport::msleep_gui(200);

        double contrast = 0;

        do {
            if (contrast > max) {
                max = contrast;
                zmax = zPos->value();
            }

            int frm = frames;
            time = 2000;
            timer.start();
            while (frm == frames && !isInterrupted && !timer.hasExpired(time)) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }
            if (timer.hasExpired(time)) {
                new Notification(tr("Autofocus") + ": " + tr("No frames from camera"), Notification::Warn, 10);
                doBeep();

                destructStatus(bar);
                isInterrupted = true;
            }

            if (isInterrupted)
                continue;

            //async  move & calc
            zPos->setValue(zPos->value() + speed);
            zPos->setNewValue();

            contrast = 0;
            QImage img = ritem->image;
            img.detach();
            for (int i = 1; i < img.width(); i++)
                for (int j = 1; j < img.height(); j++) {
                    int diff1 = DIFF(img.pixel(i, j), img.pixel(i-1, j));
                    if (abs(diff1) > 32)
                        contrast += abs(diff1)/128.0;

                    int diff2 = DIFF(img.pixel(i, j), img.pixel(i, j-1));
                    if (abs(diff2) > 32)
                        contrast += abs(diff2)/128.0;
                }

            qDebug() << "AF_0" << contrast << zPos->value() - speed;
        } while (zPos->value() < z2 && !isInterrupted);

        double dz = (z2 - z1)/16;
        double zrange = zPos->edit->maximum();

        z1 = qBound(0., zmax - dz/2, zrange);
        z2 = qBound(0., zmax + dz/2, zrange);

        speed /= 8;
        counter++;
    } while (speed >= 0.05 && counter < 3 && !isInterrupted);

    qDebug() << "AF" << counter << zmax;

    zPos->setValue(zmax);
    zPos->setNewValue();

    destructStatus(bar);

    if (!old_start)
        m_BStart->change_state_man_with_sig(false);

    setEnabled(true);
}

void VideoPlayer::cancelOperation()
{
    isInterrupted = true;
}
