#include "ccdplot.h"
#include "gui/status.hpp"
#include "flowlayout.hpp"
#include <qwt_thermo.h>

#include "modules.hpp"
#include "oscilloscope.h"
#include "gui/ttabwidget.hpp"
#include "mainwindow.h"

#include "ccdmatrix.h"
#include "i_spectr.hpp"
#include "hardware/coboltcontroller.hpp"
#include "hardware/nspectr_proxy.h"

#include <QScrollArea>
#include <QComboBox>
#include <QCheckBox>
#include <QGroupBox>
#include <QLabel>
#include <QTabWidget>
#include <QTextEdit>
#include <QRadioButton>
#include <QTimer>
#include <QSettings>
#include <QCloseEvent>
#include <QProgressBar>
#include <QFormLayout>
#include <QWidgetAction>


bool CCDPlot::eventFilter(QObject *Obj, QEvent *Event)
{
    if (Event->type() == plot::TScalesChangedEvent::id && setReadMode->currentIndex() == 2) {
        sl_set_new_3D_Limits(((plot::TScalesChangedEvent *) Event)->XLimits, ((plot::TScalesChangedEvent *) Event)->YLimits);
        ccdMatrix->sl_set_new_3D_Limits(((plot::TScalesChangedEvent *) Event)->XLimits, ((plot::TScalesChangedEvent *) Event)->YLimits);
    }
    return QWidget::eventFilter(Obj, Event);
}

CCDPlot::CCDPlot(QWidget *parent) : QWidget(parent)
{
    Nspec_Initalized=false;
    nspectr_mutex = 0;
    ccdWidget = 0;
    ScanCCDMode = false;
    ccdMatrix = (CCDMatrix *)ModuleManager::instance->find1("ccdmatrix")->basecast();
    ccdMatrix->curveCalib = &curveCalib;
    PMTPlot = 0;

    Nx=1024;
    Ny=256;

    deviceStatus = new StatusWidget1(tr("CCD"));
    deviceStatus->label->setToolTip(tr("Andor"));
#ifdef USE_CCD_EMULATOR
    {
        QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
        if (settings.value("ccd", false).toBool())
            deviceStatus->setEmulator(true);
    }
#endif

    deviceStatus2 = new StatusWidget1(tr("Optics"));

    acqUpd = new QTimer(this);
    temperatureUpd = new QTimer(this);
    refresh_limit_timer.invalidate();

    connect(acqUpd, SIGNAL(timeout()), SLOT(updateSpecBar()));
    connect(temperatureUpd, SIGNAL(timeout()), SLOT(slot_updateTemperature()));

    fillLaserConf();

    setWindowTitle(tr("CCD - NSpec"));

    createPlot();
    plot::ResizeOwner(curveData, 1);
    MyPl->TwoDPlot->Pure2dPlot->SourceData = plot::getUnLock(curveData);

    plot::ResizeOwner(imgData, 1);
    imgData.at(0)->setSize(Nx, Ny);

    createButtons();
    createCCDAtt();
    createLayout();
    load_data();

    monoch_status_update(false);

    QMetaObject::connectSlotsByName(this);

    installEventFilter(this);

    connect(base_line_manual, SIGNAL(clicked()),MyPl->TwoDPlot, SLOT(StartMCCurveConstructing()));
    connect(MyPl->TwoDPlot, SIGNAL(ManualConstructingDone()), this, SLOT(SetManualBaseLine()));
    connect(MyPl, SIGNAL(AddDataToContainer(plot::T3dDataOwner,plot::TMetaStorage)), SLOT(AddDataToContainer_slot(plot::T3dDataOwner,plot::TMetaStorage)));
    connect(MyPl, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)), SLOT(AddDataToContainer_slot(plot::TCurvesDataOwner,plot::TMetaStorage)));

    //ccdmatrix connections  //TODO module connected()?
    ccdMatrix->setOutput(&imgData.at(0)->Array);

    connect(ccdMatrix, SIGNAL(sig_Update()), SLOT(sl_Plot_Refresh()), Qt::QueuedConnection);
    connect(ccdMatrix, SIGNAL(new2dData(QVector<double>,QVector<double>)), SLOT(new2dData(QVector<double>,QVector<double>)), Qt::QueuedConnection);
    connect(ccdMatrix, SIGNAL(sigUpdateNxNy()), SLOT(UpdateNxNy()));
    connect(ccdMatrix, SIGNAL(ccd_finished()), SLOT(sl_aq_but_off()));
    connect(ccdMatrix, SIGNAL(sig_cur_kin_ser_num(int)), SLOT(sl_set_current_kin_num(int)));
    connect(ccdMatrix, SIGNAL(sig_cur_ac_num(int)), SLOT(sl_set_current_acc_num(int)));

    connect(lineSetTemp, SIGNAL(newValue(int)), ccdMatrix, SLOT(s_setTemp(int)));
    connect(lineExpTime, SIGNAL(newValue(double)), ccdMatrix, SLOT(sl_set_Exp_time(double)));
    connect(setAcqMode, SIGNAL(currentIndexChanged(int)), ccdMatrix, SLOT(sl_set_acqMode(int)));
    connect(setReadMode, SIGNAL(currentIndexChanged(int)), ccdMatrix, SLOT(sl_set_readMode(int)));

    connect(lineNumAcc, SIGNAL(newValue(int)), ccdMatrix, SLOT(sl_set_Acc_num(int)));
    connect(lineAccCycleTime, SIGNAL(newValue(double)), ccdMatrix, SLOT(sl_set_Acc_time(double)));
    connect(Num_Kin_ser, SIGNAL(newValue(int)), ccdMatrix, SLOT(sl_set_Kin_ser_num(int)));
    connect(Time_Kin_ser, SIGNAL(newValue(double)), ccdMatrix, SLOT(sl_set_Kin_ser_time(double)));

    connect(setTriggerMode, SIGNAL(currentIndexChanged(int)), ccdMatrix, SLOT(sl_set_Triger(int)));
    connect(lineCenterRowTrack, SIGNAL(newValue(int)), ccdMatrix, SLOT(sl_set_Center_row(int)));
    connect(lineHeighTrack, SIGNAL(newValue(int)), ccdMatrix, SLOT(sl_set_Height(int)));

    connect(One_Cont, SIGNAL(state_changed(bool)), ccdMatrix, SLOT(s_isContScan(bool)));
    connect(Cut_Pikes, SIGNAL(state_changed(bool)), ccdMatrix, SLOT(sl_cut_pikes(bool)));
    connect(base_line_save, SIGNAL(clicked()), ccdMatrix, SLOT(sl_save_base_line()));
    connect(Base_line, SIGNAL(state_changed(bool)), ccdMatrix, SLOT(sl_set_baseline_on_off(bool)));

    {//ccd settings
        QSettings settings("Param/ccd.conf", QSettings::IniFormat);

        if (settings.value("use_ormins", false).toBool())
            deviceStatus->label->setToolTip(tr("Ormins"));
    }

    devStatusSync = new QTimer(this);
    devStatusSync->setSingleShot(true);

    prevDevStatus = -1;
    syncDevStatus = EDno;
    setDeviceStatus();
    setDeviceStatus2(EDno);

    connect(ccdMatrix, SIGNAL(updateCCDStatus(int)), this, SLOT(setDeviceStatus_handler(int)), Qt::QueuedConnection);
    connect(ccdMatrix, SIGNAL(monoch_status(bool)), SLOT(monoch_status_update(bool)), Qt::QueuedConnection);
    connect(devStatusSync, SIGNAL(timeout()), this, SLOT(setDeviceStatus()));


    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    bool b1on = settings.value("ccd/intensity_calib_on_off", false).toBool();
    bool b2on = settings.value("ccd/intensity_calib_pixel_on_off", false).toBool();
    QString name1 = settings.value("ccd/intensity_calib_name", QString()).toString();
    QString name2 = settings.value("ccd/intensity_calib_pixel_name", QString()).toString();
    if (!name1.isEmpty()) intensity_calib_open_wl(name1);
    if (!name2.isEmpty()) intensity_calib_open_px(name2);
    calibEn->setChecked(b1on);
    pixelCalibEn->setChecked(b2on);
}

void CCDPlot::AddDataToContainer_slot(plot::TCurvesDataOwner src, plot::TMetaStorage attributes)
{
    attributes.insert("Name", MyPl->TwoDPlot->Pure2dPlot->Title);
    attributes.insert("Source", "CCD");

    I_MODULE *oscm = ModuleManager::instance->find1("oscilloscope");
    Oscilloscope *osc = (Oscilloscope *)oscm->basecast();

    if (osc) {
        attributes.insert("Stage X, um", osc->DacsWidget->X2->value());
        attributes.insert("Stage Y, um", osc->DacsWidget->Y2->value());
    }
    attributes.insert("CCD NX", Nx);
    attributes.insert("CCD NY", Ny);
    attributes.insert("CCD Central Wavelength, nm", ccdMatrix->Cent_WL);
    attributes.insert("CCD Dispersion", ccdMatrix->Disp);
    attributes.insert("CCD Central Pixel", lineCentralPx->value());
    attributes.insert("Exposure time", lineExpTime->value());
    if (ccdMatrix->m_spectrInterface)
    attributes.insert("Grating", CBgrating_select->currentText());
    if (Nspec_Initalized && nspectr_mutex)
    attributes.insert("ND filter", NDEdit->value());

    foreach (CoboltGui *it, cobolt_gui_list)
        attributes.insert("Laser Power, mW - " + it->windowTitle(), it->cpower->text());


    attributes.unite(curveCalib);

    emit AddDataToContainer(src, attributes);
}

void CCDPlot::AddDataToContainer_slot(plot::T3dDataOwner src, plot::TMetaStorage attributes)
{
    attributes.insert("Name", MyPl->ThreeDPlot->ThreeDPlots.at(0)->Title);
    attributes.insert("Source", "CCD");

    emit AddDataToContainer(src, attributes);
}

CCDPlot::~CCDPlot()
{
    foreach (CoboltGui *it, cobolt_gui_list)
        delete it;
    cobolt_gui_list.clear();
}

void CCDPlot::closeEvent(QCloseEvent *Event)
{
    PBToNewWindow->click();
    Event->ignore();
}

void CCDPlot::slot_updateTemperature()
{
    int Temper = ccdMatrix->getCameraTemperature();
    if (Temper > 15) Termometer->setFillBrush(Qt::red);
    else if (Temper >= 0) {
        Termometer->setFillBrush(Qt::green);
        temp_pic->setPixmap(QPixmap(":/CCD/plus.png"));
    } else {
        Termometer->setFillBrush(Qt::blue);
        temp_pic->setPixmap(QPixmap(":/CCD/minus.png"));
    }
    Termometer->setValue(Temper);
    lb_temp->setText(QString::number(Temper));
}

void CCDPlot::UpdateNxNy()
{
    Nx = ccdMatrix->Nx;
    Ny = ccdMatrix->Ny;

    imgData.at(0)->setSize(Nx, Ny);
}
void CCDPlot::sl_set_new_3D_Limits(const Range &/*XLimits*/, const Range &YLimits)
{
    double up_limit;
    if (YLimits[1] > (Ny-1))
        up_limit = Ny-1;
    else
        up_limit = YLimits[1];

    int cent_row = (int)(0.5*(up_limit + YLimits[0]));
    int track_height = (int)(up_limit - YLimits[0] - 1);
    if (track_height < 3) track_height = 3;

    lineCenterRowTrack->setValue(cent_row);
    lineHeighTrack->setValue(track_height);
    ccdMatrix->m_centralRow = cent_row;
    ccdMatrix->m_trackHeight = track_height;
}
void CCDPlot::new2dData(QVector<double> ydata, QVector<double> xdata)
{
    if (spec_bar != 0) {
        spec_bar->bar->setUpdatesEnabled(false);
        spec_bar->bar->setMaximum(ccdMatrix->m_exposureTime*1000);
        spec_bar->bar->reset();
        spec_bar->bar->setUpdatesEnabled(true);
    }

    curveData[0]->XSeq = xdata;
    curveData[0]->YSeq = ydata;

    if (!refresh_limit_timer.isValid() || refresh_limit_timer.elapsed() > 100) { //10 refreshes per second max
        refresh_limit_timer.start();

        MyPl->Refresh();

        if (PMTPlot && PMTPlot->isVisibleTo(this)) {
            PMTPlot->hide();
            MyPl->show();
        }
    }
}

void CCDPlot::sl_Plot_Refresh()
{
    if (spec_bar != 0) {
        spec_bar->bar->setUpdatesEnabled(false);
        spec_bar->bar->setMaximum(ccdMatrix->m_exposureTime*1000);
        spec_bar->bar->reset();
        spec_bar->bar->setUpdatesEnabled(true);
    }

    if (!refresh_limit_timer.isValid() || refresh_limit_timer.elapsed() > 100) { //10 refreshes per second max
        refresh_limit_timer.start();

        imgData.at(0)->XLimits[0] = ccdMatrix->imageX_min;
        imgData.at(0)->XLimits[1] = ccdMatrix->imageX_max;
        imgData.at(0)->YLimits[0] = ccdMatrix->imageY_min;
        imgData.at(0)->YLimits[1] = ccdMatrix->imageY_max;
        imgData.at(0)->NXElems = ccdMatrix->imageX_max - ccdMatrix->imageX_min + 1;
        imgData.at(0)->NYElems = ccdMatrix->imageY_max - ccdMatrix->imageY_min + 1;
        imgData.at(0)->UpdateDxDy();
        imgData.at(0)->resetMathLims();

        MyPl->ThreeDPlot->SetPlotsXScales(ccdMatrix->imageX_min, ccdMatrix->imageX_max);
        MyPl->ThreeDPlot->SetPlotsYScales(ccdMatrix->imageY_min, ccdMatrix->imageY_max);
        MyPl->ThreeDPlot->ThreeDPlots[0]->SourceData = plot::getUnLock(imgData);

        MyPl->Refresh();

        if (PMTPlot && PMTPlot->isVisibleTo(this)) {
            PMTPlot->hide();
            MyPl->show();
        }
    }
}
void CCDPlot::sl_set_current_acc_num(int num)
{
    lineNumAcc->blockSignals(true);
    lineNumAcc->setValue(num);
    lineNumAcc->blockSignals(false);
}
void CCDPlot::sl_set_current_kin_num(int num)
{
    Num_Kin_ser->blockSignals(true);
    Num_Kin_ser->setValue(num);
    Num_Kin_ser->blockSignals(false);

    if (setReadMode->currentIndex() != 2) {
        bool last = num >= ccdMatrix->Kin_ser_number;
        if (!last) {
            curveData.prepend(QSharedPointer<plot::TCurveData>(new plot::TCurveData));
            curveData[0]->Pen.setColor(QColor((int)((double)qrand() / (double)RAND_MAX * 240. ), (int)((double)qrand() / (double)RAND_MAX * 240. ), (int)((double)qrand() / (double)RAND_MAX * 240. )));
            MyPl->TwoDPlot->Pure2dPlot->SourceData = plot::getUnLock(curveData);
        }

        if (kinSerDACscan->currentIndex() > 0) {
            I_MODULE *oscm = ModuleManager::instance->find1("oscilloscope");
            if (!oscm)
                return;
            Oscilloscope *osc = (Oscilloscope *)oscm->basecast();
            DoubleChanger *dc = osc->DacsWidget->get_dacs_channel(kinSerDACscan);
            if (num == 1)
                kinSerDACinitialValue = dc->value();
            if (last)
                dc->setValue(kinSerDACinitialValue);
            else
                dc->setValue(dc->value() + kinSerDACstep->value());
            dc->setNewValue();
        }
    }
}
void CCDPlot::enable_options_rm(int mode)
{
    box_singleTrack->setVisible(mode == 1);
    BPanoram->setEnabled(mode != 2);
    Base_line->setEnabled(mode != 2);
    base_line_save->setEnabled(mode != 2);
    base_line_manual->setEnabled(mode != 2);

    if (mode == 2) {
        MyPl->SwitchPlots(plot::TPlot::_3d);
        MyPl->ThreeDPlot->SetXYDescAndSIUnit(QObject::tr("pixels"), QString());
        MyPl->ThreeDPlot->SetZDescAndSIUnit(QObject::tr("Counts"), QString());

        static bool first = true;
        if (first) {
            first = false;
            MyPl->ThreeDPlot->SetDefaultXLimits(ccdMatrix->imageX_min, ccdMatrix->imageX_max);
            MyPl->ThreeDPlot->SetDefaultYLimits(ccdMatrix->imageY_min, ccdMatrix->imageY_max);
        }
        MyPl->ThreeDPlot->SetPlotsXScales(ccdMatrix->imageX_min, ccdMatrix->imageX_max);
        MyPl->ThreeDPlot->SetPlotsYScales(ccdMatrix->imageY_min, ccdMatrix->imageY_max);
        MyPl->ThreeDPlot->SetTitle(tr("CCD Image"));

        MyPl->ThreeDPlot->ThreeDPlots[0]->ResizeQPlot();
        sl_Plot_Refresh();

        if (setAcqMode->currentIndex() == 2)
            setAcqMode->setCurrentIndex(0);
        setAcqMode->removeItem(2);

        if (BPanoram->on_off)
            BPanoram->change_state_man_with_sig(false);
        if (Base_line->on_off)
            Base_line->change_state_man_with_sig(false);
    } else {
        MyPl->SwitchPlots(plot::TPlot::_2d);
        MyPl->TwoDPlot->Pure2dPlot->SetDefaultYLimits(0, 65535);
        sl_set_X_scale(scaleBox->currentIndex());

        if (mode == 1)
            MyPl->TwoDPlot->Pure2dPlot->SetTitle(tr("CCD Single Track Curve"));
        else
            MyPl->TwoDPlot->Pure2dPlot->SetTitle(tr("CCD Full V Binning Curve"));

        if (setAcqMode->count() != 3)
            setAcqMode->addItem(tr("Kinetic Series"));
    }

    if (PMTPlot && PMTPlot->isVisibleTo(this)) {
        PMTPlot->hide();
        MyPl->show();
    }
}
void CCDPlot::enable_options_acm(int mode)
{
    box_accumulation->setVisible(mode != 0);
    Kin_series->setVisible(mode == 2);

    I_MODULE *oscm = ModuleManager::instance->find1("oscilloscope");
    Oscilloscope *osc = (Oscilloscope *)oscm->basecast();
    if (mode == 2 && osc) {
        kinSerDACscan->clear();
        kinSerDACscan->addItem(tr("None"), -1);
        osc->DacsWidget->fill_dacs_channel_CB(kinSerDACscan);
    }
}

void CCDPlot::s_triggerMode(int mode)
{
    lineAccCycleTime->setEnabled(mode == 0);
    Time_Kin_ser->setEnabled(mode == 0);
}


void CCDPlot::sl_calib_change(bool value)
{
    bool auto_spectr = value && ccdMatrix->m_spectrInterface;
    lineDispersion->setVisible(!auto_spectr);
    lb_dispersion->setVisible(!auto_spectr);
    lineCentrWL->setVisible(!auto_spectr);
    lb_centralWL->setVisible(!auto_spectr);
    ccdMatrix->sl_set_Auto_calib(auto_spectr);

    if (auto_spectr) {
        lb_auto->setText(tr("Status: auto"));
    } else {
        lb_auto->setText(tr("Status: manual"));
        s_apply_calib_Settings();
        AutoCalibration->change_state_man(false);
    }
}
void CCDPlot::sl_set_X_scale(int sc)
{
    if (setReadMode->currentIndex() == 2)
        return;

    switch (sc) {
    case 0:
        MyPl->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("pixels"));
        MyPl->TwoDPlot->Pure2dPlot->SetAllXScales(0,ccdMatrix->Nx_real);
        lineLaserWl->hide();
        lb_laser->hide();
        break;
    case 1:
        MyPl->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("WL"), QObject::tr("m"));
        MyPl->TwoDPlot->Pure2dPlot->SetAllXScales(0,2000);
        lineLaserWl->hide();
        lb_laser->hide();
        break;
    case 2:
        MyPl->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("Raman Shift, 1/cm"));
        MyPl->TwoDPlot->Pure2dPlot->SetAllXScales(-10000,10000);
        lineLaserWl->show();
        lb_laser->show();
        break;
    }

    s_apply_calib_Settings();
}

void CCDPlot::s_apply_calib_Settings()
{
    save_data();
    ccdMatrix->s_setXAxisScale(scaleBox->currentIndex(), lineLaserWl->value(), lineCentralPx->value(), lineCentrWL->value(), lineDispersion->value());

    if (ccdMatrix->Scale != scaleBox->currentIndex())
        for (int i = 0; i < curveData.size(); i++)
            set_spectra_calib(*curveData[i].data(), curveCalib, scaleBox->currentIndex());
    MyPl->Refresh();
}

void CCDPlot::load_data()
{
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);

    int scale_ind = settings.value("scale_ind", 0).toInt();
    lineLaserWl->setValue(settings.value("laser_wl", 532).toDouble());
    lineCentrWL->setValue(settings.value("central_wl", 532).toDouble());
    lineDispersion->setValue(settings.value("dispersion", 10).toDouble());
    lineCentralPx->setValue(settings.value("central_px", 512).toDouble());

    scaleBox->setCurrentIndex(scale_ind);
    sl_set_X_scale(scale_ind);
}

void CCDPlot::save_data()
{
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);

    settings.setValue("scale_ind", scaleBox->currentIndex());
    settings.setValue("laser_wl", lineLaserWl->text());
    settings.setValue("central_wl", lineCentrWL->text());
    settings.setValue("central_px", lineCentralPx->text());
    settings.setValue("dispersion", lineDispersion->text());
}


void CCDPlot::startAcquisition(bool value)
{
    foreach (QWidget *wid, ccdControlWidgets)
        wid->setEnabled(!value && !ScanCCDMode);

    if (value) {
        MyPl->HideSelectionIfAny();
        if (MyPl->TwoDPlot->PlotControl->BMarkers->isChecked())
            MyPl->TwoDPlot->PlotControl->BMarkers->click();

        QColor color = Qt::black;
        if (curveData.size() == 1)
            color = curveData[0]->Pen.color();

        plot::ResizeOwner(curveData, 1);
        curveData[0]->Pen.setColor(color);
        MyPl->TwoDPlot->Pure2dPlot->SourceData = plot::getUnLock(curveData);

        ccdMatrix->startAcquisition();

        //if (ccdMatrix->m_exposureTime*1000 >= 100) {
        if (!spec_bar) {
            spec_bar = new StatusBar(tr("CCD exposure"), ccdMatrix->m_exposureTime*1000);
            spec_bar->bar->setFormat(tr("%p% of %m ms"));
            connect(spec_bar, SIGNAL(canceled()), m_BStart, SLOT(click()));
            spec_bar->add();
        } else {
            spec_bar->bar->setUpdatesEnabled(false);
            spec_bar->bar->setMaximum(ccdMatrix->m_exposureTime*1000);
            spec_bar->bar->reset();
            spec_bar->bar->setUpdatesEnabled(true);
        }
        spec_bar->cancel->setEnabled(m_BStart->isEnabled());

        acqUpd->start(100);
    } else {
        ccdMatrix->s_abortAcq();

        acqUpd->stop();
        if (!ScanCCDMode)
            destructStatus(spec_bar);

        static bool first_spectra = true; //guard auto-send filler data started in main.c
        if (!first_spectra && AutoSend && MyPl->CurrentPlotType == plot::TPlot::_2d && !ScanCCDMode) {
            MyPl->TwoDPlot->on_SyncTimer_timeout();
            MyPl->TwoDPlot->PlotControl->BSave->animateClick();
        }
        first_spectra = false;


        lineNumAcc->blockSignals(true);
        lineNumAcc->setValue(ccdMatrix->numOfAccumulations);
        lineNumAcc->blockSignals(false);
        Num_Kin_ser->blockSignals(true);
        Num_Kin_ser->setValue(ccdMatrix->Kin_ser_number);
        Num_Kin_ser->blockSignals(false);

        if (ccdMatrix->m_spectrInterface && ccdMatrix->panoram) //update central wl
            lineCentralWlMonoch->setValue(ccdMatrix->m_spectrInterface->getWl());
    }
}

void CCDPlot::updateSpecBar()
{
    if (spec_bar != 0 && ccdMatrix->m_ccdStart) {
        spec_bar->setValue(spec_bar->bar->value() + 100);
    }
}

void CCDPlot::sl_aq_but_off()
{
    if (m_BStart->on_off)
        m_BStart->change_state_man_with_sig(false);
}


void CCDPlot::onPanoram(bool val)
{
#ifdef USE_CCD_EMULATOR
    QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
    if (!settings.value("ccd", false).toBool())
#endif
    if (val && !ccdMatrix->m_spectrInterface) {
        new Notification(tr("Panoram") + ": " + tr("Initialize optics before use."), Notification::Warn, 10);
        doBeep();

        BPanoram->change_state_man(false);
        return;
    }
    ccdMatrix->panoram = val;

    //manual calib not applicable
    if (val && BNLCFile->isChecked())
        BNLCAuto->click();
    BNLCFile->setEnabled(!val);

    panoramControl->setVisible(val);
    if (One_Cont->on_off)
        One_Cont->change_state_man_with_sig(false);
    One_Cont->setEnabled(!val);
}
void CCDPlot::onPanoramMin(double val)
{
    if (val >= ccdMatrix->panoram_max) {
        val = ccdMatrix->panoram_max - 10;
        PanoramMin->setValue(val);
    }
    ccdMatrix->panoram_min = val;
}
void CCDPlot::onPanoramMax(double val)
{
    if (val <= ccdMatrix->panoram_min) {
        val = ccdMatrix->panoram_min + 10;
        PanoramMax->setValue(val);
    }
    ccdMatrix->panoram_max = val;
}


void CCDPlot::createPlot()
{
    MyPl = new plot::TPlot(this, this);

    MyPl->ThreeDPlot->TwoDPlot->Pure2dPlot->SetAutoLegendEnabled(false);

    MyPl->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("pixels"));
    MyPl->TwoDPlot->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(QObject::tr("Counts"));
    MyPl->TwoDPlot->Pure2dPlot->SetTitle(tr("CCD Full V Binning Curve"));
    MyPl->TwoDPlot->Pure2dPlot->SetAllXScales(0, Nx - 1);
    MyPl->TwoDPlot->Pure2dPlot->SetAllYScales(0, 65535);
}

void CCDPlot::createButtons()
{
    PBToNewWindow =  new TwoStateButton(":/Detach.png",":/Attach.png");
    PBToNewWindow->setToolTip(QObject::tr("Individual window on/off"));

    sendTemp = new QLabel(tr("Set Temperature, ")+QChar(176)+tr("C"));

    m_BStart = new TwoStateButton(":start.png",":stop.png");
    m_BStart->setToolTip(tr("Start/stop Acquisition"));
    connect(m_BStart, SIGNAL(state_changed(bool)), this, SLOT(startAcquisition(bool)));

    One_Cont= new TwoStateButton(":/CCD/oneshoot.png",":/CCD/continum.png");
    One_Cont->setToolTip(tr("single/non stop Acquisition"));

    Monochr = new QPushButton;
    Monochr->setIcon(makeHugeIcon(":/CCD/monochromator.png"));
    Monochr->setToolTip(tr("Run Centaur instrument control"));
    connect(Monochr, SIGNAL(clicked()), this, SLOT(sl_show_Monochromator()));

    BPanoram = new TwoStateButton(":/CCD/Pspecoff.png",":/CCD/PspecoN.png");
    BPanoram->setToolTip(tr("Toggle panoram mode"));
    connect(BPanoram, SIGNAL(state_changed(bool)), this, SLOT(onPanoram(bool)));

    Cut_Pikes = new TwoStateButton(":/CCD/cut_pikes_off.png",":/CCD/cut_pikes_on.png");
    Cut_Pikes->setToolTip(tr("Cut spikes on/off"));

    Base_line = new TwoStateButton(":/CCD/base_line_off.png",":/CCD/base_line_on.png");
    Base_line->setToolTip(tr("Base line substract on/off"));

    base_line_save = new QPushButton;
    base_line_save->setIcon(makeHugeIcon(":/CCD/base_line_save.png"));
    base_line_save->setToolTip(tr("Store current line as baseline"));

    base_line_manual = new QPushButton();
    base_line_manual->setIcon(makeHugeIcon(":/CCD/base_manual.png"));
    base_line_manual->setToolTip(tr("Create manual baseline"));

    BPMTScan = new QPushButton;
    BPMTScan->setIcon(makeHugeIcon(":/CCD/Sp_part.png"));
    BPMTScan->setToolTip(tr("Scan PMT spectra"));
    connect(BPMTScan, SIGNAL(clicked(bool)), this, SLOT(PMTScan()));

    if (!show_additional_gui()) {
        Cut_Pikes->hide();
        Base_line->hide();
        base_line_save->hide();
        base_line_manual->hide();
    }

    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    if (!settings.value("use_pmt", false).toBool())
        BPMTScan->hide();
}
void CCDPlot::createCCDAtt()
{
    Termometer = new QwtThermo();
    Termometer->setOrientation(Qt::Horizontal, QwtThermo::BottomScale);
    Termometer->setFillBrush(Qt::blue);
    Termometer->setRange(-100, 30, 0);

    setAcqMode = new QComboBox;
    ccdControlWidgets << setAcqMode;
    setAcqMode->addItem(tr("Single Scan"));
    setAcqMode->addItem(tr("Accumulate"));
    setAcqMode->addItem(tr("Kinetic Series"));
    connect(setAcqMode, SIGNAL(currentIndexChanged(int)), SLOT(enable_options_acm(int)));

    setReadMode = new QComboBox;
    ccdControlWidgets << setReadMode;
    setReadMode->addItem(tr("Full Vertical Binning"));
    setReadMode->addItem(tr("Single-Track"));
    setReadMode->addItem(tr("Image"));
    connect(setReadMode, SIGNAL(currentIndexChanged(int)), SLOT(enable_options_rm(int)));


    lineExpTime = new Edit(0.001, 1000, 3);
    lineExpTime->setValue(0.1);

    lineSetTemp = new Edit(-100, 100, 1);
    lineSetTemp->setValue(20);

    lb_temp = new QLabel("0");
    lb_temp->setSizePolicy(QSizePolicy::Fixed,QSizePolicy::Fixed);
    lb_temp->setFrameStyle(QFrame::Panel);

    temp_pic = new QLabel;
    temp_pic->setPixmap(QPixmap(":/CCD/plus.png"));
    temp_pic->setMaximumSize(40,40);

    lineCenterRowTrack = new Edit(0, 10000);
    lineCenterRowTrack->setValue(128);
    ccdControlWidgets << lineCenterRowTrack;

    lineHeighTrack = new Edit(0, 1000);
    lineHeighTrack->setValue(10);
    ccdControlWidgets << lineHeighTrack;

    lineNumAcc = new Edit(1, 1000);
    lineNumAcc->setImmediateMode();
    ccdControlWidgets << lineNumAcc;

    lineAccCycleTime = new Edit(0, 10000, 2);

    //--------------Kinetic series-----------------------------------------------
    Num_Kin_ser = new Edit(1, 1000);
    Num_Kin_ser->setImmediateMode();
    ccdControlWidgets << Num_Kin_ser;

    Time_Kin_ser = new Edit(0, 1000, 2);

    kinSerDACscan = new QComboBox;
    ccdControlWidgets << kinSerDACscan;
    kinSerDACscan->addItem(tr("None"), -1);
    kinSerDACstep = new Edit(0, 10000, 2);


    lb_auto = new QLabel(tr("Status: manual"));

    setTriggerMode = new QComboBox;
    ccdControlWidgets << setTriggerMode;
    setTriggerMode->addItem(tr("Internal"));
    setTriggerMode->addItem(tr("External"));
    connect(setTriggerMode, SIGNAL(currentIndexChanged(int)), SLOT(s_triggerMode(int)));

    AutoCalibration = new TwoStateButton(":/off.png",":/onn.png");
    connect(AutoCalibration, SIGNAL(state_changed(bool)), this, SLOT(sl_calib_change(bool)));
    AutoCalibration->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);

    scaleBox = new QComboBox;
    scaleBox->addItem(tr("pixels"));
    scaleBox->addItem(tr("Wavelength"));
    scaleBox->addItem(tr("Raman shift"));
    connect(scaleBox, SIGNAL(currentIndexChanged(int)), this, SLOT(sl_set_X_scale(int)));

    CBgrating_select = new QComboBox;
    CBgrating_select->addItem(tr("Default grating"));
    connect(CBgrating_select, SIGNAL(currentIndexChanged(int)), SLOT(monoch_grating()));

    CBPreMonochWl = new QComboBox;
    {
        bool show = true;
        QSettings settings("Param/laser.conf", QSettings::IniFormat);
        settings.beginGroup("list");
        show = settings.allKeys().size() != (settings.allKeys().contains("dummy") ? 1 : 0);
        foreach (QString key, settings.allKeys()) {
            if (key == "dummy")
                continue;

            CBPreMonochWl->addItem(key, settings.value(key, -1).toInt());
        }
        settings.endGroup();

        CBPreMonochWl->setCurrentIndex(-1);
        CBPreMonochWl->setVisible(show);
        if (show)
            connect(CBPreMonochWl, SIGNAL(currentIndexChanged(int)), SLOT(monoch_laser_wl()));
    }

    CBexitport = new QComboBox;
    CBexitport->addItem(tr("Axial"));
    CBexitport->addItem(tr("Lateral"));
    connect(CBexitport, SIGNAL(currentIndexChanged(int)), SLOT(monoch_mirror()));

    lb_centralWL = new QLabel(tr("Central WL (nm)"));
    lineCentrWL = new Edit(0, 10000, 2);
    lineCentrWL->setValue(488);
    connect(lineCentrWL, SIGNAL(newValue()), this, SLOT(s_apply_calib_Settings()));

    lb_dispersion = new QLabel(tr("Dispersion (nm/mm)"));
    lineDispersion = new Edit(0, 10000, 3);
    lineDispersion->setValue(10);
    connect(lineDispersion, SIGNAL(newValue()), this, SLOT(s_apply_calib_Settings()));

    lineCentralPx = new Edit(0, 4000);
    lineCentralPx->setValue(512);
    connect(lineCentralPx, SIGNAL(newValue()), this, SLOT(s_apply_calib_Settings()));

    lb_laser = new QLabel(tr("Laser") + ", " + tr("nm"));
    lb_laser->hide();

    lineLaserWl = new Edit(0, 10000, 2);
    lineLaserWl->setValue(488);
    lineLaserWl->hide();
    connect(lineLaserWl, SIGNAL(newValue()), this, SLOT(s_apply_calib_Settings()));

    lineCentralWlMonoch = new Edit(0, 10000, 2);
    lineCentralWlMonoch->setValue(488);
    connect(lineCentralWlMonoch, SIGNAL(newValue()), this, SLOT(monoch_wl()));

    NDEdit = new Edit(0, 4, 1);
    connect (NDEdit, SIGNAL(newValue(double)), this, SLOT(nd_filter(double)));

    linePreCentralWlMonoch = new Edit(0, 10000, 2);
    linePreCentralWlMonoch->setValue(532);
    connect(linePreCentralWlMonoch, SIGNAL(newValue()), this, SLOT(monoch_pre_wl()));

    lineSlitInMonoch = new Edit(0, 10000, 1);
    lineSlitInMonoch->setValue(100);
    connect(lineSlitInMonoch, SIGNAL(newValue()), this, SLOT(monoch_slit_in()));

    lineSlitOutMonoch = new Edit(0, 10000, 1);
    lineSlitOutMonoch->setValue(100);
    connect(lineSlitOutMonoch, SIGNAL(newValue()), this, SLOT(monoch_slit_out()));

    central_wl_reset = new QCheckBox(tr("Central WL reset"));

    zPos = new DoubleChanger(0, 10, 4);

    calibLbl = new QLabel("-");
    pixelCalibLbl = new QLabel("-");
    calibLbl->setWordWrap(true);
    pixelCalibLbl->setWordWrap(true);

    calibOpn = new QPushButton(tr("Open"));
    pixelCalibOpn = new QPushButton(tr("Open"));
    connect(calibOpn, SIGNAL(clicked()), SLOT(intensity_calib_open_wl()));
    connect(pixelCalibOpn, SIGNAL(clicked()), SLOT(intensity_calib_open_px()));

    calibEn = new QCheckBox(tr("Enable WL Intensity calib"));
    pixelCalibEn = new QCheckBox(tr("Enable pixel Intensity calib"));
    connect(calibEn, SIGNAL(toggled(bool)), SLOT(intensity_calib_enabled()));
    connect(pixelCalibEn, SIGNAL(toggled(bool)), SLOT(intensity_calib_enabled()));
}

void CCDPlot::createLayout()
{
    QWidget *w = this;

    QHBoxLayout *MainLayout = new QHBoxLayout(this);
    MainLayout->setContentsMargins(9, 9, 3, 9);

    QGroupBox *ccdControl = new QGroupBox(w);
    HFlowLayout *layout_ccdControl = new HFlowLayout(ccdControl);
    layout_ccdControl->spaceX = 6; layout_ccdControl->spaceY = 6;
    layout_ccdControl->setContentsMargins(3, 3, 3, 3);

    LeftScroll = new TScrollArea(w);

    MyPl->TwoDPlot->layout()->setContentsMargins(0, 0, 6, 0);
    MainLayout->addWidget(ccdControl);
    MainLayout->addWidget(LeftScroll);
    MainLayout->addWidget(MyPl, 1);


    QWidget *LeftWid = new QWidget(LeftScroll);
    LeftScroll->setWidget(LeftWid);
    QVBoxLayout *layout_ccdOptions = new QVBoxLayout(LeftWid);
    layout_ccdOptions->setSpacing(3);
    layout_ccdOptions->setContentsMargins(0, 6, 6, 6);

    tabs_options = new QTabWidget(LeftWid);
    tabs_options->setDocumentMode(true);
    layout_ccdOptions->addWidget(tabs_options);

    TTabWidget *calib_tabs = new TTabWidget(LeftWid);
    calib_tabs->setDocumentMode(true);
    layout_ccdOptions->addWidget(calib_tabs);


    QWidget *ccdsett = new QWidget(tabs_options);
    QVBoxLayout *layoutccd = new QVBoxLayout(ccdsett);
    layoutccd->setContentsMargins(0, 0, 0, 0);
    layoutccd->setSpacing(3);
    layoutccd->setAlignment(Qt::AlignTop);

    box_monochromator = new QWidget(tabs_options);
    QFormLayout *layout_monochromator = makeFormLayout(box_monochromator);

    QWidget *box_laser = new QWidget(tabs_options);
    QFormLayout *laser_l = makeFormLayout(box_laser);

    tabs_options->addTab(ccdsett, tr("Settings"));
    tabs_options->addTab(box_monochromator, tr("Monochromator"));
    tabs_options->addTab(box_laser, tr("Laser control"));


    layout_monochromator->addRow(central_wl_reset);
    layout_monochromator->addRow(tr("Wavelength (nm)"), lineCentralWlMonoch);
    layout_monochromator->addRow(tr("Grating"), CBgrating_select);
    layout_monochromator->addRow(tr("Entrance slit (nm)"), lineSlitInMonoch);
    layout_monochromator->addRow(tr("Exit slit (nm)"), lineSlitOutMonoch);
    layout_monochromator->addRow(tr("Exit mirror"), CBexitport);


    if (cobolt_gui_list.size() != 0) {
        QTabWidget *laserTabWidget = new QTabWidget(box_laser);
        laser_l->addRow(laserTabWidget);

        foreach (CoboltGui *it, cobolt_gui_list)
            laserTabWidget->addTab(it, it->windowTitle());
    }
    {
        QSettings settings("Param/ccd.conf", QSettings::IniFormat);
        if (!settings.value("use_ars", true).toBool()) {
            laser_l->addRow(tr("ND Filter"), NDEdit);
            laser_l->addRow(tr("Pre-monoch wl"), linePreCentralWlMonoch);
        }
    }
    laser_l->addRow(tr("Optics configuration"), CBPreMonochWl);


    QGroupBox *box_ccdSettings = new QGroupBox(ccdsett);
    QFormLayout *layout_ccdSettings = makeFormLayout(box_ccdSettings);

    temperature_widget = new QWidget(box_ccdSettings);
    QHBoxLayout *temperature_layout = new QHBoxLayout(temperature_widget);

    layoutccd->addWidget(box_ccdSettings);


    //common
    layout_ccdControl->addWidget(PBToNewWindow);
    layout_ccdControl->addWidget(m_BStart);
    //2d only
    addDelimeterH(layout_ccdControl, PBToNewWindow->sizeHint().width());
    layout_ccdControl->addWidget(One_Cont);
    layout_ccdControl->addWidget(Cut_Pikes);
    layout_ccdControl->addWidget(Base_line);
    layout_ccdControl->addWidget(base_line_save);
    layout_ccdControl->addWidget(base_line_manual);
    layout_ccdControl->addWidget(BPanoram);
    //common
    addDelimeterH(layout_ccdControl, PBToNewWindow->sizeHint().width());
    layout_ccdControl->addWidget(BPMTScan);
    layout_ccdControl->addWidget(Monochr);


    PanoramMin = new Edit(200, 800, 1);
    PanoramMin->setValue(200);
    connect(PanoramMin, SIGNAL(newValue(double)), this, SLOT(onPanoramMin(double)));

    PanoramMax = new Edit(200, 800, 1);
    PanoramMax->setValue(800);
    connect(PanoramMax, SIGNAL(newValue(double)), this, SLOT(onPanoramMax(double)));

    panoramControl = new QGroupBox(tr("Panoram Control"), LeftWid);
    panoramControl->hide();
    QFormLayout *layoutPanoram = makeFormLayout(panoramControl);

    QGroupBox *box_Zpos = new QGroupBox(tr("Objective Z, ")+tr("micro")+tr("m"), LeftWid);
    QHBoxLayout *layout_Zpos = new QHBoxLayout(box_Zpos);
    layout_Zpos->addWidget(zPos);

    layout_ccdOptions->addWidget(panoramControl);
    layout_ccdOptions->addWidget(box_Zpos);
    layout_ccdOptions->addStretch(1);

    layoutPanoram->addRow(tr("From, nm"), PanoramMin);
    layoutPanoram->addRow(tr("To, nm"), PanoramMax);


    layout_ccdSettings->addRow(temperature_widget);
    layout_ccdSettings->addRow(sendTemp, lineSetTemp); //custome left widget breaks minsize estimation
    layout_ccdSettings->addRow(tr("Exposition time") + ", " + tr("s"), lineExpTime);
    if (show_additional_gui()) {
        layout_ccdSettings->addRow(tr("Acquisition mode"), setAcqMode);
        layout_ccdSettings->addRow(tr("Read mode"), setReadMode);
    }

    temperature_layout->addWidget(Termometer, 1);
    temperature_layout->addWidget(temp_pic);
    temperature_layout->addWidget(lb_temp);


    box_accumulation = new QGroupBox(tr("Accumulation options"), ccdsett);
    box_accumulation->hide();
    QFormLayout *layout_accumulation = makeFormLayout(box_accumulation);

    Kin_series = new QGroupBox(tr("Kinetic Series options"), ccdsett);
    Kin_series->hide();
    QFormLayout *layout_kin_series_par = makeFormLayout(Kin_series);

    box_singleTrack = new QGroupBox(tr("Single-track options"), ccdsett);
    box_singleTrack->hide();
    QFormLayout *layout_singleTrack = makeFormLayout(box_singleTrack);

    layoutccd->addWidget(box_accumulation);
    layoutccd->addWidget(Kin_series);
    layoutccd->addWidget(box_singleTrack);

    layout_accumulation->addRow(tr("Count"), lineNumAcc);
    layout_accumulation->addRow(tr("Delay") + ", " + tr("s"), lineAccCycleTime);
    layout_accumulation->addRow(tr("Trigger mode"), setTriggerMode);

    layout_kin_series_par->addRow(tr("Count"), Num_Kin_ser);
    layout_kin_series_par->addRow(tr("Delay") + ", " + tr("s"), Time_Kin_ser);
    QHBoxLayout *kinserdacl = new QHBoxLayout;
    layout_kin_series_par->addRow(tr("DAC step"), kinserdacl);
    kinserdacl->addWidget(kinSerDACscan);
    kinserdacl->addWidget(kinSerDACstep);

    layout_singleTrack->addRow(tr("Center row"), lineCenterRowTrack);
    layout_singleTrack->addRow(tr("Height"), lineHeighTrack);



    QGroupBox *box_calibration = new QGroupBox(calib_tabs);
    QFormLayout *layout_calibration = makeFormLayout(box_calibration);
    QGroupBox *box_calibration2 = new QGroupBox(calib_tabs);
    QFormLayout *layout_calibration2 = makeFormLayout(box_calibration2);
    QGroupBox *box_calibration3 = new QGroupBox(calib_tabs);
    QGridLayout *layout_calibration3 = new QGridLayout(box_calibration3);

    calib_tabs->addTab(box_calibration, tr("Calibrations"));
    calib_tabs->addTab(box_calibration2, tr("Intensity"));
    calib_tabs->addTab(box_calibration3, tr("Nonlinear"));
    if (!show_advanced_gui()) {
        calib_tabs->setTabVisible(1, false);
        calib_tabs->setTabVisible(2, false);
    }

    layout_calibration->addRow(lb_auto, AutoCalibration);
    layout_calibration->addRow(tr("Scale"), scaleBox);
    layout_calibration->addRow(tr("Central Pixel"), lineCentralPx);
    layout_calibration->addRow(lb_laser, lineLaserWl);
    layout_calibration->addRow(lb_centralWL, lineCentrWL);
    layout_calibration->addRow(lb_dispersion, lineDispersion);

    layout_calibration2->addRow(calibEn);
    layout_calibration2->addRow(calibOpn, calibLbl);
    layout_calibration2->addRow(pixelCalibEn);
    layout_calibration2->addRow(pixelCalibOpn, pixelCalibLbl);

    layout_calibration3->setAlignment(Qt::AlignTop);
    layout_calibration3->addWidget(BNLCNone = new QRadioButton(tr("Disabled")), 0, 0, 1, 2);
    layout_calibration3->addWidget(BNLCAuto = new QRadioButton(tr("Auto - device")), 1, 0, 1, 2);
    layout_calibration3->addWidget(BNLCFormula = new QRadioButton(tr("Formula")), 2, 0, 1, 2);
    layout_calibration3->addWidget(BNLCFile = new QRadioButton(tr("Manual list")), 3, 0, 1, 2);
    BNLCAuto->setChecked(true);

    QLabel *NLCStatus;
    layout_calibration3->addWidget(BNLCCfg = new QPushButton(tr("Edit")), 5, 0);
    layout_calibration3->addWidget(NLCStatus = new QLabel("-"), 5, 1);
    BNLCCfg->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    BNLCCfg->setEnabled(false);

    NLCEdit = new CCDNLCEditor(this);
    connect(NLCEdit, SIGNAL(newStatus(QString)), NLCStatus, SLOT(setText(QString)));
    connect(BNLCCfg, SIGNAL(clicked()), NLCEdit, SLOT(show()));
    connect(BNLCCfg, SIGNAL(clicked()), NLCEdit, SLOT(raise()));

    connect(BNLCNone, SIGNAL(toggled(bool)), NLCEdit, SLOT(hide()));
    connect(BNLCAuto, SIGNAL(toggled(bool)), NLCEdit, SLOT(hide()));

    connect(BNLCNone, SIGNAL(toggled(bool)), NLCEdit, SLOT(setMode()));
    connect(BNLCAuto, SIGNAL(toggled(bool)), NLCEdit, SLOT(setMode()));
    connect(BNLCFormula, SIGNAL(toggled(bool)), NLCEdit, SLOT(setMode()));
    connect(BNLCFile, SIGNAL(toggled(bool)), NLCEdit, SLOT(setMode()));

    connect(BNLCFormula, SIGNAL(toggled(bool)), NLCEdit, SLOT(show()));
    connect(BNLCFile, SIGNAL(toggled(bool)), NLCEdit, SLOT(show()));
    connect(BNLCFormula, SIGNAL(toggled(bool)), NLCEdit, SLOT(raise()));
    connect(BNLCFile, SIGNAL(toggled(bool)), NLCEdit, SLOT(raise()));
}

QList<QWidget *> CCDPlot::getDeviceStatusList()
{
    QList<QWidget *> list;

    list << (QWidget *)2; //group id
    list << deviceStatus;
    list << deviceStatus2;

    if (cobolt_gui_list.size()) {
        list << (QWidget *)3; //group id
        list << new QLabel(QMainWindow::tr("Lasers:"));
    }
    for (int i = 0; i < cobolt_gui_list.size(); i++)
        list << cobolt_gui_list.at(i)->deviceStatus;

    return list;
}


void CCDPlot::setDeviceStatus2(int idx)
{
    if (!deviceStatus2)
        return;

    switch (idx) {
    case EDno:
        deviceStatus2->status->set(StatusWidget::Cross, Qt::red);
        deviceStatus2->status->setToolTip(tr("No device"));
        break;
    case EDok:
        deviceStatus2->status->set(StatusWidget::Circle, Qt::green);
        deviceStatus2->status->setToolTip(tr("Connected"));
        break;
    case EDplay:
        deviceStatus2->status->set(StatusWidget::Play, Qt::green);
        deviceStatus2->status->setToolTip(tr("Busy"));
        break;
    }
}

void CCDPlot::setDeviceStatus_handler(int idx)
{
    syncDevStatus = idx;
    if (!devStatusSync->isActive())
        devStatusSync->start(100);
}
void CCDPlot::setDeviceStatus()
{
    if (!deviceStatus || prevDevStatus == syncDevStatus)
        return;

    prevDevStatus = syncDevStatus;
    bool st = (syncDevStatus >= EDok);

    switch (syncDevStatus) {
    case EDno:
        deviceStatus->status->set(StatusWidget::Cross, Qt::red);
        deviceStatus->status->setToolTip(tr("No device"));
        break;
    case EDok:
        deviceStatus->status->set(StatusWidget::Circle, Qt::green);
        deviceStatus->status->setToolTip(tr("Connected"));
        break;
    case EDplay:
        deviceStatus->status->set(StatusWidget::Play, Qt::green);
        deviceStatus->status->setToolTip(tr("Acquiring"));
        break;
    }

    //ccdsett->setEnabled(st);
    //m_BStart->setEnabled(st);

    bool st1 = st && ccdMatrix->hasTemperature;

    bool temp_upd = temperature_widget->isVisibleTo(this) != st1;
    temperature_widget->setVisible(st1);
    sendTemp->setVisible(st1);
    lineSetTemp->setVisible(st1);

    if (!st1) {
        temperatureUpd->stop();
    } else {
        if (!temperatureUpd->isActive())
            temperatureUpd->start(1000);
    }

    void *prev = ccdWidget;
    if (!st && ccdWidget) {
        delete ccdWidget;
        ccdWidget = 0;
    }

    if (st && !ccdWidget) {
        ccdWidget = ccdMatrix->getCCDInterfaceWidget();
        if (ccdWidget)
            tabs_options->insertTab(1, ccdWidget, tr("CCD"));
    }

    if (prev != ccdWidget || temp_upd) {
        LeftScroll->setWidget(LeftScroll->takeWidget()); //reset sizeHint
        LeftScroll->updateGeometry(); //notify layout
    }
}

void CCDPlot::monoch_status_update(bool st)
{
    setDeviceStatus2(st ? EDok : EDplay);

    box_monochromator->setEnabled(st);
    NDEdit->setEnabled(st && Nspec_Initalized);
    linePreCentralWlMonoch->setEnabled(st && Nspec_Initalized);
    CBPreMonochWl->setEnabled(st && Nspec_Initalized);

    if (!st)
        return;

    if (Nspec_Initalized && nspectr_mutex) {
        QMutexLocker lck(nspectr_mutex);
        NDEdit->setValue(NProxy_GetExitNDFilter());
        linePreCentralWlMonoch->setValue(NProxy_GetWL());
    }

    if (CBgrating_select->isEnabled()) {
        CBgrating_select->blockSignals(true);
        CBgrating_select->setCurrentIndex(ccdMatrix->m_spectrInterface->getActiveGrating());
        CBgrating_select->blockSignals(false);
    }

    lineSlitInMonoch->setValue(ccdMatrix->m_spectrInterface->getSlitWidth(SPECTR_PORT_IN));
    lineSlitOutMonoch->setValue(ccdMatrix->m_spectrInterface->getSlitWidth(SPECTR_PORT_OUT));
    lineCentralWlMonoch->setValue(ccdMatrix->m_spectrInterface->getWl());
    CBexitport->blockSignals(true);
    CBexitport->setCurrentIndex(ccdMatrix->m_spectrInterface->getPortPosition(SPECTR_PORT_OUT));
    CBexitport->blockSignals(false);
}
void CCDPlot::monoch_laser_wl()
{
    int idx = CBPreMonochWl->currentIndex();
    if (idx == -1)
        return;

    QSettings settings("Param/laser.conf", QSettings::IniFormat);
    settings.setValue("index", idx);

    if (settings.value("set_wl", false).toBool()) {
        if (idx < 0 || idx >= laser_list_wavelength.size()) {
            qDebug() << "wl list fail";
        } else {
            lineLaserWl->setValue(laser_list_wavelength.at(idx));
            lineLaserWl->setNewValue();
            linePreCentralWlMonoch->setValue(laser_list_wavelength.at(idx));
            linePreCentralWlMonoch->setNewValue();
        }
    }

    if (nspectr_mutex) {
        QMutexLocker lck(nspectr_mutex);

        if (settings.value("set_ef", false).toBool() &&
                idx >= 0 && idx < laser_list_edgefilter_idx.size())
            NProxy_SetEdgeFilterPos(laser_list_edgefilter_idx.at(idx));

        if (settings.value("set_be", false).toBool() &&
                idx >= 0 && idx < laser_list_beamexpander_wl_idx.size() && idx < laser_list_beamexpander_diam_idx.size())
            NProxy_SetBeamExpanderPos(laser_list_beamexpander_wl_idx.at(idx),
                                      laser_list_beamexpander_diam_idx.at(idx));

        if (settings.value("set_mxyz", false).toBool() &&
                idx >= 0 && idx < laser_list_monoch_x.size() && idx < laser_list_monoch_y.size() && idx < laser_list_monoch_z.size())
            NProxy_SetMonochromatorLensPos(laser_list_monoch_z.at(idx),
                                           laser_list_monoch_x.at(idx),
                                           laser_list_monoch_y.at(idx), true);

        if (settings.value("set_cxyz", false).toBool()  &&
                idx >= 0 && idx < laser_list_confocal_pmt_x.size() && idx < laser_list_confocal_pmt_y.size() && idx < laser_list_confocal_pmt_z.size())
            NProxy_SetConfocalLensPos(laser_list_confocal_pmt_z.at(idx),
                                      laser_list_confocal_pmt_x.at(idx),
                                      laser_list_confocal_pmt_y.at(idx), true);
    }

    int laser_idx = CBPreMonochWl->itemData(CBPreMonochWl->currentIndex()).toInt();
    for (int i = 0; i < cobolt_gui_list.size(); i++) {
        if ((!(i == laser_idx) != cobolt_gui_list.at(i)->Laser_onoff->on_off) &&
                cobolt_gui_list.at(i)->isEnabled())
            cobolt_gui_list.at(i)->Laser_onoff->change_state_man_with_sig(!(i == laser_idx));
    }
}
void CCDPlot::nd_filter(double idx)
{
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    if (settings.value("use_ars", true).toBool()) {

    } else if (nspectr_mutex) {
        QMutexLocker lck(nspectr_mutex);
        NProxy_SetExitNDFilter(idx, FALSE);
    }
}
void CCDPlot::monoch_pre_wl()
{
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    if (settings.value("use_ars", true).toBool()) {
        ARS3000::setLaser(CBPreMonochWl->currentIndex());
    } else if (nspectr_mutex) {
        QMutexLocker lck(nspectr_mutex);
        NProxy_SetWL(linePreCentralWlMonoch->value(), true);
    }
}
void CCDPlot::monoch_grating()
{
    if (!ccdMatrix->m_spectrInterface)
        return;

    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    settings.setValue("central_px_gr_"+QString::number(ccdMatrix->m_spectrInterface->getActiveGrating()), lineCentralPx->text());

    ccdMatrix->m_spectrInterface->setActiveGrating(CBgrating_select->currentIndex());

    lineCentralPx->setValue(settings.value("central_px_gr_"+QString::number(CBgrating_select->currentIndex()), lineCentralPx->value()).toDouble());
    lineCentralPx->setNewValue();
}
void CCDPlot::monoch_wl()
{
    ccdMatrix->m_spectrInterface->setWl(lineCentralWlMonoch->value(), central_wl_reset->isChecked());
}
void CCDPlot::monoch_slit_in()
{
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    if (settings.value("use_ars", true).toBool()) {
        ARS3000::setSlitWidth(0, lineSlitInMonoch->value() * 1e-3);
    }
    ccdMatrix->m_spectrInterface->setSlitWidth(SPECTR_PORT_IN, lineSlitInMonoch->value());
}
void CCDPlot::monoch_slit_out()
{
    ccdMatrix->m_spectrInterface->setSlitWidth(SPECTR_PORT_OUT, lineSlitOutMonoch->value());
}
void CCDPlot::monoch_mirror()
{
    ccdMatrix->m_spectrInterface->setPortPosition(SPECTR_PORT_OUT, CBexitport->currentIndex());
}
