#include "modules.hpp"
#include <QDebug>

class EG3000Module;
class DataModule;
class CCDMatrixModule;
class CCDPlotModule;
class ScanModule;
class SPMModule;
class MotorsModule;
class OscilloscopeModule;
class VideoModule;


ModuleManager *ModuleManager::instance = 0;

I_MODULE::~I_MODULE()
{
    if (ModuleManager::instance)
        ModuleManager::instance->unload(this);
}

ModuleManager::ModuleManager()
{
    if (instance)
        throw instance;
    instance = this;
    isRunning = false;

    ModuleManager::instance->addToFactory<EG3000Module>();
    ModuleManager::instance->addToFactoryGui<DataModule>();
    ModuleManager::instance->addToFactory<CCDMatrixModule>();
    ModuleManager::instance->addToFactoryGui<CCDPlotModule>();
    ModuleManager::instance->addToFactoryGui<ScanModule>();
    ModuleManager::instance->addToFactoryGui<SPMModule>();
    ModuleManager::instance->addToFactoryGui<MotorsModule>();
    ModuleManager::instance->addToFactoryGui<OscilloscopeModule>();
    ModuleManager::instance->addToFactoryGui<VideoModule>();
}

ModuleManager::~ModuleManager()
{
    instance = 0;

    if (isRunning)
        foreach (I_MODULE *m, modules)
            m->setRunning(false);

    foreach (I_MODULE *m, modules)
        delete m;

    modules.clear();
}

I_MODULE *ModuleManager::load(const QLatin1String name, QWidget *parent)
{
    if (!factory.contains(name))
        return 0;

    QList<QLatin1String> l;
    if (factory[name].requires)
        l = factory[name].requires();
    for (int i = 0; i < l.size(); i++)
        if (!modules.contains(l.at(i)))
            if (!load(l.at(i), parent))
                qCritical() << "can't create module" << QString(l.at(i)) << "required for" << QString(name);

    I_MODULE *m;
    if (factory[name].creategui)
        m = factory[name].creategui(parent);
    else
        m = factory[name].create();
    add(name, m);
    return m;
}

void ModuleManager::add(const QLatin1String name, I_MODULE *module)
{
    QMapIterator<QLatin1String, I_MODULE *> i(modules);
    while (i.hasNext()) {
        i.next();
        i.value()->connected(name, module);
        module->connected(i.key(), i.value());
    }

    if (isRunning)
        module->setRunning(true); //TODO fix order for dependencies

    modules.insert(name, module);
}

I_MODULE *ModuleManager::find1(const QLatin1String name) const
{
    return modules.value(name);
}
const QList<I_MODULE *> ModuleManager::find(const QLatin1String name) const
{
    return modules.values(name);
}

const QLatin1String ModuleManager::find(I_MODULE *module) const
{
    foreach (QLatin1String name, modules.keys(module))
        return name;

    return QLatin1String(0);
}

void ModuleManager::unload(const QLatin1String name)
{
    QList<I_MODULE *> l = modules.values(name);
    if (isRunning)
        foreach (I_MODULE *m0, l)
            m0->setRunning(false);

    foreach (I_MODULE *m0, l)
        foreach (I_MODULE *m, modules)
            if (m != m0)
                m->disconnected(name, m0);

    modules.remove(name);

    foreach (I_MODULE *m0, l)
        delete m0;
}

void ModuleManager::unload(I_MODULE *module)
{
    QLatin1String name = find(module);
    if (!name.latin1())
        return;

    if (isRunning)
        module->setRunning(false);

    foreach (I_MODULE *m, modules)
        if (m != module)
            m->disconnected(name, module);

    modules.remove(name, module);
}

void ModuleManager::setRunning(bool running)
{
    if (isRunning == running)
        return;

    isRunning = running;

    foreach (I_MODULE *m, modules)
        m->setRunning(isRunning);
}

//---------------------------------------------------------------------------------------
#include <QSettings>
#include "mainwindow.h"
#include "mobile/remote_mobile.hpp"
#include "gui/status.hpp"

#include "hardware/EG3000.h"
class EG3000Module : public EG3000_CONTROL, public I_GUI_MODULE_BASE
{
public:
    EG3000Module() {
#ifdef USE_EG3K_EMULATOR
        QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
        if (settings.value("usb", false).toBool()) {
            SWusb1->setEmulator(true);
            return;
        }
#endif

        start();
    }
    static const QLatin1String name() { return QLatin1String("eg3000"); }
    void *basecast(const QLatin1String name) {
        if (name == QLatin1String("gui"))
            return static_cast<I_GUI_MODULE_BASE *>(this);
        return name.latin1() != 0 ? 0 : static_cast<EG3000_CONTROL *>(this); }
    void connected(const QLatin1String name, I_MODULE *module);
    void setRunning(bool on) {
        if (on) {
            m_waitMainInit = false;
#ifdef USE_EG3K_EMULATOR
            QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
            if (settings.value("usb", false).toBool())
                emit usbStatus(EDok);
#endif
        } else
            thread_close();
    }

    QList<QWidget *> getStatusWidgets() {
        QList<QWidget *> l;
        l << (QWidget *)0 << SWusb1;
        return l;
    }
};

#include "dataplot.h"
class DataModule : public I_GUI_MODULE<DataPlot>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(DataModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        Container->SetAutoSaveTime(settings.value("autosave_time", 30).toInt() * 60 * 1000); //to msec
    }
    static const QLatin1String name() { return QLatin1String("data"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<DataPlot>::basecast(name); }
    void setRunning(bool on) { if (!on) Container->thread_close(); }

    //gui
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/Data_plot.png"), MainWindow::tr("Data")); }

    //net
    void initialSend(MobileInstance *client);
    void setValue(int id, QVariant val, QObject *sender);
};

#include "ccd/ccdmatrix.h"
#include "ccd/i_spectr.hpp"
class CCDMatrixModule : public CCDMatrix, public I_MODULE
{
public:
    static const QLatin1String name() { return QLatin1String("ccdmatrix"); }
    void *basecast(const QLatin1String name) { return name.latin1() != 0 ? 0 : static_cast<CCDMatrix *>(this); }
    void setRunning(bool on) {
        if (on)
            m_waitMainInit = false;
        else {
            thread_close();

            if (m_spectrInterface)
                m_spectrInterface->shutdown();
        }
    }
};

#include "ccd/ccdplot.h"
class CCDPlotModule : public I_GUI_MODULE<CCDPlot>
{
public:
    GUI_MODULE_CTOR(CCDPlotModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        AutoSend = settings.value("scan/autosend_data", true).toBool();
    }
    static const QLatin1String name() { return QLatin1String("ccdplot"); }
    static const QList<QLatin1String> requires() { return QList<QLatin1String>() << QLatin1String("ccdmatrix"); }
    void connected(const QLatin1String name, I_MODULE *module);
    void setRunning(bool on) { if (on) m_BStart->click(); }

    QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/ccd.png"), MainWindow::tr("Spectroscopy")); }
    QList<QWidget *> getStatusWidgets() { return QList<QWidget *>() << getDeviceStatusList(); }
};

#include "scanplot.h"
class ScanModule : public I_GUI_MODULE<ScanPlot>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(ScanModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        Autosend = settings.value("scan/autosend_data", true).toBool();
    }
    static const QLatin1String name() { return QLatin1String("scan"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<ScanPlot>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module);

    //gui
    QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/Scan.png"), MainWindow::tr("Scan")); }

    //net
    void initialSend(MobileInstance *client);
    void setValue(int id, QVariant val, QObject *sender);
};

#include "spmtunning.h"
#include "deflect_widget.hpp"
class SPMModule : public I_GUI_MODULE<SPMTunning>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(SPMModule) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        AutoSend = settings.value("scan/autosend_data", true).toBool();
    }
    static const QLatin1String name() { return QLatin1String("spm"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<SPMTunning>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module);
    void setRunning(bool on) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (on) {
            BLaserOnOff->change_state_man_with_sig(
                        settings.value("state/spm/laser_on", true).toBool());

            bool head, stage;
            {
                QSettings settings("Param/gui.conf", QSettings::IniFormat);
                head = settings.value("display/oscilloscope/head", true).toBool();
                stage = settings.value("display/oscilloscope/stage", true).toBool();
            }
            int fb_en = 0;
            if (!head) fb_en = 2;
            if (!stage) fb_en = 1;
            if (fb_en != 0) {
                CBFBScanner->setEnabled(false);
                CBFBScanner->hide();
                disconnect(CBFBScanner, 0, LaserSpot, 0);
                disconnect(CBFBScanner, 0, LaserSpotWindow, 0);
                LaserSpot->setFbscanner(-1);
                LaserSpotWindow->setFbscanner(-1);
            }
            if (fb_en == 2) CBFBScanner->setCurrentIndex(1);
            if (fb_en == 0) {
                QSettings settings("Param/gui.conf", QSettings::IniFormat);
                CBFBScanner->setCurrentIndex(settings.value("state/spm/fb_scanner", 0).toInt());
            }
        } else {
            settings.setValue("state/spm/laser_on", BLaserOnOff->on_off);
            settings.setValue("state/spm/fb_scanner", CBFBScanner->currentIndex());
        }
    }

    //gui
    QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/MainToolbar/Spm.png"), MainWindow::tr("SPM")); }
    QList<QWidget *> getStatusWidgets() {
        QList<QWidget *> l;
        l << (QWidget *)0 << SWspmfb;
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (settings.value("display/shuttle_status", true).toBool())
            l << (QWidget *)1 << shuttleStatus;
        return l;
    }

    //net
    void initialSend(MobileInstance *client);
    void setValue(int id, QVariant val, QObject *sender);
};

#include "motors.hpp"
class MotorsModule : public I_GUI_MODULE<Motors>, public I_NET_MODULE_BASE
{
    Q_OBJECT
public slots:
    void on_Motors_clicked();
public:
    GUI_MODULE_CTOR(MotorsModule) { }
    static const QLatin1String name() { return QLatin1String("motors"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<Motors>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module);

    //gui
    /*QPushButton *createPanelButton() {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (!settings.value("display/motors", false).toBool())
            return 0;

        return new QPushButton(makeHugeIcon(":/SPMTuning_window/Motors_leveling.png"), MainWindow::tr("Motors"));
    }*/
    QList<QWidget *> getStatusWidgets() { return QList<QWidget *>() << (QWidget *)0 << new QLabel(MainWindow::tr("Motors:")) << SWmotor1; }

    //net
    void setValue(int id, QVariant val, QObject *sender);
};

#include "oscilloscope.h"
#include "autocalib.hpp"
class OscilloscopeModule : public I_GUI_MODULE<Oscilloscope>
{
public:
    GUI_MODULE_CTOR(OscilloscopeModule) { }
    static const QLatin1String name() { return QLatin1String("oscilloscope"); }
    void connected(const QLatin1String name, I_MODULE *module);
    void setRunning(bool on) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (on) {
            for (int i = 0; i < 4; i++)
                CBChannel[i]->setCurrentIndex(
                            qMax(0, CBChannel[i]->findData(
                                     settings.value("state/osc/ch" + QString::number(i), 0))));

            DacsWidget->DacsChannel->setCurrentIndex(
                        qMax(0, DacsWidget->DacsChannel->findText(
                                 settings.value("state/osc/ch_control").toString())));

            CalibWidget->LVMLoad();
            CalibWidget->read_calib_par(CalibWidget->defaultFileName);
        } else {
            for (int i = 0; i < 4; i++)
                settings.setValue("state/osc/ch" + QString::number(i), CBChannel[i]->itemData(CBChannel[i]->currentIndex()));

            settings.setValue("state/osc/ch_control", DacsWidget->DacsChannel->currentText());
        }
    }

    QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (!settings.value("display/oscilloscope", true).toBool())
            return 0;
        return new QPushButton(makeHugeIcon(":/MainToolbar/Osc.png"), MainWindow::tr("Oscilloscope"));
    }
};

#include "USBVideo/videoplayer.h"
class VideoModule : public I_GUI_MODULE<VideoPlayer>, public I_NET_MODULE_BASE
{
public:
    GUI_MODULE_CTOR(VideoModule) { }
    static const QLatin1String name() { return QLatin1String("video"); }
    static const QList<QLatin1String> requires() { return QList<QLatin1String>() << QLatin1String("managedthreadedcleaner"); }
    void *basecast(const QLatin1String name) { return name == QLatin1String("net") ? static_cast<I_NET_MODULE_BASE *>(this) : I_GUI_MODULE<VideoPlayer>::basecast(name); }
    void connected(const QLatin1String name, I_MODULE *module);
    void setRunning(bool on) { if (on) resetCameras(); }

    //gui
    QPushButton *detachButton() { return PBToNewWindow; }
    QPushButton *createPanelButton() { return new QPushButton(makeHugeIcon(":/SPMTuning_window/Video.png"), MainWindow::tr("Video") + (inst_idx == 0 ? QString() : " 2")); }
    QList<QWidget *> getStatusWidgets() { return QList<QWidget *>() << (QWidget *)5 << deviceStatus; }

    //net
    void setValue(int id, QVariant val, QObject *sender);
};

//---------------------------------------------------------------------------------------
//guidelines:
//connect from user module to common service provider module
//connect from user module to hook/notify module
//be wary of possible order dependencies

#include "hardware/relay.hpp"
void EG3000Module::connected(const QLatin1String name, I_MODULE *module)
{
#ifdef USE_RELAY
    if (name == QLatin1String("relay")) {
        RelayModule *m = (RelayModule *)module;
        if (m->relayCount() < 1)
            return;

        m->enableAll();
        QAction *a;
        SWusb1->addAction(a = new QAction(tr("Power up controller"), SWusb1));
        connect(a, SIGNAL(triggered()), m, SLOT(enableAll()));
        SWusb1->addAction(a = new QAction(tr("Power down controller"), SWusb1));
        connect(a, SIGNAL(triggered()), m, SLOT(disableAll()));
    }
#endif
}

#include "gui/nstslider.h"
void CCDPlotModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();

        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)), m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
        connect(this, SIGNAL(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)), m->Container, SLOT(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)));
    }
    if (name == QLatin1String("scan")) {
        ScanPlot *m = (ScanPlot *)module->basecast();

        IExternalScan *sc;
        m->addExternalScan(MainWindow::tr("CCD"), sc = getNewCCDIExtScan(this));
        if (DataModule *data = (DataModule *)ModuleManager::instance->find1("data"))
            connect(sc, SIGNAL(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)),
                    data->Container, SLOT(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)));
    }
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        linkDoubleChangers(m->DacsWidget->Zobj, zPos);
        connect(m->DacsWidget->Zobj->slider, SIGNAL(valueChanged(double)), zPos, SLOT(setValue(double))); //sync value change on setValue
    }
}

#include <QCheckBox>
#include <QTimer>
void ScanModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        EG3000_CONTROL *m = (EG3000_CONTROL *)module->basecast();

        m->rThread.readers.add(this);
        m->wThread.writers.add(this);

        connect(&m->rThread, SIGNAL(capBits(int,long)), this, SLOT(capBits(int,long)));
    }
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();

        connect(m->Container, SIGNAL(restoreAttributes(plot::TMetaStorage)),
                this, SLOT(restoreAttributes(plot::TMetaStorage)));

        //TODO cant connect externalScans here, race. need modules for extscanmodules
        //for now connect individually
        /*foreach (IExternalScan *sc, externalScans)
            if (sc->metaObject()->indexOfSignal("AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)") != -1)
                connect(sc, SIGNAL(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)),
                        m->Container, SLOT(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)));*/

        connect(this, SIGNAL(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::T3dDataOwner, plot::TMetaStorage)));
        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
    if (name == QLatin1String("spm")) {
        SPMTunning *m = (SPMTunning *)module->basecast();

        if (SPMTunning::getLandingMode() != lm01)
            BMotorLevel->show();

        connect(m->DCfbPrec, SIGNAL(newValue()), this, SLOT(check_speed()));
        connect(m->DCfbHeight, SIGNAL(newValue()), this, SLOT(check_speed()));
        connect(m->BFeedbackOnOffMode, SIGNAL(state_changed(bool)), this, SLOT(check_speed()));

        //TODO add tab from spm::connected?
        linkDoubleChangers(m->DCfbP, DCfbP, false);
        linkDoubleChangers(m->DCfbI, DCfbI, false);
        linkDoubleChangers(m->DCfbSetPoint, DCfbSetPoint);
        connect(m->DCfbSetPoint->slider, SIGNAL(valueChanged(double)), DCfbSetPoint, SLOT(setValue(double))); //sync value change on setValue - gains, etc.
    }
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        connect(m->CalibWidget, SIGNAL(sigScanError()), this, SLOT(s_scanError()));
        connect(this, SIGNAL(recalibrateSensors(int)), m->acf, SLOT(doXYcalib(int)));
        connect(m->acf, SIGNAL(calibDone(bool)), this, SLOT(recalibDone(bool)));

        connect(m->DacsWidget, SIGNAL(xUpdate(double)), this, SLOT(sl_setCurrentX(double)));
        connect(m->DacsWidget, SIGNAL(yUpdate(double)), this, SLOT(sl_setCurrentY(double)));
        connect(m->DacsWidget, SIGNAL(x2Update(double)), this, SLOT(sl_setCurrentX2(double)));
        connect(m->DacsWidget, SIGNAL(y2Update(double)), this, SLOT(sl_setCurrentY2(double)));

        connect(this, SIGNAL(sync_scan_x(double)), m->DacsWidget, SIGNAL(xUpdate(double)));
        connect(this, SIGNAL(sync_scan_y(double)), m->DacsWidget, SIGNAL(yUpdate(double)));
        connect(this, SIGNAL(sync_scan_x2(double)), m->DacsWidget, SIGNAL(x2Update(double)));
        connect(this, SIGNAL(sync_scan_y2(double)), m->DacsWidget, SIGNAL(y2Update(double)));

        connect(this, SIGNAL(sig_XYPosition(double, double)), m->DacsWidget, SLOT(Go_to_XY(double, double)));
        connect(this, SIGNAL(sig_XYPosition2(double, double)), m->DacsWidget, SLOT(Go_to_XY2(double, double)));

        connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_set_XYZ_range(double, double, double)));
        connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_set_XYZ_range2(double, double, double)));

        connect(m->CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(sl_CB_update()));
        connect(m->CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(SetSectionUnits()));
    }
}

void SPMModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        EG3000_CONTROL *m = (EG3000_CONTROL *)module->basecast();

        m->rThread.readers.add(this);
        m->wThread.writers.add(this);

        connect(&m->rThread, SIGNAL(capBits(int,long)), this, SLOT(capBits(int,long)));
    }
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();

        connect(m->Container, SIGNAL(restoreAttributes(plot::TMetaStorage)),
                this, SLOT(restoreAttributes(plot::TMetaStorage)));

        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
    if (name == QLatin1String("scan")) {
        ScanPlot *m = (ScanPlot *)module->basecast();

        if (show_advanced_gui()) {
            IExternalScan *sc;
            m->addExternalScan(MainWindow::tr("JM force curve"), sc = getNewJMIExtScan(this));
            if (DataModule *data = (DataModule *)ModuleManager::instance->find1("data"))
                connect(sc, SIGNAL(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)),
                        data->Container, SLOT(AddDataToContainer(ccd::TSetOfSpectrum*,plot::TMetaStorage)));
        }
    }
    if (name == QLatin1String("motors")) {
        Motors *m = (Motors *)module->basecast();
        motors = m;

        m->addCustomSett((QWidget *)OptionsForMotor);

        connect(m, SIGNAL(m123stopped()), this, SLOT(slot_m123stopped()));
        connect(BMotors_leveling, SIGNAL(clicked()), m, SLOT(on_Motors_clicked()));
        connect(BMotors_Options, SIGNAL(clicked()), m, SLOT(show_sett()));
        connect(BMotors_Options, SIGNAL(clicked()), m, SLOT(on_Motors_clicked()));
    }
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        connect(m->acf, SIGNAL(calibInProgress(bool)), this, SLOT(s_setFBmasked(bool)));

        DoubleChanger *Head_Dac = new DoubleChanger(-10, 10);
        Head_Dac->setProperty("id", DataID::SPM_Vtip);
        linkDoubleChangers(Vtip, Head_Dac);
        m->DacsWidget->insertChannel(tr("V Tip"), tr("V"), Head_Dac, 0);

        connect(this, SIGNAL(sig_osc_start()), m, SLOT(sl_osc_start()));
        connect(this, SIGNAL(sig_osc_stop()), m, SLOT(sl_osc_stop()));
        connect(this, SIGNAL(sig_osc_channel(int)), m, SLOT(setChA(int)));
        connect(m->SBT, SIGNAL(newValue(int)), SBT, SLOT(setValue(int)));
        connect(SBT, SIGNAL(newValue(int)), m->SBT, SLOT(setValue(int)));
        connect(m, SIGNAL(oscRefreshed()), this, SLOT(osc_refresh()));


        connect(m->DacsWidget, SIGNAL(zUpdate(double)), NS_Z_spectr, SLOT(setValue(double)));

        connect(this, SIGNAL(sig_Zpos(double)), m->DacsWidget, SLOT(Go_to_Z(double)));
        connect(this, SIGNAL(sig_Z2pos(double)), m->DacsWidget, SLOT(Go_to_Z2(double)));

        connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_set_XYZ_range(double, double, double)));
        connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_set_XYZ_range2(double, double, double)));

        connect(this, SIGNAL(input_calib_updated()), m->CalibWidget, SIGNAL(input_calib_updated()));
        connect(m->CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(sl_setPlotMode()));
    }
}

void MotorsModule::on_Motors_clicked()
{
    MainWindow *mainWindow = (MainWindow *)ModuleManager::instance->find1("mainwindow");
    layout()->activate();
    move(mainWindow->mapToGlobal(mainWindow->rect().center() - rect().center()));
    show();
    raise();
}
void MotorsModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        EG3000_CONTROL *m = (EG3000_CONTROL *)module->basecast();

        m->rThread.readers.add(this);
        m->wThread.writers.add(this);

        connect(&m->rThread, SIGNAL(capBits(int,long)), this, SLOT(capBits(int,long)));
    }
    if (name == QLatin1String("mainwindow")) {
        MainWindow *m = (MainWindow *)module->basecast();

        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (!settings.value("display/motors", false).toBool())
            return;

        QPushButton *pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/Motors_leveling.png"), MainWindow::tr("Motors"));
        pb->setFlat(true);
        connect(pb, SIGNAL(clicked()), this, SLOT(on_Motors_clicked()));

        QToolBar *TBMain = m->findChild<QToolBar *>("MainTB");
        TBMain->addWidget(pb);
    }
}

void OscilloscopeModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("eg3000")) {
        EG3000_CONTROL *m = (EG3000_CONTROL *)module->basecast();

        m->rThread.readers.add(this);
        m->rThread.readers.add(acf);
        m->wThread.writers.add(this);
        m->wThread.writers.add(DacsWidget);

        connect(&m->wThread, SIGNAL(started()), acf, SLOT(reset()));
        connect(&m->rThread, SIGNAL(capBits(int,long)), acf, SLOT(capBits(int,long)));

        connect(m, SIGNAL(device_status(int)), this, SLOT(device_status(int)));

        //calib
        m->rThread.readers.add(CalibWidget);
        m->wThread.writers.add(CalibWidget);

        connect(&m->rThread, SIGNAL(capBits(int,long)), CalibWidget, SLOT(capBits(int,long)));

        connect(CalibWidget, SIGNAL(program_controller(QString)), m, SLOT(program(QString)));
    }
    if (name == QLatin1String("data")) {
        DataPlot *m = (DataPlot *)module->basecast();
        connect(this, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)),
                m->Container, SLOT(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
}

void VideoModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == QLatin1String("oscilloscope")) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        linkDoubleChangers(m->DacsWidget->Zobj, zPos);
        connect(m->DacsWidget->Zobj->slider, SIGNAL(valueChanged(double)), zPos, SLOT(setValue(double))); //sync value change on setValue

        connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_set_XYZ_range(double, double, double)));
        connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_set_XYZ_range2(double, double, double)));
    }
}


//---------------------------------------------------------------------------------------
#include "container/datalist.hpp"
void DataModule::initialSend(MobileInstance *client)
{
    QDir dir(TContainer::getDataDir());
    client->serv->setPropIdVarS(client->socket, DataID::DataOpenList, dir.entryList(QStringList() << "*.nstdat"));
}
void DataModule::setValue(int id, QVariant val, QObject *)
{
    if (id == DataID::DataOpenName) {
        QString fname = TContainer::getDataDir() + "/" + val.toString();
        if (val.toString().startsWith("NST_TAB_IDX/")) {
            bool ok = false;
            int idx = val.toString().split("/").last().toInt(&ok);
            if (!ok) return;
            Container->CBDataLists->setCurrentIndex(idx);
            Container->SendContainerData();
            return;
        }
        for (int i = 1; i < Container->CBDataLists->count(); i++) {
            TListWidget *w = ((TListWidget *)Container->CBDataLists->widget(i));
            if (QFileInfo(w->filename).absoluteFilePath() == QFileInfo(fname).absoluteFilePath()) {
                Container->CBDataLists->setCurrentIndex(i);
                Container->SendContainerData();
                return;
            }
        }
        Container->OpenFile(fname);
    }
    if (id == DataID::ActiveData) {
        int id = qMin(Container->CurrentDataList->findItems("*", Qt::MatchWildcard).size()-1, val.toInt());
        if (id < 0) return;
        TListItem *it = Container->CurrentDataList->FindItem(Container->CurrentDataList->findItems("*", Qt::MatchWildcard).at(id));
        if (!it) return;
        Container->CurrentDataList->on_itemPressed(it->Item);
    }
}

void ScanModule::initialSend(MobileInstance *client)
{
    QStringList chnames;
    QList<QVariant> chdata;
    for (int i = 0; i < CBChannel[1]->count(); i++) {
        chnames.append(CBChannel[1]->itemText(i));
        chdata.append(CBChannel[1]->itemData(i));
    }
    client->serv->setPropIdVarS(client->socket, DataID::ChannelList, QList<QVariant>() << chnames << QVariant(chdata));

    client->serv->setPropIdVarS(client->socket, DataID::ScanNxNy, qSqrt(Nx_points * Ny_points));
    if (!m_BStart->on_off) {
        Send2dPlotData();
        Send3dPlotData1();
        Send3dPlotData2();
    }
}
void ScanModule::setValue(int id, QVariant val, QObject *)
{
    if (id == DataID::ScanOnToggle)
        m_BStart->animateClick();

    if (id == DataID::ScanNxNy) {
        int i = val.toInt();
        if (show_additional_gui())
            i += 3;

        CBPointsOX->blockSignals(true);
        CBPointsOY->blockSignals(true);
        CBPointsOXext->blockSignals(true);
        CBPointsOYext->blockSignals(true);
        CBPointsOX->setCurrentIndex(i);
        CBPointsOY->setCurrentIndex(i);
        CBPointsOXext->setCurrentIndex(i);
        CBPointsOYext->setCurrentIndex(i);
        CBPointsOX->blockSignals(false);
        CBPointsOY->blockSignals(false);
        CBPointsOXext->blockSignals(false);
        CBPointsOYext->blockSignals(false);
        check_nx_ny();
        sl_ext_mode();
    }
    if (id == DataID::ScanSaveCh) {
        if (val.toInt() == 4) {
            Plot->TwoDPlot->PlotControl->BSave->animateClick();
        } else {
            Plot->SetActivePlot(val.toInt());
            Plot->PlotControl->BSave->animateClick();
        }
    }

    if (id == DataID::PlotScalesChanged) {
        QVariantList lst = val.toList();
        plot::TScalesChangedEvent e(Range(), Range(), (plot::TScalesChangedEvent::TReason)lst.at(0).toInt());
        e.XLimits[0] = lst.at(1).toDouble();
        e.XLimits[1] = lst.at(2).toDouble();
        e.YLimits[0] = lst.at(3).toDouble();
        e.YLimits[1] = lst.at(4).toDouble();
        Plot->SetPlotsXScales(e.XLimits[0], e.XLimits[1]);
        Plot->SetPlotsYScales(e.YLimits[0], e.YLimits[1]);
        eventFilter(this, &e);
    }
}

void SPMModule::initialSend(MobileInstance *client)
{
    QStringList chnames;
    for (int i = 0; i < fbs->count(); i++)
        chnames.append(fbs->itemText(i));
    client->serv->setPropIdVarS(client->socket, DataID::SPM_FBch_list, chnames);
}
void SPMModule::setValue(int id, QVariant val, QObject *sender)
{
    if (id == DataID::Deflect_send)
        if (sender)
            sender->setProperty("deflect_need", val);
}

void MotorsModule::setValue(int id, QVariant val, QObject *)
{
    if (id == DataID::Motor_123)
        goShuttle(val.toInt(), 250);
    if (id == DataID::Motor_1)
        goShuttle(val.toInt(), 1);
    if (id == DataID::Motor_2)
        goShuttle(val.toInt(), 2);
    if (id == DataID::Motor_3)
        goShuttle(val.toInt(), 3);
    if (id == DataID::Motor_4)
        goShuttle(val.toInt(), 4);
}

void VideoModule::setValue(int id, QVariant val, QObject *sender)
{
    if (inst_idx != 0)
        return;

    if (id == DataID::Video_send)
        if (sender)
            sender->setProperty("video_need", val);

    if (id == DataID::VideoOnToggle)
        m_BStart->animateClick();
    if (id == DataID::VideoXYjoy)
        goXYjoy(val.toPointF());
    if (id == DataID::VideoZjoy)
        goZjoy(val.toDouble());
}



#include <modules.moc>
