#include "motors.hpp"
#include "gui/status.hpp"
#include "gui/mslider.hpp"
#include "remote_nst_dataid.hpp"
#include "mobile/remote_mobile.hpp"

#include <qmath.h>
#include <QSettings>
#include <QCheckBox>
#include <QGroupBox>
#include <QTimer>
#include <QHBoxLayout>
#include <QFormLayout>
#include <QApplication>
#include <QToolButton>
#include <QMenu>


//see also MotorRoles enum
static const char *axes_names[] = { "", QT_TR_NOOP("X Head"), QT_TR_NOOP("Y Head"), QT_TR_NOOP("Z Landing"),
                                    QT_TR_NOOP("X Stage"), QT_TR_NOOP("Y Stage"), QT_TR_NOOP("Z Focus"),
                                    QT_TR_NOOP("X3"), QT_TR_NOOP("Y3"), QT_TR_NOOP("Z3"),
                                    QT_TR_NOOP("unused") };
static const int axes_names_count = 11;


void Motors::go(int num, int dir, double speed, double steps)
{
    if (num == -1)
        num = active_motor;

    if (steps < 0) {
        steps *= -1;
        dir *= -1;
    }

    BYTE direction = smStop;
    if (dir != 0) {
        bool morient = num == 250 || num == 0 ? orient123 : orient[num-1];
        if (dir < 0) morient = !morient;
        if (!morient) direction = steps == 0 ? smDownToLimit : smDown;
        else direction = steps == 0 ? smUpToLimit : smUp;
    }

    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    double max_speed = settings_double(settings, "max_speed/" + QString::number(num), 0);
    if (max_speed > 0)
        speed = qMin(speed, max_speed);

    double k = Estepsrev->value() / Epitch->value() / 1e3;
    set_step_motors_leveling(num, direction, steps * k, speed * k / 16); //to mkm
}

void Motors::goUpE(int num, double speed)
{
    if (num == -1) {
        if (last_command == 2) return goStop();
        last_command = 2;
    }
    if (speed == 0) speed = Espeed->value();
    go(num, 1, speed, 0);
}
void Motors::goUp(int num, double speed, double dist)
{
    if (num == -1) {
        if (last_command == 1) return goStop();
        last_command = 1;
    }
    if (speed == 0) speed = Espeed->value();
    if (dist == 0) dist = Edist->value();
    go(num, 1, speed, dist);
}
void Motors::goDownE(int num, double speed)
{
    if (num == -1) {
        if (last_command == -2) return goStop();
        last_command = -2;
    }
    if (speed == 0) speed = Espeed->value();
    go(num, -1, speed, 0);
}
void Motors::goDown(int num, double speed, double dist)
{
    if (num == -1) {
        if (last_command == -1) return goStop();
        last_command = -1;
    }
    if (speed == 0) speed = Espeed->value();
    if (dist == 0) dist = Edist->value();
    go(num, -1, speed, dist);
}
void Motors::goStop(int num)
{
    last_command = 0;
    go(num, 0, 0, 0);
}

void Motors::goShuttleInt(int speedidx, int num)
{
    goShuttle(speedidx, num);
}
void Motors::goShuttle(double speedidx, int num)
{
    if (num == -1)
        num = active_motor;

    double speed = getShuttleSpeed(speedidx);
    go(num, speedidx == 0 ? 0 : (speedidx > 0 ? 1 : -1), speed, 0);

    int id = num; if (id == 250) id = 0;
    if (id >= 0 && id <= 4)
        MobileServer::sendData(DataID::Motor_123 + id, qBound(-5, qRound(speedidx), 5));
}
double Motors::getShuttleSpeed(double speedidx)
{
    double p = qFabs(speedidx);
    int spd0idx = qBound(0, qFloor(p), 5+1-2);
    p -= spd0idx;
    double speed = EspeedShuttle[spd0idx]->value() * (1-p) + EspeedShuttle[spd0idx+1]->value() * p;
    return speed;
}

void Motors::goShuttleStep(int dir, int num)
{
    double speed = EspeedShuttle[0+1]->value();
    double step = EdistShuttle->value();
    if (dir > 0)
        goUp(num, speed, step);
    else
        goDown(num, speed, step);
}


void Motors::SetCurrent(int num)
{
    if (sender())
    for (int i = 0; i < 16; i++)
        if (sender() == PBselector[i]) {
            num = i;
            break;
        }

    if (last_command) goStop();

    PBselector[active_motor-1]->setChecked(false);
    active_motor = num+1;
    last_command = 0;
    PBselector[active_motor-1]->setChecked(true);

    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    QString stridx = QString::number(active_motor);
    CBdir->setChecked(orient[num]);
    CBinv->setChecked(settings.value("invert_ends/" + stridx, false).toBool());
    LEname->setText(settings.value("name/" + stridx).toString());
    int idx = settings_double(settings, "remap_motor/" + stridx, 0);
    if (idx <= 0) idx = active_motor;
    Eidx->setValue(idx);
    EBLdist->setValue(settings_double(settings, "backlash_dist/" + stridx, 0));
    EMaxSpeed->setValue(settings_double(settings, "max_speed/" + stridx, 0));
}

void Motors::save_current()
{
    if (blocksave) return;
    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    QString stridx = QString::number(active_motor);
    settings.setValue("name/" + stridx, LEname->text());
    settings.setValue("invert_dir/" + stridx, !(orient[active_motor-1] = CBdir->isChecked()));
    settings.setValue("invert_ends/" + stridx, CBinv->isChecked());
    settings.setValue("remap_motor/" + stridx, Eidx->value());
    settings.setValue("backlash_dist/" + stridx, EBLdist->value());
    settings.setValue("max_speed/" + stridx, EMaxSpeed->value());
    PBselector[active_motor-1]->setText(QString::number(active_motor) + (!LEname->text().isEmpty() ? ": " + LEname->text() : QString()));
}

void Motors::save_params()
{
    if (blocksave) return;
    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    settings.setValue("csb_motors_count", Ecsbmotors->value());

    settings.setValue("mv_dist", Edist->value());
    settings.setValue("mv_speed", Espeed->value());
    settings.setValue("pitch", Epitch->value());
    settings.setValue("steps_rev", Estepsrev->value());
    settings.setValue("backlash_speed", EBLspeed->value());

    settings.setValue("motors_count", Enum->value());
    for (int i = 0; i < 16; i++)
        PBselector[i]->setVisible(i < Enum->value());

    {
        QSettings settings("Param/shuttle.conf", QSettings::IniFormat);
        settings.setValue("motor0/steps2", EdistShuttle->value());
        for (int i = 0; i < 5; i++)
            settings.setValue("motor0/scale2" + QString::number(i), EspeedShuttle[i+1]->value());
    }
}

void Motors::init()
{
    motorRst = 0; //wait for rst
    blocksave = true;
    last_command = 0;

    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    Espeed->setValue(settings.value("mv_speed", 1000).toDouble());
    Epitch->setValue(settings.value("pitch", 0.25).toDouble());
    Estepsrev->setValue(settings.value("steps_rev", 3200).toDouble());
    double k = Estepsrev->value() / Epitch->value();


    int csb_addr = settings_double(settings, "csb_motors_addr", -1);
    if (csb_addr > 0) {
        csb_motor_addr = csb_addr;
        updateCsbMotorAddr = true;
    }

    Ecsbmotors->setValue(settings_double(settings, "csb_motors_count", -1));
    if (Ecsbmotors->value() >= 0) {
        csb_motor_count = Ecsbmotors->value();
        updateCsbMotorCount = true;
    }

    motorRst = 1; //do rst

    const int maxidx = qMax<int>(Enum->value(), Ecsbmotors->value());

    for (int i = 0; i < (maxidx > 8 ? 16 : 8); i++) {
        orient[i] = !settings.value("invert_dir/" + QString::number(i+1), true).toBool();

        int remap = settings_double(settings, "remap_motor/" + QString::number(i+1), 0);
        if (remap < 0 || remap == i+1) remap = 0;
        if (remap <= maxidx && i >= maxidx) //prevent active config overwrite from unconfigured motors
            remap = 0; //TODO check for remap conflicts
        if (remap > 0 || (remap == 0 && i != 3)) { //send resets too, except for 4th motor - will break default remap
            remap_motor_to[i] = remap; //1-base
            updateRemapMotor[i] = true; //idx 0-base
        }

        set_raw_motors_inv_end(i, settings.value("invert_ends/" + QString::number(i+1), false).toBool());

        motor_bl_steps[i] = settings_double(settings, "backlash_dist/" + QString::number(i+1), 0) * k;
        send_bl_steps[i] = true;

        Motor[i] = smStopped;
        Motor_on[i] = false;
    }
    {
        QList<int> l = Motors::getRoleMotors(Motors::RoleZ1Land);
        if (settings.contains("invert_dir/123") && !l.size()) //legacy compat
            orient123 = !settings.value("invert_dir/123", true).toBool();
        else if (settings.value("land_mode", 4).toInt() == 1 && !l.size()) //legacy compat //TODO SPMTunning::getLandingMode()
            orient123 = orient[2];
        else {
            orient123 = true;
            bool found = false;
            int first_motor_group_count = Ecsbmotors->value() > 0 ? Ecsbmotors->value() : 4;
            for (int i = 0; i < 3 &&
                            i < first_motor_group_count; i++)
                if (!l.size() || l.contains(i+1)) {
                    orient123 = found ? orient123 && orient[i] : orient[i];
                    found = true;
                }
        }
    }

    motor_land_safe_calib = qRound(k / 1e3 * settings_double(settings, "divider/123", 1));
    sendLandSafeCalib = true;

    EBLspeed->setValue(settings.value("backlash_speed", 10).toDouble());
    setBLspeed();

    Edist->setValue(settings.value("mv_dist", 1000).toDouble());
    {
        QSettings settings("Param/shuttle.conf", QSettings::IniFormat);
        EdistShuttle->setValue(settings.value("motor0/steps2", 1).toDouble());
        static int spd[] = {10, 30, 70, 200, 500};
        for (int i = 0; i < 5; i++)
            EspeedShuttle[i+1]->setValue(settings.value("motor0/scale2" + QString::number(i), spd[i]).toDouble());
    }

    motorStatusLimit = qMax<int>(Enum->value(), 4);
    Slot_Set_Motor_Status();
    SetCurrent(active_motor-1);
    blocksave = false;
}

void Motors::setBLspeed()
{
    double k = Estepsrev->value() / 16 / Epitch->value();
    motor_bl_speed = EBLspeed->value() * k;
    sendBLspeed = true;
}


Motors::Motors(QWidget *parent) : QWidget(parent)
{
    reqsent = false;
    blocksave = true;
    active_motor = 1;
    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    setWindowTitle(tr("Motors"));

    motorStatusLimit = 4;
    SWmotor1 = new StatusWidget1("1234");
    SWmotor1->status->set(StatusWidget::Square, Qt::gray);

    for (int i = 0; i < 16; i++) {
        Motor[i] = smStopped;
        Motor_on[i] = false;
    }

    motorSync = new QTimer(this);
    motorSync->setSingleShot(true);
    connect(motorSync, SIGNAL(timeout()), this, SLOT(Slot_Set_Motor_Status()));


    QHBoxLayout *lay = new QHBoxLayout(this);
    lay->setSizeConstraint(QLayout::SetFixedSize);

    QVBoxLayout *Lselector = new QVBoxLayout;
    lay->addLayout(Lselector);
    lay->addStretch(1);
    Lselector->addWidget(new QLabel(tr("Active motors: ")));
    Enum = new Edit(1, 16);
    Enum->setImmediateMode();
    Enum->setFocusPolicy(Qt::NoFocus);
    Enum->setValue(settings_double(settings, "motors_count", 8));
    connect(Enum, SIGNAL(newValue()), this, SLOT(save_params()));
    Lselector->addWidget(Enum);
    {
        QGroupBox *delim = new QGroupBox();
        delim->setFixedHeight(1);
        Lselector->addWidget(delim);
    }
    for (int i = 0; i < 16; i++) {
        QString stridx = QString::number(i+1);
        QString name = settings.value("name/" + stridx).toString();
        PBselector[i] = new QPushButton(stridx + (!name.isEmpty() ? ": " + name : QString()));
        PBselector[i]->setStyleSheet("QPushButton {border: 1px solid gray; border-radius: 3px;} QPushButton:checked {background-color: lightBlue; border: 1px solid blue; border-radius: 3px;} ");
        PBselector[i]->setFlat(true);
        PBselector[i]->setMaximumHeight(PBselector[i]->sizeHint().height() * 2);
        PBselector[i]->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::MinimumExpanding);
        PBselector[i]->setCheckable(true);
        connect(PBselector[i], SIGNAL(clicked()), this, SLOT(SetCurrent()));
        Lselector->addWidget(PBselector[i], 1);
        PBselector[i]->setVisible(i < Enum->value());
    }
    Lselector->addStretch(1);


    WsettingsLeft = new QWidget();
    WsettingsLeft->hide();
    showhidelist << WsettingsLeft;
    QFormLayout *slay0 = makeFormLayout(WsettingsLeft);
    slay0->setContentsMargins(0, 0, 0, 0);
    slay0->addRow(new QLabel(tr("Speed (")+QObject::tr("micro")+tr("m/s)")));
    slay0->addRow(Espeed = new Edit(1, 10000));
    slay0->addRow(new QLabel(tr("Distance (")+QObject::tr("micro")+tr("m)")));
    slay0->addRow(Edist = new Edit(1, 100000));
    Espeed->setImmediateMode();
    Edist->setImmediateMode();
    connect(Espeed, SIGNAL(newValue()), this, SLOT(save_params()));
    connect(Edist, SIGNAL(newValue()), this, SLOT(save_params()));

    {
        QGroupBox *delim = new QGroupBox();
        delim->setFixedWidth(1);
        lay->addWidget(delim);
    }

    QVBoxLayout *slayb = new QVBoxLayout;
    lay->addLayout(slayb);

    QPushButton *pb;
    pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/FastUp.png"), QString());
    connect(pb, SIGNAL(clicked()), this, SLOT(goUpE()));
    slayb->addWidget(pb);
    pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/Up.png"), QString());
    connect(pb, SIGNAL(clicked()), this, SLOT(goUp()));
    slayb->addWidget(pb);
    pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/approach/landing_done.png"), QString());
    pb->setShortcut(QKeySequence(Qt::Key_Escape));
    connect(pb, SIGNAL(clicked()), this, SLOT(goStop()));
    slayb->addWidget(pb);
    pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/Down.png"), QString());
    connect(pb, SIGNAL(clicked()), this, SLOT(goDown()));
    slayb->addWidget(pb);
    pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/FastDown.png"), QString());
    connect(pb, SIGNAL(clicked()), this, SLOT(goDownE()));
    slayb->addWidget(pb);


    MotorSlider *Mslider = new MotorSlider;
    connect(Mslider, SIGNAL(newPos(int)), this, SLOT(goShuttleInt(int)));
    lay->addWidget(Mslider);


    BShowHideSett = new QPushButton(">>");
    BShowHideSett->setToolTip(tr("Show/hide settings"));
    BShowHideSett->setFlat(true);
    BShowHideSett->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Expanding);
    BShowHideSett->setFixedWidth(BShowHideSett->sizeHint().width() / 2);
    connect(BShowHideSett, SIGNAL(clicked(bool)), this, SLOT(sett_showhide()));
    lay->addWidget(BShowHideSett);


    lay->addWidget(WsettingsLeft, 2);
    lay->addStretch(1);

    WsettingsShuttle = new QGroupBox(tr("Shuttle"));
    slay0->addRow(WsettingsShuttle);
    QFormLayout *slay2 = makeFormLayout(WsettingsShuttle);
    slay2->addRow(new QLabel(tr("Step (")+QObject::tr("micro")+tr("m)")));
    slay2->addRow(EdistShuttle = new Edit(0.1, 1000, 1));
    slay2->addRow(new QLabel(tr("Speed (")+QObject::tr("micro")+tr("m/s)")));
    EspeedShuttle[0] = new Edit(0, 1); //dummy 0 speed
    for (int i = 0; i < 5; i++) {
        slay2->addRow(QString::number(i+1), EspeedShuttle[i+1] = new Edit(1, 10000));
        connect(EspeedShuttle[i+1], SIGNAL(newValue()), this, SLOT(save_params()));
    }
    connect(EdistShuttle, SIGNAL(newValue()), this, SLOT(save_params()));


    Wsettings = new QWidget;//QGroupBox(tr("Settings"));
    Wsettings->hide();
    showhidelist << Wsettings;
    lay->addStretch(1);
    lay->addWidget(Wsettings, 2);
    QFormLayout *slay10 = makeFormLayout(Wsettings);
    slay10->setContentsMargins(0, 0, 0, 0);
    WsettGB = new QGroupBox(tr("Selected"));
    slay10->addRow(WsettGB);
    QFormLayout *slay1 = makeFormLayout(WsettGB);
    QHBoxLayout *nameLay = new QHBoxLayout;
    nameLay->setSpacing(0);
    nameLay->addWidget(LEname = new QLineEdit);
    QToolButton *nameTB = new QToolButton();
    nameTB->setText(tr("role"));
    nameTB->setToolButtonStyle(Qt::ToolButtonTextOnly);
    nameTB->setPopupMode(QToolButton::InstantPopup);
    QMenu *nameMenu = new QMenu();
    connect(nameMenu, SIGNAL(triggered(QAction*)), this, SLOT(set_name_idx(QAction*)));
    for (int i = 0; i < axes_names_count; i++)
        nameMenu->addAction(axes_names[i]);
    nameTB->setMenu(nameMenu);
    nameTB->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Ignored);
    nameLay->addWidget(nameTB);
    slay1->addRow(tr("Name"), nameLay);
    slay1->addRow(tr("HW idx"), Eidx = new Edit(1, 16)); //Eidx->setSpecialValueText("123");
    slay1->addRow(CBinv = new QCheckBox(tr("Limit switch mode")));
    slay1->addRow(CBdir = new QCheckBox(tr("Direction")));
    slay1->addRow(tr("Max speed (")+QObject::tr("micro")+tr("m/s)"), EMaxSpeed = new Edit(0, 10000));
    EMaxSpeed->setSpecialValueText(tr("any"));
    slay1->addRow(tr("Backlash(mm)"), EBLdist = new Edit(0, 999, 2)); EBLdist->setSpecialValueText(tr("none"));
    connect(LEname, SIGNAL(textEdited(QString)), this, SLOT(save_current()));
    connect(Eidx, SIGNAL(newValue()), this, SLOT(save_current()));
    connect(CBinv, SIGNAL(clicked()), this, SLOT(save_current()));
    connect(CBdir, SIGNAL(clicked()), this, SLOT(save_current()));
    connect(EMaxSpeed, SIGNAL(newValue()), this, SLOT(save_current()));
    connect(EBLdist, SIGNAL(newValue()), this, SLOT(save_current()));
    connect(Eidx, SIGNAL(newValue()), this, SLOT(init()));
    connect(CBinv, SIGNAL(clicked()), this, SLOT(init()));
    connect(CBdir, SIGNAL(clicked()), this, SLOT(init()));
    connect(EBLdist, SIGNAL(newValue()), this, SLOT(init()));

    QWidget *gSett = new QGroupBox(tr("Common"));
    slay10->addRow(gSett);
    QFormLayout *slay11 = makeFormLayout(gSett);
    Ecsbmotors = new Edit(-1, 12);
    Ecsbmotors->setSpecialValueText(tr("Default"));
    slay11->addRow(tr("External motors"), Ecsbmotors);
    slay11->addRow(tr("Thread pitch(mm)"), Epitch = new Edit(0.01, 9, 2));
    slay11->addRow(tr("Steps/revolution"), Estepsrev = new Edit(1, 99000));
    slay11->addRow(tr("Backslash speed(mm/s)"), EBLspeed = new Edit(0.1, 999, 1));
    connect(Epitch, SIGNAL(newValue()), this, SLOT(save_params()));
    connect(Estepsrev, SIGNAL(newValue()), this, SLOT(save_params()));
    connect(EBLspeed, SIGNAL(newValue()), this, SLOT(save_params()));
    connect(EBLspeed, SIGNAL(newValue()), this, SLOT(setBLspeed()));
    connect(Ecsbmotors, SIGNAL(newValue()), this, SLOT(save_params()));
    connect(Ecsbmotors, SIGNAL(newValue()), this, SLOT(init()));

    BSave = new QPushButton(tr("Save to device"));
    connect(BSave, SIGNAL(clicked(bool)), this, SLOT(save_device()));
    BSave->hide();
    slay10->addRow(BSave);


    { //minimum size expander
        QWidget *delim = new QWidget();
        delim->setFixedSize(1, WsettingsLeft->sizeHint().height());
        lay->addWidget(delim);
    }


    init();
}

void Motors::sett_showhide()
{
    if (!showhidelist.size())
        return;
    foreach (QWidget *w, showhidelist)
        w->setVisible(!w->isVisibleTo(this));
    BShowHideSett->setText(showhidelist.first()->isVisibleTo(this) ? "<<" : ">>");
}
void Motors::show_sett()
{
    if (!showhidelist.size())
        return;
    if (!showhidelist.first()->isVisibleTo(this))
        sett_showhide();
}
void Motors::addCustomSett(QWidget *w)
{
    QVBoxLayout *l = (QVBoxLayout *)layout();
    if (showhidelist.size())
        w->setVisible(showhidelist.first()->isVisibleTo(this));
    l->addWidget(w, 0, Qt::AlignTop);
    showhidelist << w;
}

void Motors::set_name_idx(QAction *a)
{
    QString new_name = a->text();
    int idx = 0;
    for (int i = 1; i < axes_names_count; i++)
        if (new_name == tr(axes_names[i]) || new_name == axes_names[i]) {
            idx = i;
            break;
        }

    if (idx == RoleZ1Land && active_motor > 3) {
        new Notification(tr("Landing motors must belong to the first 3"), Notification::Warn, 10);
    }


    LEname->setText(tr(axes_names[idx]));
    save_current();
}

QList<int> Motors::getRoleMotors(int role) //see MotorRoles enum
{
    QList<int> list;
    if (role >= axes_names_count)
        return list;

    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    for (int i = 0; i < settings_double(settings, "motors_count", 8); i++) {
        QString stridx = QString::number(i+1);
        QString name = settings.value("name/" + stridx).toString();
        if (name == tr(axes_names[role]) || name == axes_names[role])
            list << (i+1);
    }

    return list;
}

QString motorStr(QString name, bool active, bool atEnd)
{
    QString str;

    str += "<span";
    if (atEnd) {
        str += " style='color:red'";
    } else {
        if (!active)
            str += " style='color:#999999'";
        else
            str += " style='color:green'";
    }
    str += ">";
    str += name;
    str += "</span>";

    return str;
}
bool Motors::processPacket(const BYTE addr, const BYTE sub_addr, const void *data, const DWORD )
{
    if (addr != ADR_CPU)
        return false;

    if (sub_addr == CPU_STEP_MOTOR_STATUS) {
        EG_StepMotorStatus *motor_satus = (EG_StepMotorStatus *)data;
        int mnumber = motor_satus->MotorN;
        int mstatus = motor_satus->Status;

        if (mnumber >= 16)
            return true;

        Motor[mnumber - 1] = mstatus;
        if (mnumber == active_motor && mstatus == smStopped)
            last_command = 0;

        QMetaObject::invokeMethod(this, "Slot_Set_Motor_Status_handler");

        return true;
    }

    return false;
}
void Motors::Slot_Set_Motor_Status_handler()
{
    if (!motorSync->isActive())
        motorSync->start(100);
}
void Motors::Slot_Set_Motor_Status()
{
    bool atEnd[16], active[16];
    QString status;
    bool deactivated = true;
    for (int i = 0; i < 16; i++) {
        atEnd[i] = (Motor[i] == smLimitUp) || (Motor[i] == smLimitDown);
        active[i] = !atEnd[i] && ((Motor[i] == smRunningUp) || (Motor[i] == smRunningDown));
        if (i < motorStatusLimit) {
            status += motorStr(QString::number(i+1), active[i], atEnd[i]);
            deactivated = deactivated && !active[i];
        }
    }

    SWmotor1->label->setText(status);

    if (deactivated) {
       SWmotor1->status->set(StatusWidget::Square, Qt::gray);
    } else {
        bool dir = false;
        for (int i = 0; i < motorStatusLimit; i++)
            dir = dir || (active[i] ? Motor[i] == (orient[i] ? smRunningDown : smRunningUp) : false);
        SWmotor1->status->set(dir ? StatusWidget::Down : StatusWidget::Up, Qt::green);
    }

    for (int i = 0; i < 4; i++)
        if (!active[i] && Motor_on[i] != active[i])
            MobileServer::sendData(DataID::Motor_1 + i, 0);
    if ((!active[0] || !active[1] || !active[2]) && (Motor_on[0] != active[0] || Motor_on[1] != active[1] || Motor_on[2] != active[2]))
        MobileServer::sendData(DataID::Motor_123, 0);

    for (int i = 0; i < 16; i++)
        Motor_on[i] = active[i];

    for (int i = 0; i < 3; i++)
        if (Motor_on[i])
            return;

    emit m123stopped();
}


void Motors::set_raw_motors_control(unsigned char motor, unsigned short control)
{
    if (motor == 250) {
        set_raw_motors_control(1, control);
        set_raw_motors_control(2, control);
        set_raw_motors_control(3, control);
        return;
    }

    motor_cont_byte[motor] = control;
    send_control[motor] = true;
}
void Motors::set_raw_motors_inv_end(unsigned char motor, unsigned short inv)
{
    if (motor == 250) {
        set_raw_motors_inv_end(0, inv);
        set_raw_motors_inv_end(1, inv);
        set_raw_motors_inv_end(2, inv);
        return;
    }

    motor_inv_byte[motor] = inv;
    update_inv_end[motor] = true;
}
void Motors::set_step_motors_leveling(int motor, unsigned char motor_cont_byte_n, double motor_step_n, double motor_speed_n)
{
    if (motor_cont_byte_n == smStop && motor != 250) {
        set_raw_motors_control(motor, motor_cont_byte_n);
        return;
    }
    if (motor == 250)
        motor = 0;

    motor_step[motor] = motor_step_n;
    motor_speed[motor] = motor_speed_n;
    motor_cont_byte[motor] = motor_cont_byte_n;
    send_control[motor] = false;
    updMotor[motor] = true;
}

#include <i_eg3000.h>
bool Motors::sendData(const void *handle)
{
    bool wrote = false;

    if (sendReset.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(1) + REG_CONTROL, 0);
        eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(2) + REG_CONTROL, 0);
        eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(3) + REG_CONTROL, 0);
        eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(4) + REG_CONTROL, 0);
    }
    if (motorRst == 1) { //do rst
        motorRst = 2; //rdy
        if (updateCsbMotorAddr.claimMessage()) {
            wrote = true;
            eg3000_send_word_packet(handle, ADR_CPU, CPU_MOTOR_CSB_ADDR, csb_motor_addr);
        }
        if (updateCsbMotorCount.claimMessage()) {
            wrote = true;
            eg3000_send_word_packet(handle, ADR_CPU, CPU_MOTOR_CSB_COUNT, csb_motor_count);
        }
    }
    if (motorRst == 2) { //ready
        if (sendBLspeed.claimMessage()) {
            wrote = true;
            eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(1) + REG_BL_SPEED, motor_bl_speed); //head
            eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(8) + REG_BL_SPEED, motor_bl_speed); //second controller i2c-motor board
        }
        if (sendLandSafeCalib.claimMessage()) {
            wrote = true;
            eg3000_send_packet(handle, ADR_CPU, CPU_LANDING_MKM, &motor_land_safe_calib, sizeof(DWORD));
        }
        for (int i = 0; i < 16; i++) {
            if (updateRemapMotor[i].claimMessage()) {
                wrote = true;
                eg3000_send_word_packet(handle, ADR_CPU, CPU_MOTOR_REMAP, (i << 8) + remap_motor_to[i]);
            }
            if (update_inv_end[i].claimMessage()) {
                wrote = true;
                eg3000_send_word_packet(handle, ADR_STEP_MOTORS,
                                        SMD_MOTOR_ITOA(i + 1) + REG_INV_END, motor_inv_byte[i]);
            }
            if (send_bl_steps[i].claimMessage()) {
                wrote = true;
                eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(i + 1) + REG_BL_STEPS_H, (motor_bl_steps[i] >> 16) & 0xFFFF);
                eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(i + 1) + REG_BL_STEPS_L, motor_bl_steps[i] & 0xFFFF);
            }
        }
        for (int i = 0; i < 16+1; i++) {
            if (updMotor[i].claimMessage()) {
                wrote = true;
                eg3000_set_motor_leveling((void *)handle, (i == 0 ? 250 : i), motor_cont_byte[i], motor_step[i], motor_speed[i]);
            }
            if (send_control[i].claimMessage()) {
                wrote = true;
                eg3000_send_word_packet(handle, ADR_STEP_MOTORS,
                                        SMD_MOTOR_ITOA(i) + REG_CONTROL, motor_cont_byte[i]);
            }
        }
        if (sendCustom.claimMessage()) {
            wrote = true;
            eg3000_send_word_packet(handle, ADR_STEP_MOTORS, rawmotoraddr, rawmotordata);
        }
    }

    return wrote;
}
void Motors::resetWriter()
{
    motorRst = 0;
    reqsent = false;
    sendReset = true;

    sendBLspeed = false;
    sendLandSafeCalib = false;
    updateCsbMotorAddr = updateCsbMotorCount = false;
    updMotor[0] = false;
    send_control[0] = false;
    for (int i = 0; i < 16; i++) {
        update_inv_end[i] = false;
        send_bl_steps[i] = false;
        updateRemapMotor[i] = false;
        updMotor[i+1] = false;
        send_control[i+1] = false;
    }
    sendCustom = false;

    QMetaObject::invokeMethod(this, "init");
}

void Motors::capBits(int n, long bits)
{
    if (n == 5+1) {
        int idx = (bits & 0x0F000000) >> 24;
        //qDebug() << idx << (void *)bits << reqsent;
        if (idx == 0) {
            int count = bits & 0x0F;
            BSave->setVisible(count > 0);
            if (count == 0) {
                reqsent = false;
                sendCustom = false;
                return;
            }

            QSettings settings("Param/motor.conf", QSettings::IniFormat);
            settings.setValue("csb_motors_addr", int(bits & 0xF0) >> 4);

            if (Enum->value() < count)
                Enum->setValue(count);
            Ecsbmotors->setValue(count);
            Ecsbmotors->setNewValue();
        } else {
            if (idx == 1) { //reset conflicting remap, to prevent unrelated ends overwrite
                QSettings settings("Param/motor.conf", QSettings::IniFormat);
                for (int i = qMax<int>(0, Ecsbmotors->value()); i < 16; i++) {
                    QString stridx = QString::number(i+1);
                    int idx = settings_double(settings, "remap_motor/" + stridx, 0);
                    if (idx <= 0) idx = i+1;
                    if (idx <= Ecsbmotors->value())
                        settings.remove("remap_motor/" + stridx);
                }
            }
            if (idx > Ecsbmotors->value())
                return;

            SetCurrent(idx-1);
            CBdir->setChecked(bits & 0x10);
            CBinv->setChecked(bits & 0x20);
            int remap = bits & 0xF;
            if (remap == 0) remap = idx;
            Eidx->setValue(remap);
            EBLdist->setValue(float((bits & 0xFF00) >> 8) / 100); //to mm
            int axis_id = (bits & 0x0F0000) >> 16;
            if (axis_id >= axes_names_count) axis_id = 0;
            LEname->setText(tr(axes_names[axis_id]));
            Eidx->setNewValue();
            //EBLdist->setNewValue();

            SetCurrent(0);
        }

        if (!reqsent) {
            reqsent = true;
            while (sendCustom) {
                qApp->processEvents();
                SleepExport::msleep2(3);
            }

            SetCurrent(0);
            if (Eidx->value() > Ecsbmotors->value()) {
                Eidx->setValue(0+1);
                Eidx->setNewValue();
            }

            rawmotoraddr = REG_FLASH_READ + SMD_MOTOR_ITOA(1);
            rawmotordata = 0;
            sendCustom = true;
        } else if (idx == Ecsbmotors->value())
            reqsent = false;
    }
}

void Motors::save_device()
{
    WORD data;
    int count = (Ecsbmotors->value() > 0 ? qMin(Enum->value(), Ecsbmotors->value()) : Enum->value());
    for (int i = 0; i < count; i++) {
        SetCurrent(i);

        while (sendCustom) {
            qApp->processEvents();
            SleepExport::msleep2(3);
        }

        rawmotoraddr = REG_FLASH_IDX + SMD_MOTOR_ITOA(1);
        rawmotordata = (i+1)*2;
        sendCustom = true;
        while (sendCustom) {
            qApp->processEvents();
            SleepExport::msleep2(3);
        }

        data = Eidx->value() + (BYTE(EBLdist->value() * 100) << 8) +
               (CBdir->isChecked() ? 0x10 : 0) + (CBinv->isChecked() ? 0x20 : 0);
        rawmotoraddr = REG_FLASH_SET + SMD_MOTOR_ITOA(1);
        rawmotordata = data;
        sendCustom = true;
        while (sendCustom) {
            qApp->processEvents();
            SleepExport::msleep2(3);
        }

        rawmotoraddr = REG_FLASH_IDX + SMD_MOTOR_ITOA(1);
        rawmotordata = (i+1)*2 + 1;
        sendCustom = true;
        while (sendCustom) {
            qApp->processEvents();
            SleepExport::msleep2(3);
        }

        int axis_id = 0;
        for (int i = 1; i < axes_names_count; i++)
            if (LEname->text() == tr(axes_names[i]) || LEname->text() == axes_names[i]) {
                axis_id = i;
                break;
            }
        data = ((i + 1) << 8) + axis_id; //high-word
        rawmotoraddr = REG_FLASH_SET + SMD_MOTOR_ITOA(1);
        rawmotordata = data;
        sendCustom = true;
    }
    SetCurrent(0);

    while (sendCustom) {
        qApp->processEvents();
        SleepExport::msleep2(3);
    }

    rawmotoraddr = REG_FLASH_IDX + SMD_MOTOR_ITOA(1);
    rawmotordata = 0;
    sendCustom = true;
    while (sendCustom) {
        qApp->processEvents();
        SleepExport::msleep2(3);
    }

    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    data = count + (qMax(0, (int)settings_double(settings, "csb_motors_addr", 5) << 4));
    rawmotoraddr = REG_FLASH_SET + SMD_MOTOR_ITOA(1);
    rawmotordata = data;
    sendCustom = true;
    while (sendCustom) {
        qApp->processEvents();
        SleepExport::msleep2(3);
    }

    rawmotoraddr = REG_FLASH_WRITE + SMD_MOTOR_ITOA(1);
    rawmotordata = (count+1)*2;
    sendCustom = true;
}
