#include "oscilloscope.h"

#include <QSettings>
#include <QMessageBox>
#include <QCloseEvent>
#include <QHBoxLayout>
#include <QStackedLayout>
#include <QCheckBox>
#include <QTimer>
#include "remote_nst_dataid.hpp"
#include "gui/status.hpp"
#include "autocalib.hpp"
#include "xyzcalibrations.hpp"

#include "hardware/EG3000.h"
#include "spmtunning.h"


QVector<QPointer<QComboBox> > cb_channels_list;
void changeCBname(int srcidx, QString name)
{
    for (int i = 0; i < cb_channels_list.size(); i++)
        if (QComboBox *box = cb_channels_list.at(i))
            for (int i = 0; i < box->count(); i++)
                if (box->itemData(i) == srcidx) {
                    box->setItemText(i, name);
                    break;
                }
}

void fillCBchannels_(QComboBox *box, bool want_xyz)
{
    cb_channels_list << box;

    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    bool spm = settings.value("display/spm", true).toBool();
    bool head = settings.value("display/oscilloscope/head", true).toBool();
    bool stage = settings.value("display/oscilloscope/stage", true).toBool();
    bool z3 = settings.value("display/oscilloscope/z3", false).toBool();
    bool add = show_additional_gui();
    bool adv = show_advanced_gui();
    bool hasADDACS1234 = false;
    {
        QSettings settings("Param/controller.conf", QSettings::IniFormat);
        hasADDACS1234 = !settings.value("use_eg4000", true).toBool();
    }

    box->addItem(QObject::tr("OFF"), srcNone);
    if (spm) {
        box->addItem(QObject::tr("Height"), srcHeight);
        box->addItem(QObject::tr("DFL"), srcDFL);
        box->addItem(QObject::tr("LFL")/*+"/"+QObject::tr("ITip")*/, srcLFL);
        box->addItem(QObject::tr("MAG"), srcMag1);
        box->addItem(QObject::tr("Phase"), srcPhase1);
    }

    if (want_xyz || (add && adv)) {
        if (head) {
            QString scanner;
            if (stage || z3)
                scanner = " " + QObject::tr("head");
            box->addItem(QObject::tr("Ux") + scanner, srcDACX);
            box->addItem(QObject::tr("Uy") + scanner, srcDACY);
            box->addItem(QObject::tr("Uz") + scanner, srcDACZ);
        }
        if (stage) {
            QString scanner;
            if (head || z3)
                scanner = " " + QObject::tr("stage");
            box->addItem(QObject::tr("Ux") + scanner, srcDACX2);
            box->addItem(QObject::tr("Uy") + scanner, srcDACY2);
            box->addItem(QObject::tr("Uz") + scanner, srcDACZ2);
        }
        if (z3) {
            QString scanner;
            if (head || stage)
                scanner = " " + QObject::tr("z3");
            box->addItem(QObject::tr("Uz") + scanner, srcDACZ3);
        }

        if (head) {
            QString scanner;
            if (stage || z3)
                scanner = " " + QObject::tr("head");
            box->addItem(QObject::tr("X Sensor") + scanner, srcCX);
            box->addItem(QObject::tr("Y Sensor") + scanner, srcCY);
            box->addItem(QObject::tr("Z Sensor") + scanner, srcCZ);
        }
        if (stage) {
            QString scanner;
            if (head || z3)
                scanner = " " + QObject::tr("stage");
            box->addItem(QObject::tr("X Sensor") + scanner, srcCX2);
            box->addItem(QObject::tr("Y Sensor") + scanner, srcCY2);
            box->addItem(QObject::tr("Z Sensor") + scanner, srcCZ2);
        }
        if (z3) {
            QString scanner;
            if (head || stage)
                scanner = " " + QObject::tr("z3");
            box->addItem(QObject::tr("Z Sensor") + scanner, srcCZ3);
        }
    } else {
        if (head) {
            QString scanner;
            if (stage)
                scanner = " " + QObject::tr("head");
            box->addItem(QObject::tr("Z Sensor") + scanner, srcCZ);
        }
        if (stage) {
            QString scanner;
            if (head)
                scanner = " " + QObject::tr("stage");
            box->addItem(QObject::tr("Z Sensor") + scanner, srcCZ2);
        }
    }

    if (add && (adv || !spm)) {
        box->addItem(QObject::tr("ADC 1"), srcADC1);
        box->addItem(QObject::tr("ADC 2"), srcADC2);
        if (spm) box->addItem(QObject::tr("ADC 3"), srcADC3);
    }
    if (hasADDACS1234) {
        box->addItem(QObject::tr("ADC 4/PMT"), srcADC4);
        box->addItem(QObject::tr("ADC 5"), srcADC5);
    }

    {
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    if (!settings.value("use_pmt", false).toBool()) {
        box->addItem(QObject::tr("ADC AUX 1"), srcADCHR1);
        box->addItem(QObject::tr("ADC AUX 2"), srcADCHR2);
    } else {
        box->addItem(QObject::tr("PMT 1(INT)"), srcADCHR1);
        box->addItem(QObject::tr("PMT 2(EXT)"), srcADCHR2);
    }
    }

    if (spm && add) {
        box->addItem(QObject::tr("Tip V"), srcVTip);
        box->addItem(QObject::tr("MAG(LFL)"), srcMag2);
    }
    if (spm && add && adv) {
        box->addItem(QObject::tr("DInt"), srcIntensity);
        box->addItem(QObject::tr("MAG Sin"), srcMagSin1);
        box->addItem(QObject::tr("MAG Cos"), srcMagCos1);
    }
}

void Oscilloscope::device_status(int code)
{
    QString text;
    int type = Notification::Warn;

    switch (code) {
    default:
        return;
    case 0:
        text = tr("Controller has no firmware.") + " " + tr("Program it from oscilloscope window.");
        type = Notification::Bad;
        break;
    case 1:
        text = tr("Controller load timeout.") + " " + tr("Reboot and program it.");
        type = Notification::Bad;
        break;
    case 2:
        text = tr("Waiting for device to program.");
        break;
    case 3:
        text = tr("Programming. Please wait. Do not remove power from the device.");
        break;
    case 4:
        text = tr("Programming succesfully completed.");
        type = Notification::Good;
        break;
    };

    destructStatus(device_status_notice);
    device_status_notice = new Notification(text, (Notification::Tint)type, 60);
    doBeep();
}

Oscilloscope::Oscilloscope(QWidget *parent) : QWidget(parent)
{
    for (int i = 0; i < 4; i++) {
        k_m_nm[i] = 1;
        osc_data_idx[i] = srcNone;
    }
    osc_on = false;
    last_osc_ch = 0;

    refreshRateLimit.invalidate();

    time_dilation = 1;
    oscRate = 4000;
    oscMaxLength = 10 * oscRate;
    current_t_count = 0;
    last_updated_idx = -1;
    OscLineLength = 256;

    connect(this, SIGNAL(newOscData_(QByteArray)), this, SLOT(newOscData(QByteArray)));

    plot::ResizeOwner(curveData, 4);
    curveData[0]->Pen.setColor(Qt::red);
    curveData[1]->Pen.setColor(QColor(0, 170, 0)); //Qt::green
    curveData[2]->Pen.setColor(Qt::blue);
    curveData[3]->Pen.setColor(Qt::black);
    for (int i = 0; i < curveData.size(); i++)
        curveData[i]->setXCalib(0, time_dilation / oscRate);

    acf = new AutoCalib();
    acf->osc = this;

    setWindowTitle(tr("Oscilloscope - NSpec"));
    GUICreate();
    connect(acp->BReturn, SIGNAL(clicked()), PBCalibMode, SIGNAL(clicked()));
    connect(CalibWidget, SIGNAL(sigSetpointError(int)), acf, SLOT(s_pidError(int)));

    connect(MyPl, SIGNAL(CurveColourChanged(QColor, int)), this, SLOT(on_MyPl_CurveColourChanged(QColor, int)));
    for (int i = 0; i < 4; i++)
        connect(CBChannel[i], SIGNAL(currentIndexChanged(int)), this, SLOT(change_ch()));
    connect(MyPl, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)), SLOT(AddDataToContainer_slot(plot::TCurvesDataOwner,plot::TMetaStorage)));

    MyPl->Pure2dPlot->SourceData = plot::getUnLock(curveData);
    MyPl->Refresh();

#ifdef USE_EG3K_EMULATOR
    QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
    if (settings.value("usb", false).toBool()) {
        QTimer *t = new QTimer;
        connect(t, SIGNAL(timeout()), this, SLOT(inject_test_packet()));
        t->start(64/4);
    }
#endif
}

void Oscilloscope::AddDataToContainer_slot(plot::TCurvesDataOwner src, plot::TMetaStorage attributes)
{
    attributes.insert("Name", MyPl->Pure2dPlot->Title);
    attributes.insert("Source", "Oscilloscope");

    for (int i = 0; i < 4; i++) {
        if (CBChannel[i]->currentIndex() <= 0)
            continue;
        attributes.insert("Channel " + QString::number(i), CBChannel[i]->currentText());
        attributes.insert("Average", LEdAverage[i]->text());
        attributes.insert("RMS", LEdStDev[i]->text());
    }

    emit AddDataToContainer(src, attributes);
}


void Oscilloscope::closeEvent(QCloseEvent *event)
{
    PBToNewWindow->click();
    event->ignore();
}


void Oscilloscope::GUICreate()
{
    QHBoxLayout *MainLayout = new QHBoxLayout(this);

    OscStackedWidget = new QStackedWidget(this);
    OscStackedWidget->setContentsMargins(0, 0, 0, 0);

    CalibWidget = new XYZCalibrations(this, this);
    connect(CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(change_ch()));

    MainLayout->addWidget(OscStackedWidget, 1);
    MainLayout->addWidget(CalibWidget);


    QWidget *leftwid = new QWidget(OscStackedWidget);
    QHBoxLayout *leftl = new QHBoxLayout(leftwid);
    leftl->setContentsMargins(0, 0, 0, 0);

    acp = new AutoCalibPlot(OscStackedWidget);
    acp->osc = this;

    OscStackedWidget->addWidget(leftwid);
    OscStackedWidget->addWidget(acp);

    QVBoxLayout *VBLLeft = new QVBoxLayout;
    VBLLeft->setAlignment(Qt::AlignTop);
    VBLLeft->setContentsMargins(0, 0, 0, 0);

    QVBoxLayout *VBLMid = new QVBoxLayout;
    VBLMid->setSpacing(3);
    VBLMid->setContentsMargins(0, 0, 6, 0);

    leftl->addLayout(VBLLeft);
    leftl->addLayout(VBLMid, 1);


    DacsWidget = new DacsControl(leftwid);
    connect(CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), DacsWidget, SLOT(sl_xyz(double,double,double)));
    connect(CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), DacsWidget, SLOT(sl_xyz2(double,double,double)));
    connect(CalibWidget, SIGNAL(sig_XYZ_range3(double,double,double)), DacsWidget, SLOT(sl_xyz3(double,double,double)));

    MyPl = new plot::T2dPlot;
    MyPl->SetGridVisible(true);
    MyPl->SetAutoScalingEnabled(false);
    MyPl->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(tr("time"), tr("s"));
    MyPl->Pure2dPlot->YAxisUnit.SetDescription(tr("None"));
    MyPl->Pure2dPlot->SetAllXScales(0, 10);
    MyPl->Pure2dPlot->SetAllYScales(-11, 11);
    MyPl->Pure2dPlot->SetTitle(tr("Osc curve"));
    MyPl->layout()->setContentsMargins(0, 0, 0, 0);

    VBLMid->addWidget(MyPl, 1);
    VBLMid->addWidget(DacsWidget);


    QHBoxLayout *BLayout = new QHBoxLayout;
    BLayout->setSpacing(6);
    BLayout->setAlignment(Qt::AlignLeft);

    QHBoxLayout *HBLT = new QHBoxLayout;

    VBLLeft->addLayout(BLayout);
    VBLLeft->addLayout(HBLT);


    //----------------
    PBToNewWindow = new TwoStateButton(":/Detach.png",":/Attach.png");
    PBToNewWindow->setToolTip(QObject::tr("Individual window on/off"));

    start = new TwoStateButton(":start.png",":stop.png");
    disconnect(start, SIGNAL(clicked()), 0, 0);
    connect(start, SIGNAL(pressed()), start, SLOT(change_state()));
    connect(start, SIGNAL(state_changed(bool)), this, SLOT(slot_Osc_Start_Stop(bool)));

    SBT = new Edit(1, 60);
    SBT->setImmediateMode();
    SBT->setValue(10);
    connect(SBT, SIGNAL(newValue(int)), this, SLOT(s_setT(int)));

    PBCalibMode = new TwoStateButton(":/Osc_window/CalibOf.png",":/Osc_window/CalibOn.png");
    PBCalibMode->setToolTip(tr("Show/hide calibration window"));
    PBCalibMode->hide();
    connect(PBCalibMode, SIGNAL(state_changed(bool)), SLOT(calib_plot_change(bool)));

    BToCalib = new QPushButton(makeHugeIcon(":/Osc_window/CalibOf.png"), QString());
    BToCalib->setToolTip(tr("Switch to calibrations"));
    connect(BToCalib, SIGNAL(clicked()), PBCalibMode, SIGNAL(clicked()));

    HBLT->addWidget(new QLabel(tr("T, sec")));
    HBLT->addWidget(SBT);
    BLayout->addWidget(BToCalib);
    BLayout->addWidget(PBToNewWindow);
    BLayout->addWidget(start, 1);

    CBShowRelative = new QCheckBox(tr("Relative scale"));
    VBLLeft->addWidget(CBShowRelative);
    connect(CBShowRelative, SIGNAL(toggled(bool)), SLOT(change_ch()));
    bshowrelative = false;


    QGroupBox *GrBChannel[4];
    GrBChannel[0] = new QGroupBox(tr("Channel A"));
    GrBChannel[1] = new QGroupBox(tr("Channel B"));
    GrBChannel[2] = new QGroupBox(tr("Channel C"));
    GrBChannel[3] = new QGroupBox(tr("Channel D"));
    LEdColour[0] = new TRectWidget(Qt::red);
    LEdColour[1] = new TRectWidget(QColor(0, 170, 0));
    LEdColour[2] = new TRectWidget(Qt::blue);
    LEdColour[3] = new TRectWidget(Qt::black);

    for (int i = 0; i < 4; i++) {
        CBChannel[i] = new QComboBox(GrBChannel[i]);
        fillCBchannels_(CBChannel[i]);

        LAverage[i] = new QLabel(tr("Average"));
        LStDev[i] = new QLabel(tr("RMS"));
        LAverage[i]->hide();
        LStDev[i]->hide();

        LEdAverage[i] = new QLineEdit;
        LEdAverage[i]->hide();
        LEdAverage[i]->setReadOnly(true);
        LEdAverage[i]->setFocusPolicy(Qt::NoFocus);

        LEdStDev[i] = new QLineEdit;
        LEdStDev[i]->hide();
        LEdStDev[i]->setReadOnly(true);
        LEdStDev[i]->setFocusPolicy(Qt::NoFocus);

        GrBChannel[i]->setFocusPolicy(Qt::StrongFocus);
        GrBChannel[i]->setFocusProxy(CBChannel[i]);
        CBChannel[i]->installEventFilter(this);


        connect(LEdColour[i], SIGNAL(clicked()), this, SLOT(colorClick()));

        VBLLeft->addSpacerItem(new QSpacerItem(12, 12, QSizePolicy::Maximum, QSizePolicy::Maximum));
        VBLLeft->addWidget(GrBChannel[i]);

        QVBoxLayout *VBLMainForGrB = new QVBoxLayout(GrBChannel[i]);
        VBLMainForGrB->setContentsMargins(6, 3, 6, 3);

        QVBoxLayout *VBLAverage = new QVBoxLayout;
        VBLAverage->setContentsMargins(0, 0, 0, 0);

        QVBoxLayout *VBLStDev = new QVBoxLayout;
        VBLStDev->setContentsMargins(0, 0, 0, 0);

        VBLAverage->addWidget(LAverage[i]);
        VBLAverage->addWidget(LEdAverage[i]);

        VBLStDev->addWidget(LStDev[i]);
        VBLStDev->addWidget(LEdStDev[i]);

        VBLMainForGrB->addLayout(VBLAverage);
        VBLMainForGrB->addLayout(VBLStDev);
        VBLMainForGrB->addWidget(CBChannel[i]);
        VBLMainForGrB->addWidget(LEdColour[i]);
    }

    BShowHideSett = new QPushButton(">>");
    BShowHideSett->setFlat(true);
    BShowHideSett->setToolTip(tr("Show/hide calibration panel"));
    connect(BShowHideSett, SIGNAL(clicked(bool)), this, SLOT(sett_showhide()));
    VBLLeft->addWidget(BShowHideSett);
}

void Oscilloscope::sett_showhide()
{
    CalibWidget->setVisible(!CalibWidget->isVisibleTo(this));
    BShowHideSett->setText(CalibWidget->isVisibleTo(this) ? ">>" : "<<");
}
void Oscilloscope::calib_plot_change(bool val)
{
    OscStackedWidget->setCurrentIndex(val ? 1 : 0);
}


void Oscilloscope::s_setT(int t)
{
    oscMaxLength = t * oscRate;
    MyPl->Pure2dPlot->SetAllXScales(0, t);
    MyPl->Refresh();
}


void Oscilloscope::inject_test_packet()
{
#ifdef USE_EG3K_EMULATOR
    if (!osc_on)
        return;

    int idx = last_updated_idx;
    do idx++;
    while (idx < 4 && !osc_data_idx[idx]);
    if (idx > 3)
        idx = 0;

    EG_OscData Osc_data;
    Osc_data.Channel = idx;
    Osc_data.Size = 256;
    Osc_data.Source = osc_data_idx[idx];
    for (int i = 0; i < 256; i++)
        Osc_data.Data[i] = 30000 + idx * 10000
                + rand() % (5 + i);

    processPacket(ADR_CPU, CPU_OSCDATA, &Osc_data, 0);
#endif
}

bool Oscilloscope::processPacket(const BYTE addr, const BYTE sub_addr, const void *data, const DWORD )
{
    if (addr != ADR_CPU)
        return false;

    if (sub_addr == CPU_TIMEDATA) {
        DWORD *buf = (DWORD *)data;
        time_dilation = qBound(0.1, double(buf[0]) / buf[1], 10.);
        return false;
    }
    if (sub_addr == CPU_OSCDATA) {
        EG_OscData *Osc_data = (EG_OscData *)data;

        int dataSize = sizeof(EG_OscData) - (OSC_LINE_SIZE - Osc_data->Size)*sizeof(WORD);
        emit newOscData_(QByteArray((char *)Osc_data, dataSize));

        return true;
    }

    return false;
}
#include <i_eg3000.h>
bool Oscilloscope::sendData(const void *handle)
{
    bool wrote = false;

    if (sendOscChannels.claimMessage()) {
        wrote = true;
        eg3000_osc_set_data((void *)handle, osc_data_idx[0], osc_data_idx[1], osc_data_idx[2], osc_data_idx[3]);
    }
    if (sendOscOnOff.claimMessage()) {
        wrote = true;
        eg3000_osc_startstop((void *)handle, osc_on);
    }
    if (doGetData.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_GET_DATA, getDataCh);
    }

    return wrote;
}
void Oscilloscope::resetWriter()
{
    sendOscChannels = true;
    sendOscOnOff = true;
    doGetData = false;

    QMetaObject::invokeMethod(this, "reset");
}

void Oscilloscope::reset()
{
    time_dilation = 1;

    change_ch();
    slot_Osc_Start_Stop(osc_on);
    MyPl->Pure2dPlot->update(); //likely to happen inside RefreshInc(), manual bailout
}

void Oscilloscope::get_curr_data(int ch)
{
    getDataCh = ch;
    doGetData = true;
}

void Oscilloscope::newOscData(QByteArray data)
{
    const EG_OscData *sdata = (const EG_OscData *)data.constData();

    if (sdata->Channel >= 4 && sdata->Channel < 8)
        lsbData = data;
    if (sdata->Channel >= 4 || !osc_data_idx[sdata->Channel])
        return;

    if (sdata->Channel <= last_updated_idx) {
        current_t_count += OscLineLength;
        if (current_t_count + OscLineLength/4 > oscMaxLength / time_dilation
                && osc_on) //prevent reset in case we want to see prev data
            current_t_count = 0;
    }
    last_updated_idx = sdata->Channel;
    if (current_t_count == 0)
        curveData[sdata->Channel]->YSeq.clear();

    if (!curveData[sdata->Channel]->YSeq.size())
        curveData[sdata->Channel]->setXCalib(double(current_t_count) / oscRate, time_dilation / oscRate);

    const EG_OscData *sdata2 = (const EG_OscData *)lsbData.constData();
    if (lsbData.isEmpty() || sdata2->Channel % 4 != sdata->Channel)
        sdata2 = 0;

    BYTE source = sdata->Source;
    if (bshowrelative)
        source = srcNone; //default common calib

    for (int i = 0; i < OscLineLength; i++)
        curveData[sdata->Channel]->AddY(k_m_nm[sdata->Channel] *
                input_calib(source, sdata->Data[i] + (sdata2 ? ((double)sdata2->Data[i]) / 0x10000 : 0)));

    lsbData.clear();

    for (int i = sdata->Channel+1; i < 4; i++)
        if (osc_data_idx[i])
            return;
    GUIRefresh();
}

void Oscilloscope::GUIRefresh()
{
    if (refreshRateLimit.isValid() && refreshRateLimit.elapsed() < 40 + 4 * oscMaxLength / oscRate)
        return;

    refreshRateLimit.start();

    MyPl->Pure2dPlot->RefreshInc();

    for (int i = 0; i < 4 && i < curveData.size(); i++) {
        if (!osc_data_idx[i]) continue;

        plot::TSIUnit tunit;
        tunit.SetSIUnit(input_units(bshowrelative ? srcNone : osc_data_idx[i]));

        double vavg = avg(curveData.at(i)->YSeq) / k_m_nm[i];
        tunit.UpdateLegend(vavg);
        LEdAverage[i]->setText(QString::number(vavg / tunit.Multiplier, 'f', 3));
        LAverage[i]->setText(tr("Average, ") + tunit.Legend);

        double vrms = rms(curveData.at(i)->YSeq) / k_m_nm[i];
        tunit.UpdateLegend(vrms);
        LEdStDev[i]->setText(QString::number(vrms / tunit.Multiplier, 'f', 3));
        LStDev[i]->setText(tr("RMS, ") + tunit.Legend);
    }

    emit oscRefreshed();
}
void Oscilloscope::sl_osc_start()
{
    if (!start->on_off)
        start->change_state_man_with_sig(true);
}
void Oscilloscope::sl_osc_stop()
{
    start->change_state_man_with_sig(false);
}
void Oscilloscope::slot_Osc_Start_Stop(bool st)
{
    if (st) {
        if (MyPl->PlotControl->BMarkers->isChecked())
            MyPl->PlotControl->BMarkers->click();

        current_t_count = 0;
        last_updated_idx = -1;
        for (int i = 0; i < curveData.size(); i++)
            curveData[i]->YSeq.clear();

        lsbData.clear();

        MyPl->Refresh();
    }
    refreshRateLimit.invalidate();

    osc_on = st;
    sendOscOnOff = true;
}
void Oscilloscope::on_MyPl_CurveColourChanged(QColor c, int N)
{
    if (N < 0)
        for (int i = 0; i < 4; i++)
            on_MyPl_CurveColourChanged(c, i);
    else
        LEdColour[N]->SetColour(c);
}
bool Oscilloscope::eventFilter(QObject *Object, QEvent *Event)
{
    if (Event->type() == QEvent::FocusIn)
        for (int i = 0; i < 4; i++)
            if (Object == CBChannel[i]) {
                last_osc_ch = i;
                blockSignals(true);
                change_ch();
                blockSignals(false);
                break;
            }

    return QWidget::eventFilter(Object, Event);
}
void Oscilloscope::change_ch()
{
    bshowrelative = CBShowRelative->isChecked();
    for (int i = 0; i < 4; i++) {
        int ch = osc_data_idx[i] = CBChannel[i]->itemData(CBChannel[i]->currentIndex()).toInt();

        LEdAverage[i]->setVisible(ch != srcNone);
        LAverage[i]->setVisible(ch != srcNone);
        LEdStDev[i]->setVisible(ch != srcNone);
        LStDev[i]->setVisible(ch != srcNone);

        if (ch == srcNone) {
            curveData[i]->YSeq.clear();
            continue;
        }
        if (bshowrelative)
            ch = srcNone; //default common calib

        double min = input_calib(ch, 0),
               max = input_calib(ch, 65535);
        if (max < min)
            qSwap(min, max);

        plot::TSIUnit tunit;
        tunit.SetSIUnit(input_units(ch));
        tunit.UpdateLegend(min, max);
        k_m_nm[i] = 1. / tunit.Multiplier;
        min = min / tunit.Multiplier;
        max = max / tunit.Multiplier;

        if (i != last_osc_ch)
            continue;

        QString name = CBChannel[i]->currentText();
        MyPl->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(name, tunit.Legend);
        MyPl->Pure2dPlot->SetAllYScales(1.1 * qMin(-10., min), 1.1 * qMax(10., max));
    }

    sendOscChannels = true;

    MyPl->Refresh();
}

void Oscilloscope::setChA(int src)
{
    CBChannel[0]->setCurrentIndex(CBChannel[0]->findData(src));
}

void Oscilloscope::colorClick()
{
    TRectWidget *w = qobject_cast<TRectWidget *>(sender());
    if (!w) return;

    int active = MyPl->Pure2dPlot->ActiveCurve;
    int idx = -1;
    for (int i = 0; i < 4; i++)
        if (w == LEdColour[i]) {
            idx = i;
            break;
        }
    if (idx == -1)
        return;

    MyPl->Pure2dPlot->ActiveCurve = idx;
    MyPl->PlotControl->AGCurveColours->actions().last()->trigger();
    //MyPl->on_CurveColour_changed();

    MyPl->Pure2dPlot->ActiveCurve = active;
}



#include <nstslider.h>
DacsControl::DacsControl(QWidget *parent) : QWidget(parent)
{
    Xsp = Ysp = Zsp = 0;
    Xsp2 = Ysp2 = Zsp2 = 0;
    Zsp3 = 0;

    QSettings settings("Param/analog.conf", QSettings::IniFormat);
    double a1 = settings_double(settings, "dac/v1_min", -10.);
    double b1 = settings_double(settings, "dac/v1_max", 10.);
    double a2 = settings_double(settings, "dac/v2_min", -10.);
    double b2 = settings_double(settings, "dac/v2_max", 10.);
    double a3 = settings_double(settings, "dac/v3hr_min", -10.);
    double b3 = settings_double(settings, "dac/v3hr_max", 10.);
    double a4 = settings_double(settings, "dac/v4hr_min", -10.);
    double b4 = settings_double(settings, "dac/v4hr_max", 10.);

    Extra_Dac_1 = new DoubleChanger(a1, b1, 3);
    Extra_Dac_2 = new DoubleChanger(a2, b2, 3);
    {
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    if (!settings.value("use_pmt", false).toBool()) {
        Extra_Dac_3 = new DoubleChanger(a3, b3, 3);
        Extra_Dac_4 = new DoubleChanger(a4, b4, 3);
    } else {
        Extra_Dac_3 = new DoubleChanger(0, b3/10, 3);
        Extra_Dac_4 = new DoubleChanger(0, b4/10, 3);
    }
    }
    Z = new DoubleChanger(0, 50, 4);
    X = new DoubleChanger(0, 150, 4);
    Y = new DoubleChanger(0, 150, 4);
    Z2 = new DoubleChanger(0, 50, 4);
    X2 = new DoubleChanger(0, 150, 4);
    Y2 = new DoubleChanger(0, 150, 4);
    Z3 = new DoubleChanger(0, 50, 4);
    Zobj = new DoubleChanger(0, 50, 4);

    Extra_Dac_1->setProperty("id", DataID::SPM_DAC1);
    Extra_Dac_2->setProperty("id", DataID::SPM_DAC2);
    Extra_Dac_3->setProperty("id", DataID::SPM_DAC3);
    Extra_Dac_4->setProperty("id", DataID::SPM_DAC4);
    X->setProperty("id", DataID::HV_DAC_X);
    Y->setProperty("id", DataID::HV_DAC_Y);
    Z->setProperty("id", DataID::HV_DAC_Z);
    X2->setProperty("id", DataID::HV_DAC_X2);
    Y2->setProperty("id", DataID::HV_DAC_Y2);
    Z2->setProperty("id", DataID::HV_DAC_Z2);
    Z3->setProperty("id", DataID::HV_DAC_Z3);
    Zobj->setProperty("id", DataID::HV_DAC_Zobj);

    DacsChannel = new QComboBox;
    connect(DacsChannel, SIGNAL(currentIndexChanged(int)), this, SLOT(s_dacs_channel(int)));

    QHBoxLayout *layout_Dacs = new QHBoxLayout(this);
    layout_Dacs->setSpacing(6);
    layout_Dacs->setContentsMargins(0, 0, 0, 0);

    box_Ex_Dac1 = new QGroupBox(this);
    layout_Ex_Dac = new QStackedLayout(box_Ex_Dac1);

    QGroupBox* box_Ex_Dac2 = new QGroupBox(tr("DACS Channel"), this);
    QHBoxLayout *layout_Ex_Dac22 = new QHBoxLayout(box_Ex_Dac2);
    layout_Ex_Dac22->setContentsMargins(6, 3, 6, 3);

    layout_Dacs->addWidget(box_Ex_Dac1);
    layout_Dacs->addWidget(box_Ex_Dac2);

    layout_Ex_Dac22->addWidget(DacsChannel);


    QString v_unit = QObject::tr("V");
    QString um_unit = tr("micro") + QObject::tr("m");

    QSettings settings1("Param/gui.conf", QSettings::IniFormat);
    QSettings settings2("Param/controller.conf", QSettings::IniFormat);
    if (!settings2.value("use_eg4000", true).toBool()) {
        insertChannel(tr("DAC 1/PMT"), v_unit, Extra_Dac_1);
        insertChannel(tr("DAC 2"), v_unit, Extra_Dac_2);
    }
    {
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    if (!settings.value("use_pmt", false).toBool()) {
        insertChannel(tr("DAC AUX 1"), v_unit, Extra_Dac_3);
        insertChannel(tr("DAC AUX 2"), v_unit, Extra_Dac_4);
    } else {
        insertChannel(tr("PMT 1(INT)"), v_unit, Extra_Dac_3);
        insertChannel(tr("PMT 2(EXT)"), v_unit, Extra_Dac_4);
    }
    }
    if (settings1.value("display/oscilloscope/head", true).toBool()) {
        insertChannel(tr("X head"), um_unit, X);
        insertChannel(tr("Y head"), um_unit, Y);
        insertChannel(tr("Z head"), um_unit, Z);
    }
    if (settings1.value("display/oscilloscope/stage", true).toBool()) {
        insertChannel(tr("X stage"), um_unit, X2);
        insertChannel(tr("Y stage"), um_unit, Y2);
        insertChannel(tr("Z stage"), um_unit, Z2);
    }
    if (settings1.value("display/oscilloscope/z3", false).toBool())
        insertChannel(tr("Z3"), um_unit, Z3);

    DacsChannel->setCurrentIndex(0);


    connect(Extra_Dac_1, SIGNAL(newValue(double)), this, SLOT(set_Ex_dac1(double)));
    connect(Extra_Dac_2, SIGNAL(newValue(double)), this, SLOT(set_Ex_dac2(double)));
    connect(Extra_Dac_3, SIGNAL(newValue(double)), this, SLOT(set_Ex_dac3(double)));
    connect(Extra_Dac_4, SIGNAL(newValue(double)), this, SLOT(set_Ex_dac4(double)));

    connect(X, SIGNAL(newValue(double)), this, SLOT(Go_to_X(double)));
    connect(Y, SIGNAL(newValue(double)), this, SLOT(Go_to_Y(double)));
    connect(Z, SIGNAL(newValue(double)), this, SLOT(Go_to_Z(double)));
    connect(X2, SIGNAL(newValue(double)), this, SLOT(Go_to_X2(double)));
    connect(Y2, SIGNAL(newValue(double)), this, SLOT(Go_to_Y2(double)));
    connect(Z2, SIGNAL(newValue(double)), this, SLOT(Go_to_Z2(double)));
    connect(Z3, SIGNAL(newValue(double)), this, SLOT(Go_to_Z3(double)));

    connect(this, SIGNAL(xUpdate(double)), X, SLOT(setValue(double)));
    connect(this, SIGNAL(x2Update(double)), X2, SLOT(setValue(double)));
    connect(this, SIGNAL(yUpdate(double)), Y, SLOT(setValue(double)));
    connect(this, SIGNAL(y2Update(double)), Y2, SLOT(setValue(double)));
    connect(this, SIGNAL(zUpdate(double)), Z, SLOT(setValue(double)));
    connect(this, SIGNAL(z2Update(double)), Z2, SLOT(setValue(double)));
    connect(this, SIGNAL(z3Update(double)), Z3, SLOT(setValue(double)));

    //zobj target
    bool useZ3 = settings1.value("display/oscilloscope/z3", false).toBool();
    DoubleChanger *zobj = useZ3 ? Z3 : Z2;

    linkDoubleChangers(zobj, Zobj);
    connect(zobj->slider, SIGNAL(valueChanged(double)), Zobj, SLOT(setValue(double))); //sync value change on setValue
}

void DacsControl::insertChannel(QString ch_name, QString unit, DoubleChanger *dc, int idx)
{
    if (idx < 0)
        idx = dacs.size();

    dacs.insert(idx, dc);
    units.insert(idx, unit);

    dc->setContentsMargins(6, 3, 6, 3);
    layout_Ex_Dac->addWidget(dc);

    DacsChannel->insertItem(idx, ch_name, QVariant::fromValue((void *)dc));
}

void DacsControl::sl_xyz(double x, double y, double z)
{
    X->setMaximum(x);
    Y->setMaximum(y);
    Z->setMaximum(z);
}
void DacsControl::sl_xyz2(double x, double y, double z)
{
    X2->setMaximum(x);
    Y2->setMaximum(y);
    Z2->setMaximum(z);
}
void DacsControl::sl_xyz3(double, double, double z)
{
    Z3->setMaximum(z);
}

void DacsControl::go_to_center(int scanner)
{
    if (scanner == 0) {
        X->setValue(X->edit->maximum() / 2);
        Y->setValue(Y->edit->maximum() / 2);
        Z->setValue(Z->edit->maximum() / 2);
        X->setNewValue();
        Y->setNewValue();
        Z->setNewValue();
    }
    if (scanner == 1) {
        X2->setValue(X2->edit->maximum() / 2);
        Y2->setValue(Y2->edit->maximum() / 2);
        Z2->setValue(Z2->edit->maximum() / 2);
        X2->setNewValue();
        Y2->setNewValue();
        Z2->setNewValue();
    }
    if (scanner == 2) {
        Z3->setValue(Z3->edit->maximum() / 2);
        Z3->setNewValue();
    }
}


DoubleChanger *DacsControl::get_dacs_channel(QComboBox *CB)
{
    DoubleChanger *dc = (DoubleChanger *)CB->itemData(CB->currentIndex()).value<void *>();
    return dc;
}
void DacsControl::fill_dacs_channel_CB(QComboBox *CB)
{
    for (int i = 0; i < DacsChannel->count(); i++)
        CB->addItem(DacsChannel->itemText(i), DacsChannel->itemData(i));
}
void DacsControl::s_dacs_channel(int ch)
{
    DoubleChanger *dc = (DoubleChanger *)DacsChannel->itemData(ch).value<void *>();
    ch = dacs.indexOf(dc);
    if (ch < 0)
        return;

    layout_Ex_Dac->setCurrentWidget(dc);
    box_Ex_Dac1->setTitle(DacsChannel->currentText() + ", " + units.at(ch));
}


void DacsControl::Go_to_XY(double Xsp_n, double Ysp_n)
{
    Xsp = Xsp_n;
    Ysp = Ysp_n;

    emit xUpdate(Xsp);
    emit yUpdate(Ysp);
    sendSP[0] = true;
}
void DacsControl::Go_to_XY2(double Xsp_n, double Ysp_n)
{
    Xsp2 = Xsp_n;
    Ysp2 = Ysp_n;

    emit x2Update(Xsp2);
    emit y2Update(Ysp2);
    sendSP[1] = true;
}
void DacsControl::Go_to_X(double Xsp_n)
{
    Xsp = Xsp_n;
    emit xUpdate(Xsp);
    sendSP[0] = true;
}
void DacsControl::Go_to_X2(double Xsp_n)
{
    Xsp2 = Xsp_n;
    emit x2Update(Xsp2);
    sendSP[1] = true;
}
void DacsControl::Go_to_Y(double Ysp_n)
{
    Ysp = Ysp_n;
    emit yUpdate(Ysp);
    sendSP[0] = true;
}
void DacsControl::Go_to_Y2(double Ysp_n)
{
    Ysp2 = Ysp_n;
    emit y2Update(Ysp2);
    sendSP[1] = true;
}
void DacsControl::Go_to_Z(double Zsp_n)
{
    Zsp = Zsp_n;
    emit zUpdate(Zsp);
    sendSP[0] = true;
}
void DacsControl::Go_to_Z2(double Zsp_n)
{
    Zsp2 = Zsp_n;
    emit z2Update(Zsp2);
    sendSP[1] = true;
}
void DacsControl::Go_to_Z3(double Zsp_n)
{
    Zsp3 = Zsp_n;
    emit z3Update(Zsp3);
    sendSP[2] = true;
}
void DacsControl::set_Ex_dac1(double value)
{
    value = analog_calib(value, "dac/v1");
    Ex_dac[0] = value;
    sendExDAC[0] = true;
}
void DacsControl::set_Ex_dac2(double value)
{
    value = analog_calib(value, "dac/v2");
    Ex_dac[1] = value;
    sendExDAC[1] = true;
}
void DacsControl::set_Ex_dac3(double value)
{
    value = analog_calib(value, "dac/v3hr");
    Ex_dac[2] = value;
    sendExDAC[2] = true;
}
void DacsControl::set_Ex_dac4(double value)
{
    value = analog_calib(value, "dac/v4hr");
    Ex_dac[3] = value;
    sendExDAC[3] = true;
}
bool DacsControl::sendData(const void *handle)
{
    bool wrote = false;

    if (sendSP[0].claimMessage()) {
        wrote = true;
        eg3000_set_point_xyz((void *)handle, 0, Xsp, Ysp, Zsp);
    }
    if (sendSP[1].claimMessage()) {
        wrote = true;
        eg3000_set_point_xyz((void *)handle, 1, Xsp2,  Ysp2, Zsp2);
    }
    if (sendSP[2].claimMessage()) {
        wrote = true;
        eg3000_set_point_xyz((void *)handle, 2, 0, 0, Zsp3);
    }

    if (sendExDAC[0].claimMessage()) {
        wrote = true;
        eg3000_set_extra_DAC1((void *)handle, Ex_dac[0]);
    }
    if (sendExDAC[1].claimMessage()) {
        wrote = true;
        eg3000_set_extra_DAC2((void *)handle, Ex_dac[1]);
    }
    if (sendExDAC[2].claimMessage()) {
        wrote = true;
        eg3000_set_extra_DAC3((void *)handle, Ex_dac[2]);
    }
    if (sendExDAC[3].claimMessage()) {
        wrote = true;
        eg3000_set_extra_DAC4((void *)handle, Ex_dac[3]);
    }

    return wrote;
}
void DacsControl::resetWriter()
{
    for (int i = 0; i < 3; i++)
        sendSP[i] = false;
    for (int i = 0; i < 4; i++)
        sendExDAC[i] = true;
}
