#include "scanplot.h"
#include "defs.h"
#include <QCloseEvent>
#include "gui/status.hpp"
#include "gui/ttabwidget.hpp"
#include "mainwindow.h"
#include "spmtunning.h"
#include "oscilloscope.h"
#include "modules.hpp"
#include "hardware/EG3000.h"

#include "flowlayout.hpp"
#include "3d/plot_3dplot.h"
#include "math/plot_MathProcessing.h"
#include <qwt_thermo.h>
#include <qwt_plot_marker.h>
#include <QCheckBox>
#include <QSettings>
#include <QTimer>
#include <QFormLayout>
#include <QApplication>
#include <QSplitter>
#include <QProgressBar>
#include <QMessageBox>


static bool is_shown = false;
extern bool skip_pid_error; //autocalib_impl.cpp
void ScanPlot::s_scanError()
{
    static bool skiperr_1 = false;
    if (is_shown || !m_BStart->on_off)
        return;

    s_stopScan();

    if (skiperr_1)
        return;
    is_shown = true;
    skip_pid_error = true;

    int b1 = QMessageBox::warning(0, tr("Controller"),
                                  tr("Scan x/y/tip PID error.") + "\n" +
                                  tr("Can't reach setpoint. Recalibrate sensors or change spm pid precision.") + "\n" +
                                  tr("Do you want to recalibrate sensors and restart scan now?"),
                                  QMessageBox::Yes, QMessageBox::No, QMessageBox::NoToAll);

    skip_pid_error = false;
    skiperr_1 = (QMessageBox::NoToAll == b1);
    if (QMessageBox::Yes != b1) {
        is_shown = false;
        return;
    }

    emit recalibrateSensors(scanner_scan);
}
void ScanPlot::recalibDone(bool ok)
{
    if (!is_shown)
        return;

    if (ok)
        m_BStart->change_state_man_with_sig(true);

    is_shown = false;
}

ScanPlot::~ScanPlot()
{
    for (int i = 0; i < 8; i++) {
        if (ScanDataPart[i])
            free(ScanDataPart[i]);
        if (ScanDataPart_lsb[i])
            free(ScanDataPart_lsb[i]);
    }

    foreach (IExternalScan *ptr, externalScans)
        delete ptr;
}
void ScanPlot::closeEvent(QCloseEvent *event)
{
    PBToNewWindow->click();
    event->ignore();
}

ScanPlot::ScanPlot(QWidget *parent): QWidget(parent)
{
    DACcurrentNum = 0;
    currExtScan = 0;
    l2k_controller = false;
    l2k_controller_v22 = false;
    for (int i = 0; i < 3; i++)
        is_optical_scanner[i] = false;
    fast_scan_correction = 0;
    fast_scan_CLZ_disabled = false;
    Autosend = false;
    sectionEnabled = true;
    scanFailed = false;
    scan_bar = 0;
    ccd_scan_bar = 0;
    scan_on_off = 0;
    ccdMode = 0;

    Xmax = Ymax = 100;
    Xmax2 = Ymax2 = 100;
    x = y = x2 = y2 = 0;

    Nx_points = Ny_points = 300;
    xLeft = yBottom = 0;
    xRight = yTop = 100.0;
    stPointX = 0;
    stPointY = 0;

    for (int i = 0; i < 4; i++) {
        scan_ch[i] = 0;
        ready[i] = true;
        nlcupd[i] = 0;
        refresh_3d_limit_timer[i].invalidate();
        refresh_limit_last_elapsed[i] = 0;
    }
    channelsActive = 1;
    Ch_activated = 0;
    for (int i = 0; i < 8; i++) {
        ScanDataPart[i] = 0;
        ScanDataPart_lsb[i] = 0;
    }

    GUICreate();


    plot::ResizeOwner(Section_data, 4);
    const Qt::GlobalColor colormap[4] = {Qt::black, Qt::red, Qt::green, Qt::blue};
    for (int i = 0; i < 4; i++)
        Section_data[i]->Pen.setColor(colormap[i]);

    dataScan_src = dataScan = new plot::T3dDataOwner;
    plot::ResizeOwner(*dataScan_src, 4);

    dataScan_X = new plot::T2dArray;
    dataScan_Xb = new plot::T2dArray;
    dataScan_Y = new plot::T2dArray;
    dataScan_Yb = new plot::T2dArray;

    AutosaveTimer = new QTimer(this);
    AutosaveTimer->setSingleShot(true);
    connect(AutosaveTimer, SIGNAL(timeout()), this, SLOT(slAutosaveData()));
    connect(AutosaveTimer, SIGNAL(timeout()), this, SLOT(sl_dacscan_advance()));

    RefreshTimer4d = new QTimer(this);
    RefreshTimer4d->setSingleShot(true);
    connect(RefreshTimer4d, SIGNAL(timeout()), this, SLOT(sl_ext_scan_update()));

    setWindowTitle(tr("Scan window - NSpec"));

    QMetaObject::connectSlotsByName(this);

    installEventFilter(this);

    connect(Plot, SIGNAL(AddDataToContainer(plot::T3dDataOwner,plot::TMetaStorage)), SLOT(AddDataToContainer_slot(plot::T3dDataOwner,plot::TMetaStorage)));
    connect(Plot, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)), SLOT(AddDataToContainer_slot(plot::TCurvesDataOwner,plot::TMetaStorage)));

    connect(Plot->TwoDPlot->Pure2dPlot, SIGNAL(MathUpdated()), this, SLOT(Send2dPlotData()));
    connect(Plot->ThreeDPlots.at(0), SIGNAL(MathUpdated()), this, SLOT(Send3dPlotData1()));
    connect(Plot->ThreeDPlots.at(1), SIGNAL(MathUpdated()), this, SLOT(Send3dPlotData2()));
    connect(Plot->ThreeDPlots.at(2), SIGNAL(MathUpdated()), this, SLOT(Send3dPlotData3()));
    connect(Plot->ThreeDPlots.at(3), SIGNAL(MathUpdated()), this, SLOT(Send3dPlotData4()));

    {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        bool head_on = settings.value("display/oscilloscope/head", true).toBool();
        bool stage_on = settings.value("display/oscilloscope/stage", true).toBool();
        if (!head_on || !stage_on) {
            CBScanner->setCurrentIndex(head_on ? 0 : 1);
            while (scanControlWidgets.contains(CBScanner))
                scanControlWidgets.remove(scanControlWidgets.indexOf(CBScanner));
            CBScanner->hide();
            LScanner->hide();
        }
        CBScanner->setEnabled(head_on && stage_on);
    }

    if (!show_additional_gui() && !show_advanced_gui()) {
        connect(CBStartPos, SIGNAL(currentIndexChanged(int)), this, SLOT(sl_CB_filters_update()));
        for (int i = 0; i < 4; i++)
            connect(CBChannel[i], SIGNAL(currentIndexChanged(int)), this, SLOT(sl_CB_filters_update()));
    }

    CBChannel[0]->setCurrentIndex(1);
    CBSec[0]->setChecked(true);

    updateDirection();

#ifdef USE_EG3K_EMULATOR
    QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
    if (settings.value("usb", false).toBool()) {
        QTimer *t = new QTimer;
        connect(t, SIGNAL(timeout()), this, SLOT(injectTestData()));
        t->start(10);
    }
#endif
}


void ScanPlot::restoreAttributes(plot::TMetaStorage attrib)
{
    if (attrib.contains("Velocity, Hz")) {
        DCScanRate->setValue(attrib.value("Velocity, Hz").toDouble());
        DCScanRate->setNewValue();
    }
    if (attrib.contains("Lift height, nm")) {
        LEd_LFmodeHeight->setValue(attrib.value("Lift height, nm").toDouble());
        LEd_LFmodeHeight->setNewValue();
    }
}

void ScanPlot::FillDataAttributes2()
{
    scan_attributes.insert("Lift mode", LiftMode->on_off);
    if (LiftMode->on_off)
        scan_attributes.insert("Lift height, nm", LEd_LFmodeHeight->text());

    scan_attributes.insert("Velocity, Hz", DCScanRate->value());

    if (is_dacscan()) {
        I_MODULE *oscm = ModuleManager::instance->find1("oscilloscope");
        Oscilloscope *osc = (Oscilloscope *)oscm->basecast();
        scan_attributes.insert("DAC(Z)scan channel", CBDACscanCh->currentText());
        scan_attributes.insert("DAC(Z)scan value", osc->DacsWidget->get_dacs_channel(CBDACscanCh)->value());
    }

    if (I_MODULE *spmm = ModuleManager::instance->find1("spm"))
        ((SPMTunning *)spmm->basecast())->appendAttrib(scan_attributes);
}
void ScanPlot::FillDataAttributes()
{
    QVariant gid_prev = scan_attributes.value("GroupId");

    scan_attributes.clear();

    QString date = TContainer::getDateHash();
    scan_attributes.insert("Date", date);

    if (is_dacscan() && DACcurrentNum > 0)
        scan_attributes.insert("GroupId", gid_prev);
    else {
        quint16 gid = qChecksum(date.toStdString().data(), date.length());
        scan_attributes.insert("GroupId", QString::number(gid, 16).toUpper());
    }

    scan_attributes.insert("Source", "Scan - " + CBScanner->currentText());

    scan_attributes.insert("Ch A", CBChannel[0]->currentText());
    scan_attributes.insert("Ch B", CBChannel[1]->currentText());
    scan_attributes.insert("Ch C", CBChannel[2]->currentText());
    scan_attributes.insert("Ch D", CBChannel[3]->currentText());

    if (LiftMode->on_off) {
        scan_attributes.insert("Ch A lift", LmSwitch[0]->isChecked());
        scan_attributes.insert("Ch B lift", LmSwitch[1]->isChecked());
        scan_attributes.insert("Ch C lift", LmSwitch[2]->isChecked());
        scan_attributes.insert("Ch D lift", LmSwitch[3]->isChecked());
    }

    //tr("micro")
    scan_attributes.insert(QString("Scan width, ")+QChar(181)+QString("m"), LEdDeltaX->value());
    scan_attributes.insert(QString("Scan height, ")+QChar(181)+QString("m"), LEdDeltaY->value());

    scan_attributes.insert("Scan x points", CBPointsOX->currentText());
    scan_attributes.insert("Scan y points", CBPointsOY->currentText());
    scan_attributes.insert("Scan direction", scan_dir_st_point);

    I_MODULE *oscm = ModuleManager::instance->find1("oscilloscope");
    if (!oscm)
        return;
    Oscilloscope *osc = (Oscilloscope *)oscm->basecast();
    if (CBScanner->currentIndex() == scStage) {
        scan_attributes.insert(QString("Head X, ")+QChar(181)+QString("m"), osc->DacsWidget->X->value());
        scan_attributes.insert(QString("Head Y, ")+QChar(181)+QString("m"), osc->DacsWidget->Y->value());
    } else {
        scan_attributes.insert(QString("Stage X, ")+QChar(181)+QString("m"), osc->DacsWidget->X2->value());
        scan_attributes.insert(QString("Stage Y, ")+QChar(181)+QString("m"), osc->DacsWidget->Y2->value());
    }

    scan_attributes.insert("XY PID", (CBScanner->currentIndex() == scHead ? osc->CalibWidget->CL_status_xy : osc->CalibWidget->CL2_status_xy));
}

void ScanPlot::AddDataToContainer_slot(plot::TCurvesDataOwner src, plot::TMetaStorage attributes)
{
    attributes.insert("Name", tr("Scan section"));//Plot->TwoDPlot->Pure2dPlot->Title);

    QString date = TContainer::getDateHash();
    quint16 id = qChecksum((date + attributes.value("Name").toString() + "Scan").toStdString().data(), (date + attributes.value("Name").toString() + "Scan").length());
    attributes.insert("Id", QString::number(id, 16).toUpper());

    int j = 0;
    while (j < Plot->ThreeDPlots.size() && !Plot->ThreeDPlots.at(j)->BCrossectionIsChecked)
        j++;

    if (j < Plot->ThreeDPlots.size()) {
        QString date = scan_attributes.value("Date").toString();

        quint16 gid = qChecksum((date + Plot->ThreeDPlots.at(j)->Title + "Scan").toStdString().data(), (date + Plot->ThreeDPlots.at(j)->Title + "Scan").length());
        attributes.insert("Parent", QString::number(gid, 16).toUpper());
    }

    attributes.insert("GroupId", scan_attributes.value("GroupId"));

    attributes.insert("Source", "Scan");

    if (j < Plot->ThreeDPlots.size()) {
        attributes.insert("Marker A x", Plot->ThreeDPlots.at(Plot->ActivePlot)->MarkerA->xValue()*1e6);
        attributes.insert("Marker A y", Plot->ThreeDPlots.at(Plot->ActivePlot)->MarkerA->yValue()*1e6);

        attributes.insert("Marker B x", Plot->ThreeDPlots.at(Plot->ActivePlot)->MarkerB->xValue()*1e6);
        attributes.insert("Marker B y", Plot->ThreeDPlots.at(Plot->ActivePlot)->MarkerB->yValue()*1e6);
    }

    emit AddDataToContainer(src, attributes);
}

void ScanPlot::AddDataToContainer_slot(plot::T3dDataOwner src, plot::TMetaStorage attributes)
{
    QString name = Plot->ThreeDPlots.at(Plot->ActivePlot)->Title;
    attributes.insert("Name", name);

    QString date = scan_attributes.value("Date").toString();

    quint16 id = qChecksum((date + name + "Scan").toStdString().data(), (date + name + "Scan").length());
    attributes.insert("Id", QString::number(id, 16).toUpper());

    attributes.unite(scan_attributes);

    emit AddDataToContainer(src, attributes);
}


void ScanPlot::Set_CBPoints(QComboBox *box)
{
    box->clear();

    if (show_additional_gui() || show_advanced_gui()) {
        box->addItem(QString::number(5));
        box->addItem(QString::number(10));
        box->addItem(QString::number(20));
    }

    box->addItem("30");
    box->addItem("50");

    for (int i = 1; i * 100 <= 1000; i++)
        box->addItem(QString::number(i * 100));

    if (show_advanced_gui()) {
        box->addItem(QString::number(2000));
        box->addItem(QString::number(3000));
        box->addItem(QString::number(4000));
        box->addItem(QString::number(5000));
        box->addItem(QString::number(6000));
        box->addItem(QString::number(8000));
        box->addItem(QString::number(10000));
    }

    box->setCurrentIndex(show_additional_gui() || show_advanced_gui() ? 7 : 4);
}

void ScanPlot::sl_CB_update()
{
    for (int i = 0; i < 4; i++) {
        Plot->SetTitle(CBChannel[i]->currentText(), i);

        int idx = CBChannel[i]->itemData(CBChannel[i]->currentIndex()).toInt();

        QString units;
        if (i == 1 && currExtScan)
            units = currExtScan->attributes.value("ZUnit", QObject::tr("Counts")).toString();
        else
            units = input_units(idx);

        Plot->SetZDescAndSIUnit(QString(), units, i);
    }
    Plot->UpdateMinMaxInSlider();
}

void ScanPlot::sl_CB_filters_update()
{
    for (int i = 0; i < 4; i++) {
        int idx = CBChannel[i]->itemData(CBChannel[i]->currentIndex()).toInt();
        if (!idx)
            continue;

        bool want_math = false;
        if (idx == srcHeight || idx == srcDACZ || idx == srcDACZ2 || idx == srcDACZ3
                || idx == srcCZ || idx == srcCZ2 || idx == srcCZ3)
            want_math = true;

        bool ydir = (CBStartPos->currentIndex() % 2 != 0);

        if (!want_math || Plot->ThreeDPlots.at(i)->CurrentSubtrOrder < 1) {
            Plot->ThreeDPlots[i]->CurrentSubtrOrder = (want_math ? 2 : 0) - 1;
            Plot->ThreeDPlots[i]->Refresh();
            if (Plot->ActivePlot == i)
                Plot->PlotControl->AGSubtr->actions()[want_math ? 2 : 0]->trigger();
        }

        Plot->ThreeDPlots[i]->MakeCorrectLineByMedianaXDir(want_math && !ydir);
        Plot->ThreeDPlots[i]->MakeCorrectLineByMedianaYDir(want_math && ydir);
        if (Plot->ActivePlot == i) {
            Plot->PlotControl->MMathProcessing->actions()[0]->setChecked(want_math && !ydir);
            Plot->PlotControl->MMathProcessing->actions()[1]->setChecked(want_math && ydir);
        }
    }
}

void ScanPlot::sl_set_section_enabled(bool on)
{
    sectionEnabled = on;
    Plot->TwoDPlot->Pure2dPlot->MathProcessing->Work[plot::math::TMathProcessing::Scale2d].arg = sectionEnabled ? Ch_activated : -1;
    SetSectionUnits();

    Plot->TwoDPlot->Pure2dPlot->SourceData.clear();
    if (sectionEnabled)
        Plot->TwoDPlot->Pure2dPlot->SourceData.resize(4); //TODO show?
}

void ScanPlot::sl_section_legend(int value)
{
    Ch_activated = (unsigned int)value;
    Plot->TwoDPlot->Pure2dPlot->MathProcessing->Work[plot::math::TMathProcessing::Scale2d].arg = sectionEnabled ? Ch_activated : -1;
    SetSectionUnits();
}

void ScanPlot::SetSectionUnits()
{
    int idx = CBChannel[Ch_activated]->itemData(CBChannel[Ch_activated]->currentIndex()).toInt();

    QString units = input_units(idx);
    double max = 100000;
    if (idx == srcCCD)
        units = QObject::tr("Counts");
    else if (idx == srcHeight)
        max = 1;
    else
        max = 50;

    if (currExtScan && (!sectionEnabled || Ch_activated == 1)) {
        Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(currExtScan->attributes.value("XLabel").toString(), currExtScan->attributes.value("XUnit").toString());
        Plot->TwoDPlot->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(currExtScan->attributes.value("YLabel").toString(), currExtScan->attributes.value("YUnit").toString());
        Plot->TwoDPlot->Pure2dPlot->SetDefaultYLimits(-100000, 100000);
        return;
    }
    Plot->TwoDPlot->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(QString(), units);
    Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QString(), tr("m"));
    Plot->TwoDPlot->Pure2dPlot->SetDefaultYLimits(-max, max);
}


void ScanPlot::GUICreate()
{
    bool show_additional, show_advanced, show_spm;
    {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        show_additional = show_additional_gui();
        show_advanced = show_advanced_gui();
        show_spm = settings.value("display/spm", true).toBool();
    }


    QHBoxLayout *cl = new QHBoxLayout(this);
    cl->setSpacing(6);
    cl->setContentsMargins(9, 9, 3, 9);

    QGroupBox *scanControl = new QGroupBox(this);
    HFlowLayout *layout_scanControl = new HFlowLayout(scanControl);
    layout_scanControl->spaceX = 6; layout_scanControl->spaceY = 6;
    layout_scanControl->setContentsMargins(3, 3, 3, 3);

    LeftScroll = new TScrollArea(this);
    LeftScroll->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Minimum);

    cl->addWidget(scanControl);
    cl->addWidget(LeftScroll, 4);

    QWidget *LeftWidget = new QWidget(LeftScroll);
    QVBoxLayout *LeftLay = new QVBoxLayout(LeftWidget);
    LeftLay->setSpacing(3);
    LeftLay->setContentsMargins(6, 3, 6, 6);

    LeftScroll->setWidget(LeftWidget);


    //--------------------------------------------------------------------------------------------
    tabgroup1 = new TTabWidget(LeftWidget);
    LeftLay->addWidget(tabgroup1);

    TTabWidget *tabgroup2 = new TTabWidget(LeftWidget);
    LeftLay->addWidget(tabgroup2);


    QWidget *GrBScanControl = new QWidget;
    QGridLayout *GLUpForGrBgrid = new QGridLayout(GrBScanControl);
    GLUpForGrBgrid->setAlignment(Qt::AlignTop);
    GLUpForGrBgrid->setColumnStretch(0, 2);
    GLUpForGrBgrid->setColumnStretch(1, 1);
    QFormLayout *GLUpForGrB = makeFormLayout();
    GLUpForGrBgrid->addLayout(GLUpForGrB, 1, 0);

    BoxExternal = new QWidget;
    QVBoxLayout *lBoxExt0 = new QVBoxLayout(BoxExternal);
    QHBoxLayout *lBoxExt1 = new QHBoxLayout;
    QFormLayout *lBoxExternal = makeFormLayout();
    lBoxExt0->setAlignment(Qt::AlignTop);
    lBoxExt0->addLayout(lBoxExt1);
    lBoxExt1->addLayout(lBoxExternal, 2);

    QWidget* box_FB = new QWidget;
    QVBoxLayout *layout_FB = new QVBoxLayout(box_FB);
    layout_FB->setAlignment(Qt::AlignTop);
    layout_FB->setSpacing(6);

    QWidget* box_add = new QWidget;
    QVBoxLayout *layout_add = new QVBoxLayout(box_add);
    layout_add->setAlignment(Qt::AlignTop);
    layout_add->setSpacing(6);

    QWidget* box_sync = new QWidget;
    QVBoxLayout *layout_sync = new QVBoxLayout(box_sync);
    layout_sync->setAlignment(Qt::AlignTop);
    layout_sync->setSpacing(6);

    QGroupBox *box_fbi = new QGroupBox(tr("FB I"));


    QWidget *WScanFW = new QWidget;
    QFormLayout *GLChannelsForGrB = makeFormLayout(WScanFW);
    QWidget *WScanBW = new QWidget;
    QFormLayout *GLChannelsForGrB2 = makeFormLayout(WScanBW);


    tabgroup1->addTab(GrBScanControl, tr("SCAN"));
    tabgroup1->addTab(BoxExternal, tr("EXT"));
    tabgroup1->addTab(box_FB, tr("FEEDBACK")); //TODO add tab from spm::connected()
    tabgroup1->addTab(box_add, tr("Additional"));
    tabgroup1->addTab(box_sync, tr("Sync"));
    tabgroup1->setTabVisible(1, false);

    tabgroup2->addTab(WScanFW, tr("Forward Channels"));
    tabgroup2->addTab(WScanBW, tr("Backward Channels"));


    if (!show_spm)
        tabgroup1->setTabVisible(2, false);
    if (!show_additional) {
        box_fbi->hide();
        tabgroup1->setTabVisible(3, false);
        tabgroup2->setTabVisible(1, false);
    }
    if (!show_advanced) {
        tabgroup1->setTabVisible(4, false);
    }


    //-------- GrBCurrentPoint ---------------------------------------------------------------------
    LEdX = new Edit(0, 100, 3);
    LEdY = new Edit(0, 100, 3);
    connect(LEdX, SIGNAL(newValue()), this, SLOT(sl_Update_XYZ_pos()));
    connect(LEdY, SIGNAL(newValue()), this, SLOT(sl_Update_XYZ_pos()));

    QWidget *GrBCurrentPoint = new QGroupBox(tr("XY Pos"));
    scanControlWidgets.append(GrBCurrentPoint);
    QGridLayout *MainLayoutForCurPos = new QGridLayout(GrBCurrentPoint);
    MainLayoutForCurPos->addWidget(new QLabel(tr("Xpos, ")+tr("micro")+tr("m")), 1, 0);
    MainLayoutForCurPos->addWidget(LEdX, 1, 1);
    MainLayoutForCurPos->addWidget(new QLabel(tr("Ypos, ")+tr("micro")+tr("m")), 2, 0);
    MainLayoutForCurPos->addWidget(LEdY, 2, 1);

    //-------- ZPos ---------------------------------------------------------------------
    Zposition = new QwtThermo();
    Zposition->setOrientation(Qt::Horizontal, QwtThermo:: NoScale);
    Zposition->setFillBrush(Qt::green);
    Zposition->setRange(0, 1, 0);
    Zposition->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Fixed);
    ZposValue = new QLabel("Z pos = 100%");

    QGroupBox *box_zposition = new QGroupBox(tr("Current Tip Position"));
    QHBoxLayout *layout_zposition = new QHBoxLayout(box_zposition);
    layout_zposition->setContentsMargins(6, 6, 6, 6);
    layout_zposition->addWidget(Zposition, 1);
    layout_zposition->addWidget(ZposValue);


    //-------- GrBDACscan ---------------------------------------------------------------------
    GrBDACscan = new QGroupBox(tr("DAC(Z) scan"));
    GrBDACscan->setCheckable(true);
    GrBDACscan->setChecked(false);
    GrBDACscan->hide();
    scanControlWidgets.append(GrBDACscan);
    QFormLayout *layout_dacscan = makeFormLayout(GrBDACscan);
    layout_dacscan->setContentsMargins(6, 6, 6, 6);

    CBDACscanCh = new QComboBox;
    CBDACscanCh->addItem(tr("None "), -1);
    EDACstep = new Edit(0, 10000, 2);
    EDACcount = new Edit(2, 100);
    EDACcount->setImmediateMode();
    layout_dacscan->addRow(tr("Channel"), CBDACscanCh);
    layout_dacscan->addRow(tr("Count"), EDACcount);
    layout_dacscan->addRow(tr("Step"), EDACstep);


    if (show_additional)
        LeftLay->addWidget(GrBCurrentPoint);
    if (!show_spm)
        box_zposition->hide();
    else
        LeftLay->addWidget(box_zposition);

    LeftLay->addSpacerItem(new QSpacerItem(1, 6));
    LeftLay->addWidget(GrBDACscan);
    LeftLay->addStretch(1);


    //-------- GrBScanControl ---------------------------------------------------------------------
    m_BStart = new TwoStateButton(":start.png",":stop.png");
    connect(m_BStart, SIGNAL(state_changed(bool)), this, SLOT(scan_start_stop(bool)));

    m_BStart->setProperty("id", DataID::ScanOn);

    CBStartPos = new QComboBox;
    scanControlWidgets.append(CBStartPos);
    CBStartPos->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    CBStartPos->addItem(makeHugeIcon(":/ScanPlot/Top_left_X.png"), QString());
    CBStartPos->addItem(makeHugeIcon(":/ScanPlot/Top_left_Y.png"), QString());
    CBStartPos->addItem(makeHugeIcon(":/ScanPlot/Top_right_X.png"), QString());
    CBStartPos->addItem(makeHugeIcon(":/ScanPlot/Top_right_Y.png"), QString());
    CBStartPos->addItem(makeHugeIcon(":/ScanPlot/Bottom_left_X.png"), QString());
    CBStartPos->addItem(makeHugeIcon(":/ScanPlot/Bottom_left_Y.png"), QString());
    CBStartPos->addItem(makeHugeIcon(":/ScanPlot/Bottom_right_X.png"), QString());
    CBStartPos->addItem(makeHugeIcon(":/ScanPlot/Bottom_right_Y.png"), QString());
    connect(CBStartPos, SIGNAL(currentIndexChanged(int)), this, SLOT(updateDirection()));

    LiftMode = new TwoStateButton(":/ScanPlot/LF_off.png", ":/ScanPlot/LF_on.png");
    scanControlWidgets.append(LiftMode);
    LiftMode->setToolTip(tr("Switch to Lift Mode"));
    if (!show_additional)
        LiftMode->hide();
    connect(LiftMode, SIGNAL(state_changed(bool)), this, SLOT(LmOnOff(bool)));

    LEd_LFmodeHeight = new Edit(-2000, 20000, -1);
    LEd_LFmodeHeight->setValue(1000);
    connect(LEd_LFmodeHeight, SIGNAL(newValue(double)), this, SLOT(scan_lm_height(double)));
    LblLiftMode = new QLabel(tr("LM height, nm"));
    LblLiftMode->hide();
    LEd_LFmodeHeight->hide();

    for (int i = 0; i < 4; i++) {
        LmSwitch[i] = new QCheckBox;
        LmSwitch[i]->hide();
    }

    LmSwitch[0]->setText(tr("Lift mode A"));
    LmSwitch[1]->setText(tr("Lift mode B"));
    LmSwitch[2]->setText(tr("Lift mode C"));
    LmSwitch[3]->setText(tr("Lift mode D"));


    CBExternalMode = new QComboBox;
    scanControlWidgets.append(CBExternalMode);
    CBExternalMode->addItem(tr("None"));
    connect(CBExternalMode, SIGNAL(currentIndexChanged(int)), this, SLOT(sl_ext_mode()));

    BCCDXYLock = new TwoStateButton(":/ScanPlot/LockOFF.png",":/ScanPlot/LockON.png");
    scanControlWidgets.append(BCCDXYLock);
    BCCDXYLock->setToolTip(tr("Same X/Y points as scan"));
    BCCDXYLock->change_state_man(true);
    connect(BCCDXYLock, SIGNAL(state_changed(bool)), this, SLOT(sl_ext_mode()));

    CBPointsOXext = new QComboBox;
    CBPointsOYext = new QComboBox;
    scanControlWidgets.append(CBPointsOXext);
    scanControlWidgets.append(CBPointsOYext);
    Set_CBPoints(CBPointsOXext);
    Set_CBPoints(CBPointsOYext);
    connect(CBPointsOXext, SIGNAL(currentIndexChanged(int)), this, SLOT(sl_ext_mode()));
    connect(CBPointsOYext, SIGNAL(currentIndexChanged(int)), this, SLOT(sl_ext_mode()));

    lBoxExternal->addRow(tr("Source"), CBExternalMode);
    lBoxExt1->addWidget(BCCDXYLock, 1);
    if (!show_additional) {
        lBoxExternal->addRow(tr("Points"), CBPointsOXext);
    } else {
        lBoxExternal->addRow(tr("X points"), CBPointsOXext);
        lBoxExternal->addRow(tr("Y points"), CBPointsOYext);
    }


    LEdDeltaX =  new Edit(10e-9 * 1e6, 200, 3);
    LEdDeltaX->setEnabled(false);

    LEdDeltaY = new Edit(10e-9 * 1e6, 200, 3);
    LEdDeltaY->setEnabled(false);

    CBScanner = new QComboBox;
    scanControlWidgets.append(CBScanner);
    CBScanner->setProperty("id", DataID::ScanScanner);
    CBScanner->addItem(tr("Head"));
    CBScanner->addItem(tr("Stage"));
    connect(CBScanner, SIGNAL(currentIndexChanged(int)), this, SLOT(setScanner(int)));

    CBPointsOX = new QComboBox;
    CBPointsOY = new QComboBox;
    scanControlWidgets.append(CBPointsOX);
    scanControlWidgets.append(CBPointsOY);

    Set_CBPoints(CBPointsOX);
    Set_CBPoints(CBPointsOY);

    GLUpForGrB->addRow(LScanner = new QLabel(tr("Scanner")), CBScanner);
    if (!show_additional) {
        GLUpForGrB->addRow(tr("Points"), CBPointsOX);
        GLUpForGrB->addRow(tr("Size, ")+tr("micro")+tr("m"), LEdDeltaX);
    } else {
        GLUpForGrB->addRow(tr("X points"), CBPointsOX);
        GLUpForGrB->addRow(tr("Y points"), CBPointsOY);
        GLUpForGrB->addRow(tr("Width, ")+tr("micro")+tr("m"), LEdDeltaX);
        GLUpForGrB->addRow(tr("Height, ")+tr("micro")+tr("m"), LEdDeltaY);
    }
    GLUpForGrB->addRow(LblLiftMode, LEd_LFmodeHeight);

    for (int i = 0; i < 4; i++) {
        CBChannel[i] = new QComboBox;
        scanControlWidgets.append(CBChannel[i]);
        fillCBchannels_(CBChannel[i], false);
        CBChannel[i]->setProperty("id", DataID::ScanChA+i);
        CBChannel[i]->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    }
    CBChannel[1]->addItem(QObject::tr("CCD")+"/EXT", srcCCD);

    //-------- GrBVelocity ---------------------------------------------------------------------
    LScanRateWarn = new QLabel;
    LScanRateWarn->setPixmap(qApp->style()->standardPixmap(QStyle::SP_MessageBoxWarning));
    LScanRateWarn->setEnabled(false);

    DCScanRate = new DoubleChanger(0, 10, 2);
    DCScanRate->edit->setSpecialValueText(tr("min"));
    DCScanRate->setProperty("id", DataID::ScanV);
    DCScanRate->setValue(user_scan_speed = requested_scan_speed = 1);
    {
        QSettings settings("Param/controller.conf", QSettings::IniFormat);
        DCScanRate->setMaximum(settings_double(settings, "limit_scan_rate", 10));
    }
    connect(DCScanRate, SIGNAL(newValue()), this, SLOT(check_speed()));

    QGroupBox *GrBVelocity = new QGroupBox(tr("Velocity, Hz"));
    QHBoxLayout *MainLayoutForVelocity = new QHBoxLayout(GrBVelocity);
    MainLayoutForVelocity->setContentsMargins(6, 6, 6, 6);
    GLUpForGrBgrid->addWidget(GrBVelocity, 2, 0, 1, 2);
    MainLayoutForVelocity->addWidget(DCScanRate);
    MainLayoutForVelocity->addWidget(LScanRateWarn);

    QPalette plt;
    CBSec[0] = new QCheckBox(tr("Section") + " A");
    CBSec[0]->setProperty("id", DataID::ScanChSecA);
    plt.setColor(QPalette::WindowText, Qt::black); CBSec[0]->setPalette(plt);
    CBSec[1] = new QCheckBox(tr("Section") + " B");
    CBSec[1]->setProperty("id", DataID::ScanChSecB);
    plt.setColor(QPalette::WindowText, Qt::red); CBSec[1]->setPalette(plt);
    CBSec[2] = new QCheckBox(tr("Section") + " C");
    CBSec[2]->setProperty("id", DataID::ScanChSecC);
    plt.setColor(QPalette::WindowText, Qt::green); CBSec[2]->setPalette(plt);
    CBSec[3] = new QCheckBox(tr("Section") + " D");
    CBSec[3]->setProperty("id", DataID::ScanChSecD);
    plt.setColor(QPalette::WindowText, Qt::blue); CBSec[3]->setPalette(plt);

    GLChannelsForGrB->addRow(QObject::tr("Channel A"), CBChannel[0]);
    GLChannelsForGrB->addRow(QObject::tr("Channel B"), CBChannel[1]);
    GLChannelsForGrB->addRow(CBSec[0], CBSec[1]);
    GLChannelsForGrB->addRow(LmSwitch[0], LmSwitch[1]);

    GLChannelsForGrB2->addRow(QObject::tr("Channel C"), CBChannel[2]);
    GLChannelsForGrB2->addRow(QObject::tr("Channel D"), CBChannel[3]);
    GLChannelsForGrB2->addRow(CBSec[2], CBSec[3]);
    GLChannelsForGrB2->addRow(LmSwitch[2], LmSwitch[3]);

    //---------FBControl--------------------------------------------------------------------------
    DCfbP = new DoubleChanger(0, 10, 2);
    DCfbI = new DoubleChanger(0, 1, 3);
    DCfbSetPoint = new DoubleChanger(-10, 10, 2);

    DCfbP->setProperty("id", DataID::SPM_P);
    DCfbI->setProperty("id", DataID::SPM_I);
    DCfbSetPoint->setProperty("id", DataID::SPM_SetPoint);

    spmPrecision = new DoubleChanger(0, 20, 5);
    spmPrecision->setValue(20);
    connect(spmPrecision, SIGNAL(newValue(double)), this, SLOT(spm_pid_precision(double)));

    firstDelay = new DoubleChanger(0, 2, 1);
    lineDelay = new DoubleChanger(0, 0.5, 2);
    pointDelay = new DoubleChanger(0, 1, 4);
    pointDelay->setProperty("id", DataID::ScanPointDelay);
    connect(firstDelay, SIGNAL(newValue(double)), this, SLOT(scan_first_delay(double)));
    connect(lineDelay, SIGNAL(newValue(double)), this, SLOT(scan_line_delay(double)));
    connect(pointDelay, SIGNAL(newValue(double)), this, SLOT(scan_point_delay(double)));

    fastReturn = new QCheckBox(tr("Fast backward line"));
    fastReturn->setToolTip(tr("Good with low scan rate or per-point EXT mode(e.g. CCD)")
                           + "\n" + tr("Restricts data to Channels A and B only"));
    connect(fastReturn, SIGNAL(toggled(bool)), this, SLOT(scan_fast_return(bool)));

    pointSync = new QComboBox;
    pointSync->addItem(tr("Internal"));
    pointSync->addItem(tr("Rising edge"));
    pointSync->addItem(tr("Falling edge"));
    connect(pointSync, SIGNAL(currentIndexChanged(int)), this, SLOT(scan_sync_mode(int)));


    QGroupBox *box_fbp = new QGroupBox(tr("FB P"));
    QHBoxLayout *layout_fbp = new QHBoxLayout(box_fbp);
    layout_fbp->setContentsMargins(6, 6, 6, 6);
    layout_fbp->addWidget(DCfbP);

    QHBoxLayout *layout_fbi = new QHBoxLayout(box_fbi);
    layout_fbi->setContentsMargins(6, 6, 6, 6);
    layout_fbi->addWidget(DCfbI);

    QGroupBox *box_setPoint = new QGroupBox(tr("Set point"));
    QHBoxLayout *layout_setPoint = new QHBoxLayout(box_setPoint);
    layout_setPoint->setContentsMargins(6, 6, 6, 6);
    layout_setPoint->addWidget(DCfbSetPoint);

    layout_FB->addWidget(box_setPoint);
    layout_FB->addWidget(box_fbp);
    layout_FB->addWidget(box_fbi);


    QGroupBox *box_spm_prec = new QGroupBox(tr("Set point precision"));
    QHBoxLayout *layout_spm_prec = new QHBoxLayout(box_spm_prec);
    layout_spm_prec->addWidget(spmPrecision);
    layout_spm_prec->setContentsMargins(6, 6, 6, 6);
    if (!show_spm)
        box_spm_prec->hide();

    QGroupBox *box_firstDelay = new QGroupBox(tr("Scan First point delay, s"));
    QHBoxLayout *layout_firstDelay = new QHBoxLayout(box_firstDelay);
    layout_firstDelay->addWidget(firstDelay);
    layout_firstDelay->setContentsMargins(6, 6, 6, 6);

    QGroupBox *box_lineDelay = new QGroupBox(tr("Line First point delay, s"));
    QHBoxLayout *layout_lineDelay = new QHBoxLayout(box_lineDelay);
    layout_lineDelay->addWidget(lineDelay);
    layout_lineDelay->setContentsMargins(6, 6, 6, 6);

    layout_add->addWidget(box_firstDelay);
    layout_add->addWidget(box_lineDelay);
    layout_add->addWidget(box_spm_prec);
    layout_add->addWidget(fastReturn);


    QGroupBox *box_pointDelay = new QGroupBox(tr("Point delay, s"));
    QHBoxLayout *layout_pointDelay = new QHBoxLayout(box_pointDelay);
    layout_pointDelay->addWidget(pointDelay);
    layout_pointDelay->setContentsMargins(6, 6, 6, 6);

    layout_sync->addWidget(pointSync);
    layout_sync->addWidget(box_pointDelay);
    QLabel *sync_warn = new QLabel(tr("Sync output is enabled for nonzero delay or with external sync input."));
    sync_warn->setWordWrap(true);
    layout_sync->addWidget(sync_warn);



    BMotorLevel = new QPushButton();
    BMotorLevel->setIcon(makeHugeIcon(":/SPMTuning_window/Motors_leveling.png"));
    BMotorLevel->setToolTip(tr("Level Head by scan data"));
    BMotorLevel->hide();
    connect(BMotorLevel, SIGNAL(clicked()), this, SLOT(sl_motor_level()));

    FastScanCorrection = new QToolButton;
    scanControlWidgets.append(FastScanCorrection);
    FastScanCorrection->setPopupMode(QToolButton::InstantPopup);
    FastScanCorrection->setIcon(makeHugeIcon(":/ScanPlot/fscanOf.png"));
    FastScanCorrection->setToolTip(tr("Enable fast scan correction"));
    FastScanCorrection->setMenu(new QMenu);
    QActionGroup *fast_correct_ag = new QActionGroup(FastScanCorrection);
    fast_correct_ag->addAction(FastScanCorrection->menu()->addAction(tr("None")))->setCheckable(true);
    fast_correct_ag->addAction(FastScanCorrection->menu()->addAction(tr("Simple")))->setCheckable(true);
    fast_correct_ag->addAction(FastScanCorrection->menu()->addAction(tr("Hybrid")))->setCheckable(true);
    fast_correct_ag->addAction(FastScanCorrection->menu()->addAction(tr("Voltage")))->setCheckable(true);
    fast_correct_ag->actions()[0]->setChecked(true);
    connect(FastScanCorrection, SIGNAL(triggered(QAction*)), this, SLOT(sl_ScanCorrection()));

    BPreciseScan = new TwoStateButton(":/ScanPlot/PresizeScan.png",":/ScanPlot/PresizeScanOn.png");
    BPreciseScan->setToolTip(tr("Enable precise scanning"));

    BContinuousScan = new TwoStateButton(":/CCD/oneshoot.png",":/CCD/continum.png");
    BContinuousScan->setToolTip(tr("Enable continuous scanning") + "/" + tr("DAC(Z)-scan"));
    connect(BContinuousScan, SIGNAL(state_changed(bool)), this, SLOT(sl_dacscan()));
    connect(BContinuousScan, SIGNAL(state_changed(bool)), GrBDACscan, SLOT(setVisible(bool)));

    if (!show_advanced) {
        box_spm_prec->hide();
        FastScanCorrection->hide();
        BPreciseScan->hide();
        BContinuousScan->hide();
        LScanRateWarn->hide();
    }

    BXYLock = new TwoStateButton(":/ScanPlot/LockOFF.png",":/ScanPlot/LockON.png");
    scanControlWidgets.append(BXYLock);
    BXYLock->setToolTip(tr("Sync XY size & points"));
    BXYLock->change_state_man(true);
    if (!show_additional)
        BXYLock->hide();

    PBToNewWindow = new TwoStateButton(":/Detach.png",":/Attach.png");
    PBToNewWindow->setToolTip(QObject::tr("Individual window on/off"));

    lb_ccd_scan = new QLabel;
    lb_ccd_scan->setMaximumSize(PBToNewWindow->sizeHint().width(), PBToNewWindow->sizeHint().width());
    lb_ccd_scan->setScaledContents(true);
    lb_ccd_scan->setPixmap(QPixmap(":/ScanPlot/CCD.png"));
    lb_ccd_scan->hide();

    lb_ccd_scan_sm = new QLabel;
    lb_ccd_scan_sm->setMaximumSize(PBToNewWindow->sizeHint().width(), PBToNewWindow->sizeHint().width());
    lb_ccd_scan_sm->setScaledContents(true);
    lb_ccd_scan_sm->setPixmap(QPixmap(":/ScanPlot/CCD_Smart.png"));
    lb_ccd_scan_sm->hide();


    layout_scanControl->addWidget(PBToNewWindow);
    layout_scanControl->addWidget(m_BStart);
    layout_scanControl->addWidget(BContinuousScan);
    addDelimeterH(layout_scanControl, PBToNewWindow->sizeHint().width());
    layout_scanControl->addWidget(FastScanCorrection);
    layout_scanControl->addWidget(BPreciseScan);
    layout_scanControl->addWidget(LiftMode);
    layout_scanControl->addWidget(CBStartPos);
    layout_scanControl->addWidget(BMotorLevel);
    layout_scanControl->addWidget(lb_ccd_scan);
    layout_scanControl->addWidget(lb_ccd_scan_sm);


    //
    Plot = new plot::T3dPlot(4, this, this);
    for (int i = 0; i < 4; i++)
        Plot->ThreeDPlots[i]->blockRefresh = true;

    Plot->SetDefaultXLimits(0, Xmax * 1e-6);
    Plot->SetDefaultYLimits(0, Ymax * 1e-6);
    Plot->SetXYDescAndSIUnit(QString(), tr("m"), -1);
    Plot->SetTitle(tr("OFF"),-1);
    Plot->SetPlotsXScales(0, 1e-4);  Plot->SetPlotsYScales(0, 1e-4);
    Plot->TwoDPlot->Pure2dPlot->SetDefaultXLimits(0, 200);
    Plot->TwoDPlot->Pure2dPlot->SetDefaultYLimits(-50, 50);
    connect(Plot, SIGNAL(ActivePlotChanged(int)), this, SLOT(sl_section_legend(int)));


    QPushButton *BSendToGraphEditor = new QPushButton;
    BSendToGraphEditor->setObjectName("BSendToGraphEditor");
    BSendToGraphEditor->setIcon(makeHugeIcon(":/MainToolbar/litho.png"));
    BSendToGraphEditor->setToolTip(QObject::tr("use area for lithography"));
    connect(BSendToGraphEditor, SIGNAL(clicked()), this, SLOT(BSendToGraphEditor_clicked()));
    BSendToGraphEditor->hide();

    Plot->PlotControl->ControlMainLayout->addWidget(BSendToGraphEditor);

    if (!show_additional) {
        Plot->Containers.at(1)->hide();
    }

    for (int i = 0; i < 4; i++) {
        scanLineInd[i] = new QwtPlotMarker;
        scanLineInd[i]->setXValue(0);
        scanLineInd[i]->setYValue(0);
        scanLineInd[i]->setLinePen(QPen(QColor(255, 0, 0), 0, Qt::SolidLine));
        scanLineInd[i]->setLineStyle(QwtPlotMarker::HLine);
        scanLineInd[i]->attach(Plot->ThreeDPlots.at(i)->QPlot);
    }
    for (int i = 0; i < 4; i++)
        Plot->ThreeDPlots[i]->blockRefresh = false;
    Plot->Refresh();

    QGridLayout *blay = new QGridLayout;
    GLUpForGrBgrid->addLayout(blay, 1, 1, 1, 1, Qt::AlignTop);
    blay->addWidget(BXYLock, 0, 0);
    blay->addWidget(Plot->PlotControl->BRestoreScale, 1, 0);
    blay->addWidget(Plot->PlotControl->BSelectByRect, 2, 0);
    blay->addWidget(Plot->PlotControl->BApply, 1, 0);
    blay->addWidget(Plot->PlotControl->BCancel, 2, 0);

    scanControlWidgets << Plot->PlotControl->BMarkers <<
                          Plot->PlotControl->BCross <<
                          Plot->PlotControl->BSelectByRect <<
                          BSendToGraphEditor <<
                          Plot->PlotControl->BFixX <<
                          Plot->PlotControl->BFixY <<
                          Plot->PlotControl->BRestoreScale <<
                          Plot->TwoDPlot->PlotControl->CBChooseActiveCurve <<
                          Plot->TwoDPlot->PlotControl->TBCurveColour;

    cl->addWidget(Plot, 11);

    //------------------------
    connect(LEdDeltaX, SIGNAL(newValue()), this, SLOT(onLEdDeltaX_editingFinished()));
    connect(LEdDeltaY, SIGNAL(newValue()), this, SLOT(onLEdDeltaY_editingFinished()));

    connect(CBPointsOX, SIGNAL(currentIndexChanged(int)), this, SLOT(check_nx_ny()));
    connect(CBPointsOY, SIGNAL(currentIndexChanged(int)), this, SLOT(check_nx_ny()));

    for (int i = 0; i < 4; i++)
        connect(CBChannel[i], SIGNAL(currentIndexChanged(int)), this, SLOT(sl_CB_update()));
}


void ScanPlot::setScanner(int)
{
    for (int i = 0; i < 4; i++)
        Plot->ThreeDPlots[i]->SourceData = plot::getUnLock(scanData[CBScanner->currentIndex()][i]);
    Plot->Refresh();

    sl_set_XYZ_range_sync();

    if (CBScanner->currentIndex() == scHead) {
        LEdX->setValue(x);
        LEdY->setValue(y);
    } else {
        LEdX->setValue(x2);
        LEdY->setValue(y2);
    }
    sl_syncXY();

    check_speed();
}
void ScanPlot::updateDirection()
{
    bool prev_scan_direction = scan_direction;
    scan_direction = (CBStartPos->currentIndex() % 2 != 0); //0-fast axis x, 1-y


    if (prev_scan_direction == scan_direction)
        return;

    double active_range0 = !prev_scan_direction ? xRight - xLeft : yTop - yBottom;
    double active_range1 = !scan_direction ? xRight - xLeft : yTop - yBottom;
    if (active_range1 > 0) {
        user_scan_speed *= active_range0 / active_range1; //preserve tip speed
        if (DCScanRate->value() != user_scan_speed)
            DCScanRate->setValue(user_scan_speed);
    }

    check_speed();
}

void ScanPlot::LmOnOff(bool value)
{
    for (int i = 0; i < 4; i++)
        LmSwitch[i]->setVisible(value);
    LblLiftMode->setVisible(value);
    LEd_LFmodeHeight->setVisible(value);
    LeftScroll->setWidget(LeftScroll->takeWidget()); //reset sizeHint
    LeftScroll->updateGeometry(); //notify layout
}


void ScanPlot::acquireNextPoint()
{
    if (!m_BStart->on_off || CBExternalMode->currentIndex() == 0)
        return;

    if (!currExtScan->acquireNextPoint()) {
        new Notification(tr("External scan") + ": " + tr("Scan error:") + " " + currExtScan->lastError, Notification::Bad, 60);
        doBeep();

        m_BStart->change_state_man_with_sig(false);
    }
}
void ScanPlot::sl_ext_scan_acquired()
{
    if (!m_BStart->on_off || CBExternalMode->currentIndex() == 0 || sender() != currExtScan)
        return;

    sendExtReady = true;

    if (ccd_scan_bar != 0)
        ccd_scan_bar->setValue(ccd_scan_bar->bar->value() + 1);

    sl_ext_scan_update();
}
void ScanPlot::sl_ext_scan_update()
{
    if (!refresh_3d_limit_timer[1].isValid() || refresh_3d_limit_timer[1].elapsed() > 100 + refresh_limit_last_elapsed[1]) { //~10 refreshes per second max
        refresh_3d_limit_timer[1].start();
        RefreshTimer4d->stop();

        currExtScan->updateData();
        refresh_limit_last_elapsed[1] = qMax<qint64>(refresh_limit_last_elapsed[1], refresh_3d_limit_timer[1].elapsed() / 0.8 * channelsActive); //time budget, %
    } else if (sender() != RefreshTimer4d && !RefreshTimer4d->isActive())
        RefreshTimer4d->start(200);
}
void ScanPlot::sl_ext_scan_refresh()
{
    if (CBExternalMode->currentIndex() == 0)
        return;

    Plot->SetZDescAndSIUnit(QString(), currExtScan->attributes.value("ZUnit", QObject::tr("Counts")).toString(), 1);

    SetSectionUnits();
    sl_scan_refresh_ch(1); //TODO for any ch
}
void ScanPlot::sl_ext_mode()
{
    IExternalScan *scan = CBExternalMode->currentIndex() != 0 ? externalScans.at(CBExternalMode->currentIndex() - 1) : 0;
    if (currExtScan != scan) {
        if (currExtScan)
            currExtScan->clear();

        currExtScan = scan;

        CBChannel[1]->setCurrentIndex(CBChannel[1]->findData(currExtScan ? currExtScan->getDataChannel() : srcNone));
        //CBChannelB->setEnabled(!currExtScan); //TODO onoff sync

        for (int i = 0; i < externalScanWidgets.size(); i++)
            if (QWidget *wid = externalScanWidgets.at(i))
                wid->setVisible(i + 1 == CBExternalMode->currentIndex());

        LeftScroll->setWidget(LeftScroll->takeWidget()); //reset sizeHint
        LeftScroll->updateGeometry(); //notify layout
    }

    lb_ccd_scan->setVisible(currExtScan && BCCDXYLock->on_off);
    lb_ccd_scan_sm->setVisible(currExtScan && !BCCDXYLock->on_off);

    //xy sync
    CBPointsOX->blockSignals(true);
    CBPointsOY->blockSignals(true);

    Set_CBPoints(CBPointsOX);
    Set_CBPoints(CBPointsOY);

    CBPointsOX->setCurrentIndex(CBPointsOXext->currentIndex());
    CBPointsOY->setCurrentIndex(CBPointsOYext->currentIndex());

    int nx = CBPointsOX->currentText().toInt();
    int ny = CBPointsOY->currentText().toInt();
    MobileServer::sendData(DataID::ScanNxNy, qSqrt(nx*ny));

    if (currExtScan && !BCCDXYLock->on_off) {
        int count = 0;
        do{
            if (CBPointsOX->itemText(count).toInt() % nx != 0)
                CBPointsOX->removeItem(count);
            else
                count++;
        } while (count < CBPointsOX->count());

        count = 0;
        do {
            if (CBPointsOY->itemText(count).toInt() % ny != 0)
                CBPointsOY->removeItem(count);
            else
                count++;
        } while (count < CBPointsOY->count());
    }

    CBPointsOX->blockSignals(false);
    CBPointsOY->blockSignals(false);

    if (BXYLock->on_off) {
        if (sender() == CBPointsOXext)
            CBPointsOYext->setCurrentIndex(CBPointsOXext->currentIndex());
        if (sender() == CBPointsOYext)
            CBPointsOXext->setCurrentIndex(CBPointsOYext->currentIndex());
    }
}

void ScanPlot::addExternalScan(QString name, IExternalScan *iface)
{
    CBExternalMode->addItem(name);

    iface->displayPlot = Plot;

    externalScans.append(iface);
    externalScanWidgets.append(iface->createInterfaceWidget());
    connect(iface, SIGNAL(acquired()), this, SLOT(sl_ext_scan_acquired()));
    connect(iface, SIGNAL(updated()), this, SLOT(sl_ext_scan_refresh()));
    connect(iface, SIGNAL(setSectionEnabled(bool)), this, SLOT(sl_set_section_enabled(bool)));

    if (externalScanWidgets.last()) {
        ((QVBoxLayout *)BoxExternal->layout())->addWidget(externalScanWidgets.last());
        externalScanWidgets.last()->hide();
    }

    tabgroup1->setTabVisible(1, true);
}


bool ScanPlot::is_dacscan()
{
    return BContinuousScan->on_off &&
            GrBDACscan->isChecked() &&
            CBDACscanCh->currentIndex() > 0;
}
void ScanPlot::sl_dacscan()
{
    GrBDACscan->setChecked(false);
    DACcurrentNum = 0;

    I_MODULE *oscm = ModuleManager::instance->find1("oscilloscope");
    if (!oscm)
        return;
    Oscilloscope *osc = (Oscilloscope *)oscm->basecast();

    CBDACscanCh->clear();
    CBDACscanCh->addItem(tr("None "), -1);
    osc->DacsWidget->fill_dacs_channel_CB(CBDACscanCh);
}
void ScanPlot::sl_dacscan_advance()
{
    if (!BContinuousScan->on_off || !GrBDACscan->isChecked()) {
        DACcurrentNum = 0;
        return;
    }

    if (CBDACscanCh->currentIndex() <= 0)
        return;

    I_MODULE *oscm = ModuleManager::instance->find1("oscilloscope");
    Oscilloscope *osc = (Oscilloscope *)oscm->basecast();
    DoubleChanger *dc = osc->DacsWidget->get_dacs_channel(CBDACscanCh);
    if (DACcurrentNum == 0)
        DACinitialValue = dc->value();

    bool last = DACcurrentNum == EDACcount->value()-1;
    if (last) {
        dc->setValue(DACinitialValue);
        DACcurrentNum = 0;
    } else {
        dc->setValue(dc->value() + EDACstep->value());
        DACcurrentNum++;
    }
    dc->setNewValue();

    if (!last)
        m_BStart->change_state_man_with_sig(true);
}


void ScanPlot::onLEdDeltaX_editingFinished()
{
    Plot->SetSelectWidth(LEdDeltaX->value() / 1e6);
    if (BXYLock->on_off) {
        LEdDeltaY->setValue(LEdDeltaX->value());
        Plot->SetSelectHeight(LEdDeltaY->value() / 1e6);
    }
}
void ScanPlot::onLEdDeltaY_editingFinished()
{
    Plot->SetSelectHeight(LEdDeltaY->value() / 1e6);
    if (BXYLock->on_off) {
        LEdDeltaX->setValue(LEdDeltaY->value());
        Plot->SetSelectWidth(LEdDeltaX->value() / 1e6);
    }
}

void ScanPlot::sendXYPos()
{
    if (CBScanner->currentIndex() == scHead)
        emit sig_XYPosition(LEdX->value(), LEdY->value());
    else
        emit sig_XYPosition2(LEdX->value(), LEdY->value());
}
void ScanPlot::sl_setCurrentX(double v)
{
    x = v;
    if (CBScanner->currentIndex() == scHead) {
        LEdX->setValue(v);
        sl_syncXY();
    }
}
void ScanPlot::sl_setCurrentY(double v)
{
    y = v;
    if (CBScanner->currentIndex() == scHead) {
        LEdY->setValue(v);
        sl_syncXY();
    }
}
void ScanPlot::sl_setCurrentX2(double v)
{
    x2 = v;
    if (CBScanner->currentIndex() == scStage) {
        LEdX->setValue(v);
        sl_syncXY();
    }
}
void ScanPlot::sl_setCurrentY2(double v)
{
    y2 = v;
    if (CBScanner->currentIndex() == scStage) {
        LEdY->setValue(v);
        sl_syncXY();
    }
}
void ScanPlot::sl_syncXY()
{
    int active = Plot->ActivePlot;
    for (int i = 0; i < 4; i++) {
        Plot->ActivePlot = i;
        Plot->SetCrossCoords(LEdX->value() / 1e6, LEdY->value() / 1e6);
    }
    Plot->ActivePlot = active;
}
void ScanPlot::sl_Update_XYZ_pos()
{
    sendXYPos();
    sl_syncXY();
}
void ScanPlot::slUpdZscanner(double zpos)
{
    if (zpos < 0.33) Zposition->setFillBrush(Qt::red);
    else if (zpos > 0.66) Zposition->setFillBrush(Qt::green);
    else Zposition->setFillBrush(Qt::blue);
    Zposition->setValue(zpos);
    ZposValue->setText(tr("Z pos = ")+(QString::number((int)(100*zpos),'f',0))+tr("%"));
}

void ScanPlot::sl_set_XYZ_range(double xm, double ym, double)
{
    bool xyu = qFuzzyCompare(Xmax, xm) &&
            qFuzzyCompare(Ymax, ym);

    Xmax = xm;
    Ymax = ym;

    if (CBScanner->currentIndex() != scHead)
        return;

    if (!xyu)
        sl_set_XYZ_range_sync();

    check_speed();
}
void ScanPlot::sl_set_XYZ_range2(double xm, double ym, double)
{
    bool xyu = qFuzzyCompare(Xmax2, xm) &&
            qFuzzyCompare(Ymax2, ym);

    Xmax2 = xm;
    Ymax2 = ym;

    if (CBScanner->currentIndex() != scStage)
        return;

    if (!xyu)
        sl_set_XYZ_range_sync();

    check_speed();
}
void ScanPlot::sl_set_XYZ_range_sync()
{
    double _xLeft = xLeft / LEdX->maximum();
    double _xRight = xRight / LEdX->maximum();
    double _yTop = yTop / LEdY->maximum();
    double _yBottom = yBottom / LEdY->maximum();

    double xm = Xmax, ym = Ymax;
    if (CBScanner->currentIndex() == scStage) {
        xm = Xmax2; ym = Ymax2;
    }

    if (_xRight - _xLeft > 0.8 && _yTop - _yBottom > 0.8) {
        _xLeft = _yBottom = 0;
        _xRight = _yTop = 1;
    }
    _xLeft *= xm;
    _xRight *= xm;
    _yTop *= ym;
    _yBottom *= ym;

    LEdDeltaX->setMaximum(xm);
    LEdDeltaY->setMaximum(ym);
    LEdX->setMaximum(xm);
    LEdY->setMaximum(ym);


    Plot->SetAllXScales(0, xm * 1e-6);
    Plot->SetAllYScales(0, ym * 1e-6);
    for (int i = 0; i < 4; i++) {
        Plot->ThreeDPlots[i]->SetPlotsXScales(_xLeft * 1e-6, _xRight * 1e-6);
        Plot->ThreeDPlots[i]->SetPlotsYScales(_yBottom * 1e-6, _yTop * 1e-6);
        Plot->ThreeDPlots[i]->ResizeQPlot();
        Plot->ThreeDPlots[i]->PlotWasChanged();
    }
    Plot->Refresh();
    qApp->processEvents(); //flush event filter

    Plot->TwoDPlot->Pure2dPlot->SetDefaultXLimits(0, qMax(xm, ym) * 1e-6);

    xLeft = _xLeft;
    xRight = _xRight;
    yTop = _yTop;
    yBottom = _yBottom;
    LEdDeltaX->setValue(xRight - xLeft);
    LEdDeltaY->setValue(yTop - yBottom);

    check_speed();
}

void ScanPlot::sendXYsync(double x, double y, int sc)
{
    if (sc == scHead) {
        emit sync_scan_x(x);
        emit sync_scan_y(y);
    } else {
        emit sync_scan_x2(x);
        emit sync_scan_y2(y);
    }
}


void ScanPlot::Send2dPlotData()
{
    send2dPlot(PlotID::ScanSection, Plot->TwoDPlot);
}
void ScanPlot::Send3dPlotData1()
{
    send3dPlot(PlotID::Scan1, Plot, 0);
    if (refresh_limit_need_math_update[0]) {
        refresh_limit_need_math_update[0] = false;
        refresh_limit_last_elapsed[0] = refresh_3d_limit_timer[0].elapsed() / 0.8 * channelsActive; //time budget, %
    }
}
void ScanPlot::Send3dPlotData2()
{
    send3dPlot(PlotID::Scan2, Plot, 1);
    if (refresh_limit_need_math_update[1]) {
        refresh_limit_need_math_update[1] = false;
        refresh_limit_last_elapsed[1] = refresh_3d_limit_timer[1].elapsed() / 0.8 * channelsActive; //time budget, %
    }
}
void ScanPlot::Send3dPlotData3()
{
    //send3dPlot(PlotID::Scan3, Plot, 2);
    if (refresh_limit_need_math_update[2]) {
        refresh_limit_need_math_update[2] = false;
        refresh_limit_last_elapsed[2] = refresh_3d_limit_timer[2].elapsed() / 0.8 * channelsActive; //time budget, %
    }
}
void ScanPlot::Send3dPlotData4()
{
    //send3dPlot(PlotID::Scan4, Plot, 3);
    if (refresh_limit_need_math_update[3]) {
        refresh_limit_need_math_update[3] = false;
        refresh_limit_last_elapsed[3] = refresh_3d_limit_timer[3].elapsed() / 0.8 * channelsActive; //time budget, %
    }
}

#include <qwt_plot_canvas.h>
void ScanPlot::BSendToGraphEditor_clicked()
{
    emit SendBackground(QPixmap::grabWidget(Plot->ThreeDPlots.at(Plot->ActivePlot)->QPlot->canvas()).toImage());
    emit SendRect(plot::ConstructRectF(Plot->ThreeDPlots.at(Plot->ActivePlot)->CurrentXLimits,
                                 Plot->ThreeDPlots.at(Plot->ActivePlot)->CurrentYLimits));
}


void ScanPlot::scan_first_delay(double time)
{
    m_scan_first_delay = time*1000.;
    updateFirstDelay = true;
}
void ScanPlot::scan_line_delay(double time)
{
    m_scan_line_delay = time*1000.;
    updateLineDelay = true;
}
void ScanPlot::scan_point_delay(double time)
{
    m_scan_point_delay = time*1000.*10.;
    updatePointDelay = true;
}
void ScanPlot::scan_fast_return(bool on)
{
    m_scan_fast_return = on;
    updateFastReturn = true;
}
void ScanPlot::scan_sync_mode(int mode)
{
    m_scan_sync_mode = mode;
    updateSyncMode = true;
}
void ScanPlot::spm_pid_precision(double delta)
{
    m_spm_pid_precision = 65535. * delta / 20. ;
    updateSPMPrecision = true;
}
void ScanPlot::scan_lm_height(double h)
{
    scan_lm_h = h/1000.;
    sendLMheight = true;
}

#include <i_eg3000.h>
bool ScanPlot::sendData(const void *handle)
{
    bool wrote = false;

    if (updateFirstDelay.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_FIRST_DELAY, m_scan_first_delay);
    }
    if (updateLineDelay.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_LINE_DELAY, m_scan_line_delay);
    }
    if (updatePointDelay.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_SCAN_POINT_DELAY, m_scan_point_delay);
    }
    if (updateFastReturn.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_FAST_RETURN, m_scan_fast_return ? 1 : 0);
    }
    if (updateSyncMode.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_SCAN_WAIT_SYNC, m_scan_sync_mode);
    }
    if (updateSPMPrecision.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_SPM_PRECISION, m_spm_pid_precision);
    }

    if (sendSpeed.claimMessage()) {
        wrote = true;
        eg3000_scan_speed((void *)handle, scanner_scan, requested_scan_speed);
    }
    if (sendLMheight.claimMessage()) {
        wrote = true;
        eg3000_send_packet(handle, ADR_CPU, CPU_SET_SCAN_LM_HEIGHT, &scan_lm_h, sizeof(float));
    }
    if (sendConfig.claimMessage()) {
        wrote = true;
        eg3000_scan_params((void *)handle, scanner_scan, Nx_points, Ny_points, requested_scan_speed,
                                Step_X,  Step_Y,  stPointX,  stPointY,
                                scan_dir_st_point, scan_ch[0], scan_ch[1], scan_ch[2], scan_ch[3],
                                LmEn, LmEn_ch[0], LmEn_ch[1], LmEn_ch[2], LmEn_ch[3], LmHeight, fast_scan_correction, scan_precise_mode);
    }
    if (sendExtConfig.claimMessage()) {
        wrote = true;
        eg3000_ccd_scan_params((void *)handle, scanner_scan, ccdMode, ccdNX, ccdNY, Step_X, Step_Y, stPointX, stPointY, scan_dir_st_point);
    }
    if (sendStartStop.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, (scanner_scan == scStage) ? CPU_SCAN_ON_OFF_STAGE : CPU_SCAN_ON_OFF, scan_on_off);
    }
    if (sendExtStartStop.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_SCAN_ON_OFF_CCD, scan_on_off);
    }
    if (sendExtReady.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_CCD_READY, 1);
    }

    return wrote;
}
void ScanPlot::resetWriter()
{
    updateSPMPrecision = false;
    updateFirstDelay = false;
    updateLineDelay = false;
    updatePointDelay = false;
    updateFastReturn = false;
    updateSyncMode = false;

    sendSpeed = false;
    sendLMheight = false;
    sendConfig = scan_on_off != 0;
    sendExtConfig = sendConfig && ccdMode != 0;
    sendStartStop = false;
    sendExtStartStop = false;
    sendExtReady = false;

    QMetaObject::invokeMethod(this, "reset");
}
void ScanPlot::reset()
{
    if (m_BStart->on_off)
        m_BStart->change_state_man_with_sig(false);

    spmPrecision->setNewValue();
    firstDelay->setNewValue();
    lineDelay->setNewValue();
    pointDelay->setNewValue();
    pointSync->setCurrentIndex(0);
    scan_sync_mode(0);
    scan_fast_return(fastReturn->isChecked());

    LEd_LFmodeHeight->setNewValue();

    for (int i = 0; i < 4; i++) {
        scanData[0][i].resize(qMin(scanData[0][i].size(), 1));
        scanData[1][i].resize(qMin(scanData[1][i].size(), 1));
    }
    Plot->Refresh();
}

void ScanPlot::capBits(int n, long bits)
{
    if (n == 0) {
        l2k_controller = bits & 4;
        l2k_controller_v22 = bits & 16;
        if (l2k_controller || l2k_controller_v22)
            check_speed();
    }
    if (n >= 2 && n <= 4) {
        is_optical_scanner[n-2] = bool(bits & 1);
    }
}

//emulator --------------------
#ifdef USE_EG3K_EMULATOR
static int sc_ch;
static bool ccdswitch;
static int ccdCounter;
static QElapsedTimer etmr;
static double test_plane_func(double xp, double yp)
{
    double a0=1000, a1=200, a2=200;
    return a0 + a1*xp + a2*yp;
}
#endif
void ScanPlot::reset_scan_test()
{
#ifdef USE_EG3K_EMULATOR
    QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
    if (!settings.value("usb", false).toBool())
        return;

    sc_ch = 0;
    ccdswitch = ccdMode != 0;
    ccdCounter = 0;
    sendExtReady = true;
    etmr.start();
#endif
}
void ScanPlot::injectTestData()
{
#ifdef USE_EG3K_EMULATOR
    if (!scan_on_off)
        return;

    if (ccdswitch) {
        if (!sendExtReady)
            return;
        sendExtReady = false;
        if (ccdCounter < ccdNX*ccdNY) {
            processPacket(ADR_CPU, CPU_CCD_START, 0, 0);
            if (ccdCounter > 0 && ccdCounter % (scan_direction ? ccdNY : ccdNX) == 0)
                ccdswitch = false;
            ccdCounter++;
            if (ccdswitch)
                return;
        } else
            ccdswitch = false;
    }

    int scan_wait = qMax<int>(1, 1000/4/user_scan_speed);
    if (etmr.elapsed() < scan_wait)
        return;
    etmr.restart();

    int prev_sc_ch = sc_ch;
    sc_ch++;
    if (sc_ch > 3) {
        sc_ch = 0;
        if (ccdMode != 0) {
            ccdswitch = true;
            sendExtReady = true;
        }
    }

    EG_ScanData Scan_Data;
    Scan_Data.Source = scan_ch[prev_sc_ch];
    if (Scan_Data.Source == srcNone)
        return;

    Scan_Data.Axis = scan_direction;
    Scan_Data.Direction = prev_sc_ch >= 2 ? dirBW : dirFW;
    Scan_Data.Channel = prev_sc_ch % 2;
    Scan_Data.Scanner = scanner_scan;
    Scan_Data.Line = line_number[prev_sc_ch];
    Scan_Data.Size = qMin<int>(1000, (scan_direction ? Ny_points : Nx_points));
    Scan_Data.ScanSize = qMin<int>(1000, (scan_direction ? Nx_points : Ny_points));
    WORD line_shift = (WORD)(rand() % 5000);
    for (int i = 0; i < Scan_Data.Size; i++)
    {
        double cur_x_point = xLeft;
        double cur_y_point = yBottom;
        if (scan_direction) {
            cur_x_point += line_number[prev_sc_ch] * Step_X;
            cur_y_point += i * Step_Y;
        } else {
            cur_x_point += i * Step_X;
            cur_y_point += line_number[prev_sc_ch] * Step_Y;
        }

        double ca[4] = {1, 0.5, 0.25, 0.125};
        double c = ca[prev_sc_ch];
        WORD func = 10000. *
                sin(c * cur_x_point) * sin(c * cur_x_point) *
                sin(c * cur_y_point) * sin(c * cur_y_point);
        func += test_plane_func(cur_x_point, cur_y_point) + line_shift;
        Scan_Data.Data[i] = func;
    }

    processPacket(ADR_CPU, CPU_SCANDATA, &Scan_Data, 0);
#endif
}
